Package: BEND
Title: Bayesian Estimation of Nonlinear Data (BEND)
Version: 2.0.1
Authors@R: c(
  person("Corissa T.", "Rohloff", email = "corissa.wurth@gmail.com", role = c("aut","cre"), comment = c(ORCID = "0000-0003-3228-4653")),
  person("Rik", "Lamm", email = "rzlamm@gmail.com", role = "aut", comment = c(ORCID = "0000-0002-3317-6243")),
  person("Yadira", "Peralta", email = "yadira.peralta@cide.edu", role = "aut", comment = c(ORCID = "0000-0003-4823-6939")),
  person("Nidhi", "Kohli", email = "nkohli@umn.edu", role = "aut", comment = c(ORCID = "0000-0003-4690-2854")),
  person("Eric F.", "Lock", email = "elock@umn.edu", role = "aut", comment = c(ORCID = "0000-0003-4663-2356")))
Description: Provides a set of models to estimate nonlinear longitudinal data using Bayesian estimation methods. These models include the: 1) Bayesian Piecewise Random Effects Model (Bayes_PREM()) which estimates a piecewise random effects (mixture) model for a given number of latent classes and a latent number of possible changepoints in each class, and can incorporate class and outcome predictive covariates (see Lamm (2022) <https://hdl.handle.net/11299/252533> and Lock et al., (2018) <doi:10.1007/s11336-017-9594-5>), 2) Bayesian Crossed Random Effects Model (Bayes_CREM()) which estimates a linear, quadratic, exponential, or piecewise crossed random effects models where individuals are changing groups over time (e.g., students and schools; see Rohloff et al., (2024) <doi:10.1111/bmsp.12334>), and 3) Bayesian Bivariate Piecewise Random Effects Model (Bayes_BPREM()) which estimates a bivariate piecewise random effects model to jointly model two related outcomes (e.g., reading and math achievement; see Peralta et al., (2022) <doi:10.1037/met0000358>). 
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.2
URL: https://github.com/crohlo/BEND
BugReports: https://github.com/crohlo/BEND/issues
Depends: R (>= 3.6.3)
Imports: coda (>= 0.19.4.1), label.switching (>= 1.8), rjags (>= 4.17)
LazyData: true
NeedsCompilation: no
Packaged: 2026-04-19 03:08:06 UTC; CorissaW
Author: Corissa T. Rohloff [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-3228-4653>),
  Rik Lamm [aut] (ORCID: <https://orcid.org/0000-0002-3317-6243>),
  Yadira Peralta [aut] (ORCID: <https://orcid.org/0000-0003-4823-6939>),
  Nidhi Kohli [aut] (ORCID: <https://orcid.org/0000-0003-4690-2854>),
  Eric F. Lock [aut] (ORCID: <https://orcid.org/0000-0003-4663-2356>)
Maintainer: Corissa T. Rohloff <corissa.wurth@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-19 03:20:02 UTC
Built: R 4.6.0; ; 2026-04-25 23:53:56 UTC; unix

Package: BEND
Title: Bayesian Estimation of Nonlinear Data (BEND)
Version: 2.1.0
Authors@R: c(
  person("Corissa T.", "Rohloff", email = "corissa.wurth@gmail.com", role = c("aut","cre"), comment = c(ORCID = "0000-0003-3228-4653")),
  person("Rik", "Lamm", email = "rzlamm@gmail.com", role = "aut", comment = c(ORCID = "0000-0002-3317-6243")),
  person("Yadira", "Peralta", email = "yadira.peralta@cide.edu", role = "aut", comment = c(ORCID = "0000-0003-4823-6939")),
  person("Nidhi", "Kohli", email = "nkohli@umn.edu", role = "aut", comment = c(ORCID = "0000-0003-4690-2854")),
  person("Eric F.", "Lock", email = "elock@umn.edu", role = "aut", comment = c(ORCID = "0000-0003-4663-2356")))
Description: Provides a set of models to estimate nonlinear longitudinal data using Bayesian estimation methods. These models include the: 1) Bayesian Piecewise Random Effects Model (Bayes_PREM()) which estimates a piecewise random effects (mixture) model for a given number of latent classes and a latent number of possible changepoints in each class, and can incorporate class and outcome predictive covariates (see Lamm (2022) <https://hdl.handle.net/11299/252533> and Lock et al., (2018) <doi:10.1007/s11336-017-9594-5>), 2) Bayesian Crossed Random Effects Model (Bayes_CREM()) which estimates a linear, quadratic, exponential, or piecewise crossed random effects models where individuals are changing groups over time (e.g., students and schools; see Rohloff et al., (2024) <doi:10.1111/bmsp.12334>), and 3) Bayesian Bivariate Piecewise Random Effects Model (Bayes_BPREM()) which estimates a bivariate piecewise random effects model to jointly model two related outcomes (e.g., reading and math achievement; see Peralta et al., (2022) <doi:10.1037/met0000358>). 
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.2
URL: https://github.com/crohlo/BEND
BugReports: https://github.com/crohlo/BEND/issues
Depends: R (>= 3.6.3)
Imports: coda (>= 0.19.4.1), label.switching (>= 1.8), rjags (>= 4.17)
LazyData: true
NeedsCompilation: no
Packaged: 2026-07-03 19:00:46 UTC; CorissaW
Author: Corissa T. Rohloff [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-3228-4653>),
  Rik Lamm [aut] (ORCID: <https://orcid.org/0000-0002-3317-6243>),
  Yadira Peralta [aut] (ORCID: <https://orcid.org/0000-0003-4823-6939>),
  Nidhi Kohli [aut] (ORCID: <https://orcid.org/0000-0003-4690-2854>),
  Eric F. Lock [aut] (ORCID: <https://orcid.org/0000-0003-4663-2356>)
Maintainer: Corissa T. Rohloff <corissa.wurth@gmail.com>
Repository: CRAN
Date/Publication: 2026-07-10 05:50:20 UTC

# Generated by roxygen2: do not edit by hand

S3method(getClassProb,PREM)
S3method(getCoef,BEND)
S3method(getFitted,BEND)
S3method(getFixEf,BPREM)
S3method(getFixEf,CREM)
S3method(getFixEf,PREM)
S3method(getKProb,PREM)
S3method(getModelFit,BEND)
S3method(getRanEf,BPREM)
S3method(getRanEf,CREM)
S3method(getVarCov,BPREM)
S3method(getVarCov,CREM)
S3method(getVarCov,PREM)
S3method(plot,BPREM)
S3method(plot,CREM)
S3method(plot,PREM)
S3method(print,BPREM)
S3method(print,CREM)
S3method(print,PREM)
S3method(print,getClassProb.PREM)
S3method(print,getCoef)
S3method(print,getFitted)
S3method(print,getFixEf.BEND)
S3method(print,getKProb.PREM)
S3method(print,getModelFit.BEND)
S3method(print,getRanEf.BPREM)
S3method(print,getRanEf.CREM)
S3method(print,getVarCov.BPREM)
S3method(print,getVarCov.CREM)
S3method(print,getVarCov.PREM)
S3method(print,summary.BPREM)
S3method(print,summary.CREM)
S3method(print,summary.PREM)
S3method(summary,BPREM)
S3method(summary,CREM)
S3method(summary,PREM)
export(Bayes_BPREM)
export(Bayes_CREM)
export(Bayes_PREM)
export(getClassProb)
export(getCoef)
export(getFitted)
export(getFixEf)
export(getKProb)
export(getModelFit)
export(getRanEf)
export(getVarCov)
export(plot_BEND)
import(graphics)
import(stats)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# BEND

<!-- badges: start -->

<!-- badges: end -->

The goal of BEND is to provide a set of models to estimate nonlinear
longitudinal data using Bayesian estimation methods. These models
include the:

1.  Bayesian Piecewise Random Effects Model (`Bayes_PREM()`) which
    estimates a piecewise random effects (mixture) model for a given
    number of latent classes and a latent number of possible
    changepoints in each class, and can incorporate class and outcome
    predictive covariates (see Lamm, 2022 and Lock et al., 2018 for more
    details).

2.  Bayesian Crossed Random Effects Model (`Bayes_CREM()`) which
    estimates a linear, quadratic, exponential, or piecewise crossed
    random effects models where individuals are changing groups over
    time (e.g., students and schools; see Rohloff et al., 2024 for more
    details).

3.  Bayesian Bivariate Piecewise Random Effects Model (`Bayes_BPREM()`)
    which estimates a bivariate piecewise random effects model to
    jointly model two related outcomes (e.g., reading and math
    achievement; see Peralta et al., 2022 for more details).

This package requires Just Another Gibbs Sampler (JAGS) to be installed
on your computer (<https://mcmc-jags.sourceforge.io/>), and depends on
the packages `rjags` and `label.switching`.

### References

Lamm, R. (2022). Incorporation of covariates in Bayesian piecewise
growth mixture models. <https://hdl.handle.net/11299/252533>

Lock, E. F., Kohli, N., & Bose, M. (2018). Detecting multiple random
changepoints in Bayesian piecewise growth mixture models. Psychometrika,
83(3), 733–750. <https://doi.org/10.1007/s11336-017-9594-5>

Peralta, Y., Kohli, N., Lock, E. F., & Davison, M. L. (2022). Bayesian
modeling of associations in bivariate piecewise linear mixed-effects
models. Psychological Methods, 27(1), 44–64.
<https://doi.org/10.1037/met0000358>

Rohloff, C. T., Kohli, N., & Lock, E. F. (2024). Identifiability and
estimability of Bayesian linear and nonlinear crossed random effects
models. British Journal of Mathematical and Statistical Psychology.
<https://doi.org/10.1111/bmsp.12334>

## Installation

You can install the development version of BEND from
[GitHub](https://github.com/) with:

``` r
# install.packages("devtools")
library(devtools)
install_github("crohlo/BEND")
```

or directly from the [CRAN](https://cran.r-project.org/) repository
with:

``` r
install.packages("BEND")
```

#' Summarize the results of a piecewise random effects model (PREM)
#'
#' @description
#' Provides a summary of a PREM model, as returned by `Bayes_PREM()`.
#'
#' @param object An object of class "PREM" (returned by `Bayes_PREM(...)`).
#' @param ... Additional arguments.
#'
#' @returns Returns a list of key parameter estimates.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_prem)
#' # result summary
#' summary(results_prem)
#'
#' @export
summary.PREM <- function(object, ...){

  x <- object

  # determine number of classes
  n_class <- length(unique(x$Class_Information$class_membership))

  # determine number of changepoints in each class (based on final model results)
  changepoints <- c()
  class_data <- data.frame()
  for(i in 1:n_class){
    class_num <- paste0("Class_", i)
    changepoints[i] <- which.max(x$Parameter_Estimates[[class_num]]$K_prob)-1
  }
  max_cp <- max(changepoints)

  # detect covariates
  n_cov_op <- length(x$Parameter_Estimates$outcome_predictive_covariates)
  n_cov_cp <- length(x$Parameter_Estimates$class_predictive_covariates)

  # get rownames
  my_rownames <- rep(NA, 4+4*max_cp)
  my_rownames[1] = "Intercept Mean"
  my_rownames[2] = "Slope Mean"
  my_rownames[max_cp*2+2+1] = "Intercept Var"
  my_rownames[max_cp*2+2+2] = "Slope Var"
  for(k in 1:max_cp){
    my_rownames[2*k+1] = paste0("Changepoint ", k, " Mean")
    my_rownames[2*k+2] = paste0("Change in Slope ", k, " Mean")
    my_rownames[k*2+(max_cp*2+2+1)] = paste0("Changepoint ", k, " Var")
    my_rownames[k*2+(max_cp*2+2+2)] = paste0("Change in Slope ", k, " Var")
  }

  # CLASS DEPENDENT PARAMETERS -----

  # pull parameter estimates for each class
  coeff_mat_list <- vector("list", n_class)
  for(i in 1:n_class){
    class_num <- paste0("Class_", i)
    cp_num <- paste0("K_", changepoints)[i]
    cur_mat <- matrix(NA, nrow=4+4*max_cp, ncol=3)
    # fixed effects
    cur_mat[1,1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[1]
    cur_mat[1,2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean_CI[1,1]
    cur_mat[1,3] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean_CI[2,1]
    cur_mat[2,1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[2]
    cur_mat[2,2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean_CI[1,2]
    cur_mat[2,3] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean_CI[2,2]
    # random effects
    cur_mat[max_cp*2+2+1,1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var[1]
    cur_mat[max_cp*2+2+1,2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var_CI[1,1]
    cur_mat[max_cp*2+2+1,3] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var_CI[2,1]
    cur_mat[max_cp*2+2+2,1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var[2]
    cur_mat[max_cp*2+2+2,2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var_CI[1,2]
    cur_mat[max_cp*2+2+2,3] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var_CI[2,2]
    # if there are changepoints (beyond linear)
    if(changepoints[i]>0){
      for(k in 1:changepoints[i]){
        # fixed effects
        cur_mat[2*k+1,1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_mean[k]
        cur_mat[2*k+1,2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_mean_CI[1,k]
        cur_mat[2*k+1,3] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_mean_CI[2,k]
        cur_mat[2*k+2,1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[k+2]
        cur_mat[2*k+2,2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean_CI[1,k+2]
        cur_mat[2*k+2,3] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean_CI[2,k+2]
        # random effects
        cur_mat[k*2+(max_cp*2+2+1),1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_var[k]
        cur_mat[k*2+(max_cp*2+2+1),2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_var_CI[1,k]
        cur_mat[k*2+(max_cp*2+2+1),3] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_var_CI[2,k]
        cur_mat[k*2+(max_cp*2+2+2),1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var[k+2]
        cur_mat[k*2+(max_cp*2+2+2),2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var_CI[1,k+2]
        cur_mat[k*2+(max_cp*2+2+2),3] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var_CI[2,k+2]
      }
    }
    colnames(cur_mat) <- c("Estimate", "Lower CI", "Upper CI")
    rownames(cur_mat) <- paste0("Class ", i, ": ", my_rownames)
    coeff_mat_list[[i]] <- cur_mat
  }
  coeff_mat <- do.call(rbind, coeff_mat_list)

  # CLASS INDEPENDENT PARAMETERS -----
  coeff_addit <- cbind(x$Parameter_Estimates$error_var, t(x$Parameter_Estimates$error_var_CI))
  rownames(coeff_addit) <- "Error Var"

  # Covariates
  if(n_cov_op > 0){
    coeff_cov_op <- cbind(as.matrix(x$Parameter_Estimates$outcome_predictive_covariates), t(x$Parameter_Estimates$outcome_predictive_covariates_CI))
    coeff_addit <- rbind(coeff_addit, coeff_cov_op)
  }
  if(n_cov_cp > 0){
    coeff_cov_cp <- cbind(as.matrix(x$Parameter_Estimates$class_predictive_covariates), t(x$Parameter_Estimates$class_predictive_covariates_CI))
    rownames(coeff_cov_cp) <- paste0(rownames(coeff_cov_cp), " (in log-odds units)")
    coeff_log_int <- cbind(x$Parameter_Estimates$logistic_intercept, t(x$Parameter_Estimates$logistic_intercept_CI[,1]))
    rownames(coeff_log_int) <- "Logistic Intercept"
    coeff_addit <- rbind(coeff_addit, coeff_cov_cp, coeff_log_int)
  }
  colnames(coeff_addit) <- c("Estimate", "Lower CI", "Upper CI")

  ## return value
  out <- list("class_dep_params" = coeff_mat,
              "class_ind_params" = coeff_addit,
              "msrf" = x$Convergence$multivariate_psrf,
              "mean_psrf" = x$Convergence$mean_psrf,
              "DIC" = x$Model_Fit$dic)
  class(out) <- c("summary.PREM", class(out))
  return(out)
}

#' @rdname summary.PREM
#' @param x An object of class "summary.PREM".
#' @export
print.summary.PREM <- function(x, ...){

  cat("Class Dependent Parameters:\n")
  print(round(x$class_dep_params,3), na.print="")
  cat("\n")
  cat("Class Independent Parameters:\n")
  print(round(x$class_ind_params,3))
  cat("\n")
  cat("Gelman's msrf:", round(x$msrf, 3), "\n")
  cat("Mean psrf:", round(x$mean_psrf, 3), "\n")
  cat("DIC:", round(x$DIC,3))
  cat("\n")

  invisible(x)
}

# Additional Functions ----

## Permute Changepoint Labels ----
Permute_CP <- function(output, n_class, max_cp, max_beta){
  # output = output from JAGS
  # n_class = number of latent clases
  # max_cp = maximum number of changepoints
  # max_beta = maximum number of intercept + slope parameters

  ### Setup -----
  n_iter = dim(output$sigma2_error)[2]
  n_chain = dim(output$sigma2_error)[3]

  ### Permute changepoint labels, if necessary, so they are ordered -----
  for(i in 1:n_iter){
    for(j in 1:n_chain){
      for(k in 1:n_class){
        if(output$n_cp[k,i,j]>0){
          ind <- c(1:max_cp)[as.logical(output$cp_indicator[k,,i,j])]
          my_order <- order(output$cp[k,ind,i,j])
          output$cp_indicator[k,,i,j] <- c(output$cp_indicator[k,ind[my_order],i,j],output$cp_indicator[k,-ind,i,j])
          output$cp[k,,i,j] <- c(output$cp[k,ind[my_order],i,j],output$cp[k,-ind,i,j])
          output$beta[k,3:max_beta,i,j] <- c(output$beta[k,ind[my_order]+2,i,j],output$beta[k,c(3:max_beta)[-ind],i,j])
        }
      }
    }
  }
  return(output)
}


## Get Mode (for initializing number of changepoints) ----
getmode <- function(v){
  # v = vector

  uniqv <- unique(v)
  uniqv[which.max(tabulate(match(v, uniqv)))]
}


## Realign ECR Function ----
Realign_ECR <- function(output, n_class, model=NULL){
  # output = output from JAGS
  # n_class = number of latent clases
  # model = model type ("PREMM", "CI_PREMM_Full", "CI_PREMM_Class_Predictive" or "CI_PREMM_Outcome_Predictive")

  ### Warnings -----
  if(is.null(model)) stop('Specify model type. Must be \'PREMM\', \'CI_PREMM_Full\', \'CI_PREMM_Class_Predictive\', or \'CI_PREMM_Outcome_Predictive\'')

  ### Setup -----
  n_subj <- dim(output$class)[1]
  n_iter <- dim(output$class)[2]
  n_chain <- dim(output$class)[3]
  if(model=="CI_PREMM_Full" | model=="CI_PREMM_Class_Predictive"){class_pred <- TRUE} else {class_pred <- FALSE}
  ## For class_pred == TRUE
  if(class_pred == TRUE){output$class_orig <- output$class}

  ### Align for the first time -----
  class_matrix <- matrix(nrow=n_iter*n_chain, ncol=n_subj)
  for(i in 1:n_subj){
    class_matrix[,i] <- as.vector(output$class[i,,])
  }
  permutes <- label.switching::ecr.iterative.1(class_matrix, n_class)$permutations

  perm_chain <- list()
  perm_chain[[1]] <- permutes[1:n_iter,]
  perm_chain[[2]] <- permutes[(n_iter+1):(2*n_iter),]
  perm_chain[[3]] <- permutes[(2*n_iter+1):(3*n_iter),]
  for(i in 1:n_chain){
    for(j in 1:n_iter){
      # this loop will ignore parameters that don't exist in output
      ## Initialization-Only
      output$beta[,,j,i] <- output$beta[perm_chain[[i]][j,],,j,i]
      output$cp[,,j,i] <- output$cp[perm_chain[[i]][j,],,j,i]
      ## PREMM only
      output$class_prob[,j,i] <- output$class_prob[perm_chain[[i]][j,],j,i]
      ## Full Model
      output$beta_mean[,,j,i] <- output$beta_mean[perm_chain[[i]][j,],,j,i]
      output$beta_sd[,,j,i] <- output$beta_sd[perm_chain[[i]][j,],,j,i]
      output$cp_mean[,,j,i] <- output$cp_mean[perm_chain[[i]][j,],,j,i]
      output$cp_sd[,,j,i] <- output$cp_sd[perm_chain[[i]][j,],,j,i]
      ## ALL
      output$cp_indicator[,,j,i] <- output$cp_indicator[perm_chain[[i]][j,],,j,i]
      output$n_cp[,j,i] <- output$n_cp[perm_chain[[i]][j,],j,i]
      inv_perm <- order(perm_chain[[i]][j,])
      output$class[,j,i] <- inv_perm[output$class[,j,i]]
    }
  }

  ### Align for a second time (for better accuracy) -----
  class_matrix <- matrix(nrow=n_iter*n_chain, ncol=n_subj)
  for(i in 1:n_subj){
    class_matrix[,i] <- as.vector(output$class[i,,])
  }
  permutes <- label.switching::ecr.iterative.1(class_matrix, n_class)$permutations

  perm_chain <- list()
  perm_chain[[1]] <- permutes[1:n_iter,]
  perm_chain[[2]] <- permutes[(n_iter+1):(2*n_iter),]
  perm_chain[[3]] <- permutes[(2*n_iter+1):(3*n_iter),]
  for(i in 1:n_chain){
    for(j in 1:n_iter){
      # this loop will ignore parameters that don't exist in output
      ## Initialization-Only
      output$beta[,,j,i] <- output$beta[perm_chain[[i]][j,],,j,i]
      output$cp[,,j,i] <- output$cp[perm_chain[[i]][j,],,j,i]
      ## PREMM only
      output$class_prob[,j,i] <- output$class_prob[perm_chain[[i]][j,],j,i]
      ## Full Model
      output$beta_mean[,,j,i] <- output$beta_mean[perm_chain[[i]][j,],,j,i]
      output$beta_sd[,,j,i] <- output$beta_sd[perm_chain[[i]][j,],,j,i]
      output$cp_mean[,,j,i] <- output$cp_mean[perm_chain[[i]][j,],,j,i]
      output$cp_sd[,,j,i] <- output$cp_sd[perm_chain[[i]][j,],,j,i]
      ## ALL
      output$cp_indicator[,,j,i] <- output$cp_indicator[perm_chain[[i]][j,],,j,i]
      output$n_cp[,j,i] <- output$n_cp[perm_chain[[i]][j,],j,i]
      inv_perm <- order(perm_chain[[i]][j,])
      output$class[,j,i] <- inv_perm[output$class[,j,i]]
    }
  }

  ### Invert logistic regression models if the class was switched -----
  if(class_pred == TRUE){
    n_cov <- length(output$class_predictive_covariate_lambda[,1,1])
    for(i in 1:n_iter){
      for(j in 1:n_chain){
        for(k in 1:n_cov){

          if(output$class[1,i,j] != output$class_orig[1,i,j]){
            cp_class1 = output$cp_indicator[1,,i,j]
            cp_class2 = output$cp_indicator[2,,i,j]

            output$cp_indicator[1,,i,j] = cp_class2
            output$cp_indicator[2,,i,j] = cp_class1

            output$logistic_intercept[1,i,j] = -output$logistic_intercept[1,i,j]
            output$class_predictive_covariate_lambda[k,i,j] = -output$class_predictive_covariate_lambda[k,i,j]
          }

        }
      }
    }
  }
  return(output)
}

#' Summarize the results of a piecewise random effects model (PREM)
#'
#' @description
#' Provides a summary of a PREM model, as returned by `Bayes_PREM()`.
#'
#' @param object An object of class "PREM" (returned by `Bayes_PREM(...)`).
#' @param ... Additional arguments.
#'
#' @returns Returns a list of key parameter estimates.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_prem)
#' # result summary
#' summary(results_prem)
#'
#' @rdname summary.PREM
#' @export
summary.PREM <- function(object, ...){

  x <- object

  # determine number of classes
  n_class <- length(unique(x$Class_Information$class_membership))

  # determine number of changepoints in each class (based on final model results)
  changepoints <- c()
  class_data <- data.frame()
  for(i in 1:n_class){
    class_num <- paste0("Class_", i)
    changepoints[i] <- which.max(x$Parameter_Estimates[[class_num]]$K_prob)-1
  }
  max_cp <- max(changepoints)

  # detect covariates
  n_cov_op <- length(x$Parameter_Estimates$outcome_predictive_covariates)
  n_cov_cp <- length(x$Parameter_Estimates$class_predictive_covariates)

  # get rownames
  my_rownames <- rep(NA, 4+4*max_cp)
  my_rownames[1] = "Intercept Mean"
  my_rownames[2] = "Slope Mean"
  my_rownames[max_cp*2+2+1] = "Intercept Var"
  my_rownames[max_cp*2+2+2] = "Slope Var"
  for(k in 1:max_cp){
    my_rownames[2*k+1] = paste0("Changepoint ", k, " Mean")
    my_rownames[2*k+2] = paste0("Change in Slope ", k, " Mean")
    my_rownames[k*2+(max_cp*2+2+1)] = paste0("Changepoint ", k, " Var")
    my_rownames[k*2+(max_cp*2+2+2)] = paste0("Change in Slope ", k, " Var")
  }

  # CLASS DEPENDENT PARAMETERS -----

  # pull parameter estimates for each class
  coeff_mat_list <- vector("list", n_class)
  for(i in 1:n_class){
    class_num <- paste0("Class_", i)
    cp_num <- paste0("K_", changepoints)[i]
    cur_mat <- matrix(NA, nrow=4+4*max_cp, ncol=3)
    # fixed effects
    cur_mat[1,1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[1]
    cur_mat[1,2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean_CI[1,1]
    cur_mat[1,3] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean_CI[2,1]
    cur_mat[2,1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[2]
    cur_mat[2,2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean_CI[1,2]
    cur_mat[2,3] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean_CI[2,2]
    # random effects
    cur_mat[max_cp*2+2+1,1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var[1]
    cur_mat[max_cp*2+2+1,2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var_CI[1,1]
    cur_mat[max_cp*2+2+1,3] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var_CI[2,1]
    cur_mat[max_cp*2+2+2,1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var[2]
    cur_mat[max_cp*2+2+2,2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var_CI[1,2]
    cur_mat[max_cp*2+2+2,3] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var_CI[2,2]
    # if there are changepoints (beyond linear)
    if(changepoints[i]>0){
      for(k in 1:changepoints[i]){
        # fixed effects
        cur_mat[2*k+1,1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_mean[k]
        cur_mat[2*k+1,2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_mean_CI[1,k]
        cur_mat[2*k+1,3] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_mean_CI[2,k]
        cur_mat[2*k+2,1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[k+2]
        cur_mat[2*k+2,2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean_CI[1,k+2]
        cur_mat[2*k+2,3] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean_CI[2,k+2]
        # random effects
        cur_mat[k*2+(max_cp*2+2+1),1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_var[k]
        cur_mat[k*2+(max_cp*2+2+1),2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_var_CI[1,k]
        cur_mat[k*2+(max_cp*2+2+1),3] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_var_CI[2,k]
        cur_mat[k*2+(max_cp*2+2+2),1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var[k+2]
        cur_mat[k*2+(max_cp*2+2+2),2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var_CI[1,k+2]
        cur_mat[k*2+(max_cp*2+2+2),3] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var_CI[2,k+2]
      }
    }
    colnames(cur_mat) <- c("Estimate", "Lower CI", "Upper CI")
    rownames(cur_mat) <- paste0("Class ", i, ": ", my_rownames)
    coeff_mat_list[[i]] <- cur_mat
  }
  coeff_mat <- do.call(rbind, coeff_mat_list)

  # CLASS INDEPENDENT PARAMETERS -----
  coeff_addit <- cbind(x$Parameter_Estimates$error_var, t(x$Parameter_Estimates$error_var_CI))
  rownames(coeff_addit) <- "Error Var"

  # Covariates
  if(n_cov_op > 0){
    coeff_cov_op <- cbind(as.matrix(x$Parameter_Estimates$outcome_predictive_covariates), t(x$Parameter_Estimates$outcome_predictive_covariates_CI))
    coeff_addit <- rbind(coeff_addit, coeff_cov_op)
  }
  if(n_cov_cp > 0){
    coeff_cov_cp <- cbind(as.matrix(x$Parameter_Estimates$class_predictive_covariates), t(x$Parameter_Estimates$class_predictive_covariates_CI))
    rownames(coeff_cov_cp) <- paste0(rownames(coeff_cov_cp), " (in log-odds units)")
    coeff_log_int <- cbind(x$Parameter_Estimates$logistic_intercept, t(x$Parameter_Estimates$logistic_intercept_CI[,1]))
    rownames(coeff_log_int) <- "Logistic Intercept"
    coeff_addit <- rbind(coeff_addit, coeff_cov_cp, coeff_log_int)
  }
  colnames(coeff_addit) <- c("Estimate", "Lower CI", "Upper CI")

  ## return value
  out <- list("class_dep_params" = coeff_mat,
              "class_ind_params" = coeff_addit,
              "msrf" = x$Convergence$multivariate_psrf,
              "mean_psrf" = x$Convergence$mean_psrf,
              "DIC" = x$Model_Fit$dic)
  class(out) <- c("summary.PREM", "summary.BEND")
  return(out)
}

#' @rdname summary.PREM
#' @param x An object of class "summary.PREM".
#' @export
print.summary.PREM <- function(x, ...){

  cat("Class Dependent Parameters:\n")
  print(round(x$class_dep_params,3), na.print="")
  cat("\n")
  cat("Class Independent Parameters:\n")
  print(round(x$class_ind_params,3))
  cat("\n")
  cat("Gelman's msrf:", round(x$msrf, 3), "\n")
  cat("Mean psrf:", round(x$mean_psrf, 3), "\n")
  cat("DIC:", round(x$DIC,3))
  cat("\n")

  invisible(x)
}

#' Summarize the results of a bivariate piecewise random effects model (BPREM)
#'
#' @description
#' Provides a summary of a BPREM model, as returned by `Bayes_BPREM()`.
#'
#' @param object An object of class "BPREM" (returned by `Bayes_BPREM(...)`).
#' @param ... Additional arguments.
#'
#' @returns Returns a list of key parameter estimates.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_bprem)
#' # result summary
#' summary(results_bprem)
#'
#' @rdname summary.BPREM
#' @export
summary.BPREM <- function(object, ...){

  x <- object

  # determine number of parameters
  n_param <- 4

  # determine number of covariances
  n_covar <- ((n_param*2)*((n_param*2)-1))/2

  # define parameters for labeling
  param_names <- c("Intercept", "Slope", "Change in Slope", "Changepoint")
  names(param_names) <- c("0", "1", "2", "g")

  # FIXED EFFECTS -----
  fix_eff_est <- x$Parameter_Estimates[1:(n_param*2),]
  names(fix_eff_est) <- c("Estimate", "Lower CI", "Upper CI")
  rownames(fix_eff_est) <- paste0(rep(c("Outcome 1: ", "Outcome 2: "), e=4), rep(param_names,2), " Mean")

  # RANDOM EFFECTS -----

  ## COVARIANCE MATRIX -----
  ran_eff_var_mat <- x$Parameter_Estimates[((n_param*2)+(1:(n_param*2))),]
  names(ran_eff_var_mat) <- c("Estimate", "Lower CI", "Upper CI")
  rownames(ran_eff_var_mat) <- paste0(rep(c("Outcome 1: ", "Outcome 2: "), e=4), rep(param_names,2), " Var")

  ran_eff_cov_mat <- x$Parameter_Estimates[(2*(n_param*2)+(1:n_covar)),]
  names(ran_eff_cov_mat) <- c("Estimate", "Lower CI", "Upper CI")
  # get rownames
  ss <- sub("cp", "g", gsub("cov_b_", "", rownames(ran_eff_cov_mat))) ## replace cp with g to make following code simpler
  rownames(ran_eff_cov_mat) <- paste0("Cov(Outcome ", substr(ss,1,1), ": ", param_names[substr(ss,2,2)],
                                      ", Outcome ", substr(ss,4,4), ": ", param_names[substr(ss,5,5)], ")")

  ## CORRELATION MATRIX -----
  ran_eff_corr_mat <- x$Parameter_Estimates[(2*(n_param*2)+n_covar+3+(1:n_covar)),]
  names(ran_eff_corr_mat) <- c("Estimate", "Lower CI", "Upper CI")
  # get rownames
  ss <- sub("cp", "g", gsub("corr_b_", "", rownames(ran_eff_corr_mat))) ## replace cp with g to make following code simpler
  rownames(ran_eff_corr_mat) <- paste0("Corr(Outcome ", substr(ss,1,1), ": ", param_names[substr(ss,2,2)],
                                       ", Outcome ", substr(ss,4,4), ": ", param_names[substr(ss,5,5)], ")")

  # ERROR -----
  error_cov_mat <- x$Parameter_Estimates[(2*(n_param*2)+n_covar+(1:3)),]
  names(error_cov_mat) <- c("Estimate", "Lower CI", "Upper CI")
  rownames(error_cov_mat) <- c("Outcome 1: Error Var", "Outcome 2: Error Var", "Cov(Outcome 1: Error, Outcome 2: Error)")

  error_cor_mat <- x$Parameter_Estimates[(2*(n_param*2)+2*n_covar+3+1),]
  names(error_cor_mat) <- c("Estimate", "Lower CI", "Upper CI")
  rownames(error_cor_mat) <- c("Corr(Outcome 1: Error, Outcome 2: Error)")

  ## return value
  out <- list("data" = x$Call$data,
              "y1_var" = x$Call$y1_var,
              "y2_var" = x$Call$y2_var,
              "fix_eff_est" = fix_eff_est,
              "ran_eff_var_mat" = ran_eff_var_mat,
              "ran_eff_cov_mat" = ran_eff_cov_mat,
              "ran_eff_corr_mat" = ran_eff_corr_mat,
              "error_cov_mat" = error_cov_mat,
              "error_corr" = error_cor_mat,
              "msrf" = x$Convergence$multivariate_psrf,
              "mean_psrf" = x$Convergence$mean_psrf,
              "DIC" = x$Model_Fit$dic)
  class(out) <- c("summary.BPREM", "summary.BEND")
  return(out)
}

#' @rdname summary.BPREM
#' @param x An object of class "summary.BPREM".
#' @export
print.summary.BPREM <- function(x, ...){

  cat("Bayesian bivariate piecewise random effects model\n")
  cat("Data:", x$data, "\n")
  cat("Outcomes:", c(x$y1_var, x$y2_var), "\n")
  cat("\n")
  cat("Fixed Effect Parameters:\n")
  print(round(x$fix_eff_est,3))
  cat("\n")
  cat("Random Effect Parameters:\n")
  cat("Variances:\n")
  print(round(x$ran_eff_var_mat,3), na.print="")
  cat("\n")
  cat("Covariances:\n")
  print(round(x$ran_eff_cov_mat,3), na.print="")
  cat("\n")
  cat("Correlations:\n")
  print(round(x$ran_eff_corr_mat,3), na.print="")
  cat("\n")
  cat("Error:\n")
  cat("Variance-Covariance:\n")
  print(round(x$error_cov_mat,3), na.print="")
  cat("\n")
  cat("Correlation:\n")
  print(round(x$error_corr,3), na.print="")
  cat("\n")
  cat("Gelman's msrf:", round(x$msrf, 3), "\n")
  cat("Mean psrf:", round(x$mean_psrf, 3), "\n")
  cat("DIC:", round(x$DIC,3))
  cat("\n")

  invisible(x)
}

#' Summarize the results of a crossed random effects model (CREM)
#'
#' @description
#' Provides a summary of a CREM model, as returned by `Bayes_CREM()`.
#'
#' @param object An object of class "CREM" (returned by `Bayes_CREM(...)`).
#' @param ... Additional arguments.
#'
#' @returns Returns a list of key parameter estimates.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_pcrem)
#' # result summary
#' summary(results_pcrem)
#'
#' @rdname summary.CREM
#' @export
summary.CREM <- function(object, ...){

  x <- object

  # determine form
  form <- x$Functional_Form

  # determine number of parameters
  if(form=="linear")                           n_param <- 2
  if(form=="quadratic" | form=="exponential")  n_param <- 3
  if(form=="piecewise")                        n_param <- 4

  # determine number of covariances
  n_covar <- (n_param*(n_param-1))/2

  # define parameters for labeling
  if(form=="linear")       param_names <- c("Intercept", "Slope")
  if(form=="linear")       names(param_names) <- c("0", "1")
  if(form=="quadratic")    param_names <- c("Intercept", "Linear Slope", "Quadratic Slope")
  if(form=="quadratic")    names(param_names) <- c("0", "1", "2")
  if(form=="exponential")  param_names <- c("Intercept", "Total Change", "Growth Rate")
  if(form=="exponential")  names(param_names) <- c("0", "1", "2")
  if(form=="piecewise")    param_names <- c("Intercept", "Slope", "Change in Slope", "Changepoint")
  if(form=="piecewise")    names(param_names) <- c("0", "1", "2", "g")

  # FIXED EFFECTS -----
  fix_eff_est <- x$Parameter_Estimates[1:n_param,]
  colnames(fix_eff_est) <- c("Estimate", "Lower CI", "Upper CI")
  rownames(fix_eff_est) <- paste0(param_names, " Mean")

  # RANDOM EFFECTS -----

  ## INDIVIDUALS -----
  ran_eff_b_var_mat <- x$Parameter_Estimates[(n_param+(1:n_param)),]
  colnames(ran_eff_b_var_mat) <- c("Estimate", "Lower CI", "Upper CI")
  rownames(ran_eff_b_var_mat) <- paste0(param_names, " Var")

  ran_eff_b_cov_mat <- x$Parameter_Estimates[(2*n_param+(1:n_covar)),]
  colnames(ran_eff_b_cov_mat) <- c("Estimate", "Lower CI", "Upper CI")
  # get rownames
  ss <- sub("cp", "g", sub("cov_b_", "", rownames(ran_eff_b_cov_mat))) ## replace cp with g to make following code simpler
  rownames(ran_eff_b_cov_mat) <- paste0("Cov(", param_names[substr(ss,1,1)], ", ", param_names[substr(ss,2,2)], ")")

  ## GROUPS -----
  ran_eff_g_var_mat <- x$Parameter_Estimates[(2*n_param+n_covar+(1:n_param)),]
  colnames(ran_eff_g_var_mat) <- c("Estimate", "Lower CI", "Upper CI")
  rownames(ran_eff_g_var_mat) <- paste0(param_names, " Var")

  ran_eff_g_cov_mat <- x$Parameter_Estimates[(3*n_param+n_covar+(1:n_covar)),]
  colnames(ran_eff_g_cov_mat) <- c("Estimate", "Lower CI", "Upper CI")
  # get rownames
  ss <- sub("cp", "g", sub("cov_g_", "", rownames(ran_eff_g_cov_mat))) ## replace cp with g to make following code simpler
  rownames(ran_eff_g_cov_mat) <- paste0("Cov(", param_names[substr(ss,1,1)], ", ", param_names[substr(ss,2,2)], ")")

  ## ERROR -----
  error_mat <- x$Parameter_Estimates[(3*n_param+2*n_covar+1),]
  colnames(error_mat) <- c("Estimate", "Lower CI", "Upper CI")
  rownames(error_mat) <- c("Error Var")

  ## return value
  out <- list("data" = x$Call$data,
              "y_var" = x$Call$y_var,
              "ind_id_var" = x$Call$ind_id_var,
              "cross_id_var" = x$Call$cross_id_var,
              "form" = x$Call$form,
              "fix_eff_est" = fix_eff_est,
              "ran_eff_b_var_mat" = ran_eff_b_var_mat,
              "ran_eff_b_cov_mat" = ran_eff_b_cov_mat,
              "ran_eff_g_var_mat" = ran_eff_g_var_mat,
              "ran_eff_g_cov_mat" = ran_eff_g_cov_mat,
              "error_var" = error_mat,
              "msrf" = x$Convergence$multivariate_psrf,
              "mean_psrf" = x$Convergence$mean_psrf,
              "DIC" = x$Model_Fit$dic)
  class(out) <- c("summary.CREM", "summary.BEND")
  return(out)
}

#' @rdname summary.CREM
#' @param x An object of class "summary.CREM".
#' @export
print.summary.CREM <- function(x, ...){

  cat("Bayesian crossed random effects model\n")
  cat("Data:", x$data, "\n")
  cat("Outcome:", x$y_var, "\n")
  cat("Individuals:", x$ind_id_var, "\n")
  cat("Group:", x$cross_id_var, "\n")
  cat("Functional Form:", x$form, "\n")
  cat("\n")
  cat("Fixed Effect Parameters:\n")
  print(round(x$fix_eff_est,3))
  cat("\n")
  cat("Random Effect Parameters:\n")
  cat("Individual Random Effects Variance-Covariance:\n")
  print(round(rbind(x$ran_eff_b_var_mat, x$ran_eff_b_cov_mat),3), na.print="")
  cat("\n")
  cat("Group Random Effects Variance-Covariance:\n")
  print(round(rbind(x$ran_eff_g_var_mat, x$ran_eff_g_cov_mat),3), na.print="")
  cat("\n")
  cat("Error Variance:\n")
  print(round(x$error_var,3), na.print="")
  cat("\n")
  cat("Gelman's msrf:", round(x$msrf, 3), "\n")
  cat("Mean psrf:", round(x$mean_psrf, 3), "\n")
  cat("DIC:", round(x$DIC,3))
  cat("\n")

  invisible(x)
}

#' Simulated data for a PREM + Extensions
#'
#' Simulated data for a piecewise random effects model (PREM) and useful extensions (CI-PREM, PREMM, CI-PREMM) with 18 timepoints collected on 30 individuals.
#'
#' \itemize{
#'   \item `id` ID for each individual.
#'   \item `time` Timepoints for each individual.
#'   \item `y` Outcome.
#'   \item `class_pred_1` First class predictive covariate (time-invariant).
#'   \item `class_pred_2` Second class predictive covariate (time-invariant).
#'   \item `outcome_pred_1` Outcome predictive covariate (time-varying).
#' }
#'
#' @docType data
#' @keywords datasets
#' @name SimData_PREM
#' @usage data(SimData_PREM)
#' @format A data frame with 540 rows and 6 variables.
NULL

#' Simulated data for a PCREM
#'
#' Simulated data for a piecewise crossed random effects model (PCREM) with 7 timepoints collected on 30 individuals.
#'
#' \itemize{
#'   \item `id` ID for each individual.
#'   \item `teacherid` ID for each teacher.
#'   \item `time` Timepoints for each individual.
#'   \item `y` Outcome.
#' }
#'
#' @docType data
#' @keywords datasets
#' @name SimData_PCREM
#' @usage data(SimData_PCREM)
#' @format A data frame with 210 rows and 4 variables.
NULL

#' Simulated data for a BPREM
#'
#' Simulated data for a bivariate piecewise random effects model (BPREM) with 7 timepoints collected on 30 individuals.
#'
#' \itemize{
#'   \item `id` ID for each individual.
#'   \item `time` Timepoints for each individual.
#'   \item `y1` Outcome 1.
#'   \item `y2` Outcome 2.
#' }
#'
#' @docType data
#' @keywords datasets
#' @name SimData_BPREM
#' @usage data(SimData_BPREM)
#' @format A data frame with 210 rows and 4 variables.
NULL

#' Fitted results for a PREM
#'
#' Fitted results for a piecewise random effects model (PREM) using `SimData_PREM`. Included to demonstrate the use of `plot_BEND()` and `summary.PREM()`.
#'
#' @docType data
#' @keywords datasets
#' @name results_prem
#' @usage data(results_prem)
#' @format A list (an object of class `PREM`) with fitted model results.
NULL

#' Fitted results for a PCREM
#'
#' Simulated data for a piecewise crossed random effects model (PCREM) using `SimData_CREM`. Included to demonstrate the use of `summary.CREM()`.
#'
#' @docType data
#' @keywords datasets
#' @name results_pcrem
#' @usage data(results_pcrem)
#' @format A list (an object of class `CREM`) with fitted model results.
NULL

#' Fitted results for a BPREM
#'
#' Simulated data for a bivariate piecewise random effects model (BPREM) using `SimData_BPREM`. Included to demonstrate the use of `summary.BPREM()`.
#'
#' @docType data
#' @keywords datasets
#' @name results_bprem
#' @usage data(results_bprem)
#' @format A list (an object of class `BPREM`) with fitted model results.
NULL

#' Extract random effects
#'
#' @description
#' Extracts the random effects from a fitted model of class "BPREM" or "CREM".
#'
#' @param x An object of class "BPREM" or "CREM".
#' @param ... Additional arguments.
#'
#' @returns Returns a list of the random effects for each individual/group.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_bprem)
#' # get random effects
#' getRanEf(results_bprem)
#'
#' @export
getRanEf <- function(x, ...) UseMethod(("getRanEf"))

#' @rdname getRanEf
#' @export
getRanEf.BPREM <- function(x, ...){

  out <- list(ranef = x$Random_Coefficients$ranef_b)
  class(out) <- c("getRanEf.BPREM", class(out))
  return(out)

}

#' @rdname getRanEf
#' @export
getRanEf.CREM <- function(x, ...){

  out <- list(ranef_b = x$Random_Coefficients$ranef_b,
              ranef_g = x$Random_Coefficients$ranef_g)
  class(out) <- c("getRanEf.CREM", class(out))
  return(out)

}

#' @rdname getRanEf
#' @export
print.getRanEf.BPREM <- function(x, ...){

  cat("Random Effects\n")
  print(x$ranef, na.print="")
  cat("\n")

  invisible(x)
}

#' @rdname getRanEf
#' @export
print.getRanEf.CREM <- function(x, ...){

  cat("Individual Random Effects\n")
  print(x$ranef_b, na.print="")
  cat("\n")

  cat("Group Random Effects\n")
  print(x$ranef_g, na.print="")
  cat("\n")

  invisible(x)
}

#' Extract model fit
#'
#' @description
#' Extracts the model fit information from a fitted model of class "BPREM", "CREM", or "PREM".
#'
#' @param x An object of class "BPREM", "CREM", or "PREM".
#' @param ... Additional arguments.
#'
#' @returns Returns a vector of the model fit information (deviance, pD, DIC).
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_prem)
#' # get model fit
#' getModelFit(results_prem)
#'
#' @export
getModelFit <- function(x, ...) UseMethod(("getModelFit"))

#' @rdname getModelFit
#' @export
getModelFit.BEND <- function(x, ...){

  out <- list(mod_fit = as.data.frame(x$Model_Fit))
  class(out) <- c("getModelFit.BEND", class(out))
  return(out)

}

#' @rdname getModelFit
#' @export
print.getModelFit.BEND <- function(x, ...){

  print(round(x$mod_fit,2), na.print="", row.names = FALSE)
  cat("\n")

  invisible(x)
}

#' Plot the results of a bivariate piecewise random effects model (BPREM)
#'
#' @description
#' Provides a fitted plot of a BPREM model, as returned by `Bayes_BPREM()`.
#'
#' @param x An object of class "BPREM" (returned by `Bayes_BPREM(...)`).
#' @param xlab X-axis label for the generated plot.
#' @param ylab Y-axis label for the generated plot.
#' @param colors Colors for each class outcome. By default, up to 2 colors are provided in the following order: "blue" (outcome 1), "red" (outcome 2).
#' @param mean_colors Colors for the trajectory defined by the mean parameters for each outcome. By default, up to 2 colors are provided in the following order: "darkblue" (outcome 1), "darkred" (outcome 2).
#' @param legend_pos (optional) Option to change legend position (default = "topright").
#' @param ... (optional) Other parameters to pass to the `plot()` function.
#'
#' @returns No return value.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_bprem)
#' # plot fitted results
#' plot(results_bprem)
#'
#' @import graphics
#'
#' @export
plot.BPREM <- function(x,
                       xlab='X', ylab='Y',
                       colors=NULL, mean_colors=NULL,
                       legend_pos="topright", ...){
  # Setup ----

  object <- x

  data <- object$Data
  id_var <- object$Call$id_var
  time_var <- object$Call$time_var
  y_var <- object$Call$y1_var
  y2_var <- object$Call$y2_var

  if(is.null(colors)) colors <- c('blue','red')
  if(is.null(mean_colors)) mean_colors = c('darkblue','darkred')

  ## outcome data - matrix form
  y <- reshape(data[,c(id_var, time_var, y_var)],
               idvar=id_var,
               timevar=time_var,
               direction='wide')
  y <- unname(as.matrix(y[,names(y)!=id_var]))

  y2 <- reshape(data[,c(id_var, time_var, y2_var)],
                idvar=id_var,
                timevar=time_var,
                direction='wide')
  y2 <- unname(as.matrix(y2[,names(y2)!=id_var]))

  ## time data - matrix form
  ## should be the same dimensions as y
  x <- matrix(data[,c(time_var)],
              byrow=TRUE,
              nrow=dim(y)[1],
              ncol=dim(y)[2])

  ## Define relevant variables
  n_subj <- nrow(y)
  n_time <- ncol(y)
  xvec <- seq(min(x), max(x), length.out=100)

  # Fitted Plot -----

  # pull fixed effect estimates
  mean_est1 <- object$Parameter_Estimates$Mean[1:4]
  mean_est2 <- object$Parameter_Estimates$Mean[5:8]

  # define piecewise function
  pw_eq <- function(x,est){
    return(est[1] + est[2]*x + est[3]*(max(0, x-est[4])))
  }

  mean_traj1 <- rep(0,100)
  mean_traj2 <- rep(0,100)
  for(i in 1:100){
    mean_traj1[i] <- pw_eq(xvec[i], mean_est1)
    mean_traj2[i] <- pw_eq(xvec[i], mean_est2)
  }

  plot(x[1, !is.na(y[1,])],
       y[1, !is.na(y[1,])],
       type = "l",
       col = "grey",
       ylim = c(min(y,na.rm=TRUE), max(y,na.rm=TRUE)),
       xlim= c(min(x,na.rm=TRUE), max(x,na.rm=TRUE)),
       xlab = xlab,
       ylab = ylab, ...)
  for(i in 2:n_subj){
    lines(x[i, !is.na(y[i,])],
          y[i, !is.na(y[i,])],
          type = "l",
          col = colors[1])
  }
  for(i in 1:n_subj){
    lines(x[i, !is.na(y2[i,])],
          y2[i, !is.na(y2[i,])],
          type = "l",
          col = colors[2])
  }
  points(xvec,
         mean_traj1,
         type='l',
         col=mean_colors[1],
         lwd=4)
  points(xvec,
         mean_traj2,
         type='l',
         col=mean_colors[2],
         lwd=4)
  legend(legend_pos, lty=1, col=colors[1:2], legend=c(y_var, y2_var))

}

#' Extract random effects variance-covariance matrix
#'
#' @description
#' Extracts the random effects variance-covariance matrix from a fitted model of class "BPREM", "CREM", or "PREM".
#'
#' @param x An object of class "BPREM", "CREM", or "PREM".
#' @param ... Additional arguments.
#'
#' @returns Returns a list of the random effects variance-covariance matrices.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_prem)
#' # get random effects variance-covariance matrices
#' getVarCov(results_prem)
#'
#' @export
getVarCov <- function(x, ...) UseMethod(("getVarCov"))

#' @rdname getVarCov
#' @export
getVarCov.BPREM <- function(x, ...){

  # determine number of parameters
  n_param <- 4
  # determine number of covariances
  n_covar <- ((n_param*2)*((n_param*2)-1))/2
  # define parameters for labeling
  param_names <- c("Intercept", "Slope", "Change in Slope", "Changepoint")

  ## COVARIANCE MATRIX
  ran_eff_cov_mat <- diag(n_param*2)
  diag(ran_eff_cov_mat) <- x$Parameter_Estimates[((n_param*2)+(1:(n_param*2))),"Mean"]

  ran_eff_cov_mat[upper.tri(ran_eff_cov_mat)] <- x$Parameter_Estimates[(2*(n_param*2)+(1:n_covar)),"Mean"] # covariances on upper triangle
  ran_eff_cov_mat[lower.tri(ran_eff_cov_mat)] <- t(ran_eff_cov_mat)[lower.tri(ran_eff_cov_mat)] # copy to lower triangle
  ran_eff_cov_mat[upper.tri(ran_eff_cov_mat)] <- NA

  colnames(ran_eff_cov_mat) <- rownames(ran_eff_cov_mat) <- paste0(rep(c("Outcome 1: ", "Outcome 2: "), e=4), rep(param_names,2))

  ## CORRELATION MATRIX
  ran_eff_corr_mat <- diag(n_param*2)

  ran_eff_corr_mat[upper.tri(ran_eff_corr_mat)] <- x$Parameter_Estimates[(2*(n_param*2)+n_covar+3+(1:n_covar)),"Mean"] # covariances on upper triangle
  ran_eff_corr_mat[lower.tri(ran_eff_corr_mat)] <- t(ran_eff_corr_mat)[lower.tri(ran_eff_corr_mat)] # copy to lower triangle
  ran_eff_corr_mat[upper.tri(ran_eff_corr_mat)] <- NA

  colnames(ran_eff_corr_mat) <- rownames(ran_eff_corr_mat) <- paste0(rep(c("Outcome 1: ", "Outcome 2: "), e=4), rep(param_names,2))

  out <- list(cov_mat = ran_eff_cov_mat,
              cor_mat = ran_eff_corr_mat)
  class(out) <- c("getVarCov.BPREM", "getVarCov.BEND", class(out))
  return(out)
}

#' @rdname getVarCov
#' @export
getVarCov.CREM <- function(x, ...){

  # determine form
  form <- x$Functional_Form
  # determine number of parameters
  if(form=="linear")                           n_param <- 2
  if(form=="quadratic" | form=="exponential")  n_param <- 3
  if(form=="piecewise")                        n_param <- 4
  # determine number of covariances
  n_covar <- (n_param*(n_param-1))/2
  # define parameters for labeling
  if(form=="linear")       param_names <- c("Intercept", "Slope")
  if(form=="quadratic")    param_names <- c("Intercept", "Linear Slope", "Quadratic Slope")
  if(form=="exponential")  param_names <- c("Intercept", "Total Change", "Growth Rate")
  if(form=="piecewise")    param_names <- c("Intercept", "Slope", "Change in Slope", "Changepoint")

  ## INDIVIDUALS
  ran_eff_b_mat <- diag(n_param)
  diag(ran_eff_b_mat) <- x$Parameter_Estimates[(n_param+(1:n_param)),"Mean"]

  ran_eff_b_mat[upper.tri(ran_eff_b_mat)] <- x$Parameter_Estimates[(2*n_param+(1:n_covar)),"Mean"] # covariances on upper triangle
  ran_eff_b_mat[lower.tri(ran_eff_b_mat)] <- t(ran_eff_b_mat)[lower.tri(ran_eff_b_mat)] # copy to lower triangle
  ran_eff_b_mat[upper.tri(ran_eff_b_mat)] <- NA

  colnames(ran_eff_b_mat) <- rownames(ran_eff_b_mat) <- param_names

  ## GROUPS
  ran_eff_g_mat <- diag(n_param)
  diag(ran_eff_g_mat) <- x$Parameter_Estimates[(2*n_param+n_covar+(1:n_param)),"Mean"]

  ran_eff_g_mat[upper.tri(ran_eff_g_mat)] <- x$Parameter_Estimates[(3*n_param+n_covar+(1:n_covar)),"Mean"] # covariances on upper triangle
  ran_eff_g_mat[lower.tri(ran_eff_g_mat)] <- t(ran_eff_g_mat)[lower.tri(ran_eff_g_mat)] # copy to lower triangle
  ran_eff_g_mat[upper.tri(ran_eff_g_mat)] <- NA

  colnames(ran_eff_g_mat) <- rownames(ran_eff_g_mat) <- param_names

  out <- list(ind_mat = ran_eff_b_mat,
              grp_mat = ran_eff_g_mat)
  class(out) <- c("getVarCov.CREM", "getVarCov.BEND", class(out))
  return(out)
}

#' @rdname getVarCov
#' @export
getVarCov.PREM <- function(x, ...){

  # determine number of classes
  n_class <- length(unique(x$Class_Information$class_membership))

  # determine number of changepoints in each class (based on final model results)
  changepoints <- c()
  for(i in 1:n_class){
    class_num <- paste0("Class_", i)
    changepoints[i] <- which.max(x$Parameter_Estimates[[class_num]]$K_prob)-1
  }

  # pull variances for each class
  varcor_list <- vector("list", n_class)
  for(i in 1:n_class){
    class_num <- paste0("Class_", i)
    cp_num <- paste0("K_", changepoints)[i]
    varcor_vec <- c()
    # random effects
    varcor_vec[1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var[1]
    varcor_vec[2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var[2]
    # if there are changepoints (beyond linear)
    if(changepoints[i]>0){
      for(k in 1:changepoints[i]){
        # random effects
        varcor_vec[2*k+1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_var[k]
        varcor_vec[2*k+2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var[k+2]
      }
    }
    n_var <- length(varcor_vec)

    # param_names
    param_names <- rep(NA, n_var)
    param_names[1] = "Intercept"
    param_names[2] = "Slope"
    if(changepoints[i]>0){
      for(k in 1:changepoints[i]){
        param_names[2*k+1] = paste0("Changepoint ", k)
        param_names[2*k+2] = paste0("Change in Slope ", k)
      }
    }
    # format matrix
    varcor_mat <- diag(n_var)
    diag(varcor_mat) <- varcor_vec
    varcor_mat[row(varcor_mat) != col(varcor_mat)] <- NA
    colnames(varcor_mat) <- rownames(varcor_mat) <- param_names

    # save to list
    varcor_list[[i]] <- varcor_mat
  }
  names(varcor_list) <- paste0("class_", 1:n_class)

  out <- varcor_list
  class(out) <- c("getVarCov.PREM", "getVarCov.BEND", class(out))
  return(out)
}

#' @rdname getVarCov
#' @export
print.getVarCov.BPREM <- function(x, ...){

  cat("Random Effects Variance-Covariance Matrix\n")
  print(round(x$cov_mat,3), na.print="")
  cat("\n")
  cat("Random Effects Correlation Matrix\n")
  print(round(x$cor_mat,3), na.print="")
  cat("\n")

  invisible(x)
}

#' @rdname getVarCov
#' @export
print.getVarCov.CREM <- function(x, ...){

  cat("Random Effects Variance-Covariance Matrix for Individuals\n")
  print(round(x$ind_mat,3), na.print="")
  cat("\n")
  cat("Random Effects Variance-Covariance Matrix for Groups\n")
  print(round(x$grp_mat,3), na.print="")
  cat("\n")

  invisible(x)
}

#' @rdname getVarCov
#' @export
print.getVarCov.PREM <- function(x, ...){

  for(i in 1:length(x)){
    cat(paste0("Random Effects Variance-Covariance Matrix for Class ",i,"\n"))
    print(round(x[[i]],3), na.print="")
    cat("\n")
  }

  invisible(x)
}

#' Print the results of a piecewise random effects model (PREM)
#'
#' @description
#' Provides a summary of a PREM model, as returned by `Bayes_PREM()`.
#'
#' @param x An object of class "PREM" (returned by `Bayes_PREM(...)`).
#' @param ... Additional arguments.
#'
#' @returns Returns a list of key parameter estimates.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_prem)
#' # print results
#' print(results_prem)
#'
#' @export
print.PREM <- function(x, ...){

  # determine number of classes
  n_class <- length(unique(x$Class_Information$class_membership))

  # determine number of changepoints in each class (based on final model results)
  changepoints <- c()
  class_data <- data.frame()
  for(i in 1:n_class){
    class_num <- paste0("Class_", i)
    changepoints[i] <- which.max(x$Parameter_Estimates[[class_num]]$K_prob)-1
  }
  max_cp <- max(changepoints)

  # detect covariates
  n_cov_op <- length(x$Parameter_Estimates$outcome_predictive_covariates)
  n_cov_cp <- length(x$Parameter_Estimates$class_predictive_covariates)

  # CLASS DEPENDENT PARAMETERS -----
  # pull fixed effects for each class
  coeff_mat <- matrix(NA, nrow=2+2*max_cp, ncol=n_class)
  for(i in 1:n_class){
    class_num <- paste0("Class_", i)
    cp_num <- paste0("K_", changepoints)[i]
    # fixed effects
    coeff_mat[1,i] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[1]
    coeff_mat[2,i] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[2]
    # if there are changepoints (beyond linear)
    if(changepoints[i]>0){
      for(k in 1:changepoints[i]){
        # fixed effects
        coeff_mat[2*k+1,i] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_mean[k]
        coeff_mat[2*k+2,i] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[k+2]
      }
    }
  }
  colnames(coeff_mat) <- paste0("Class ", 1:n_class)

  my_rownames <- rep(NA, 2+2*max_cp)
  my_rownames[1] = "Intercept"
  my_rownames[2] = "Slope"
  for(k in 1:max_cp){
    my_rownames[2*k+1] = paste0("Changepoint ", k)
    my_rownames[2*k+2] = paste0("Change in Slope ", k)
  }
  rownames(coeff_mat) <- paste0(my_rownames, " Mean")

  # pull variances for each class
  varcor_list <- vector("list", n_class)
  for(i in 1:n_class){
    class_num <- paste0("Class_", i)
    cp_num <- paste0("K_", changepoints)[i]
    varcor_vec <- c()
    # random effects
    varcor_vec[1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var[1]
    varcor_vec[2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var[2]
    # if there are changepoints (beyond linear)
    if(changepoints[i]>0){
      for(k in 1:changepoints[i]){
        # random effects
        varcor_vec[2*k+1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_var[k]
        varcor_vec[2*k+2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var[k+2]
      }
    }
    n_var <- length(varcor_vec)

    # param_names
    param_names <- rep(NA, n_var)
    param_names[1] = "Intercept"
    param_names[2] = "Slope"
    if(changepoints[i]>0){
      for(k in 1:changepoints[i]){
        param_names[2*k+1] = paste0("Changepoint ", k)
        param_names[2*k+2] = paste0("Change in Slope ", k)
      }
    }
    # format matrix
    varcor_mat <- diag(n_var)
    diag(varcor_mat) <- varcor_vec
    varcor_mat[row(varcor_mat) != col(varcor_mat)] <- NA
    colnames(varcor_mat) <- rownames(varcor_mat) <- param_names

    # save to list
    varcor_list[[i]] <- varcor_mat
  }
  names(varcor_list) <- paste0("class_", 1:n_class)

  # number of changepoints info
  coeff_n_cp <- matrix(changepoints, nrow = 1)
  rownames(coeff_n_cp) <- "Number of Changepoints"
  coeff_mat <- rbind(coeff_n_cp, coeff_mat)

  # class probability info
  class_probs <- matrix(prop.table(table(x$Class_Information$class_membership)), nrow=1)
  rownames(class_probs) <- "Empirical Class Probabilities"
  coeff_mat <- rbind(class_probs, coeff_mat)

  # CLASS INDEPENDENT PARAMETERS -----
  coeff_addit <- data.frame("Estimate" = x$Parameter_Estimates$error_var)
  rownames(coeff_addit) <- "Error Var"

  # Covariates
  if(n_cov_op > 0){
    coeff_cov_op <- data.frame("Estimate" = x$Parameter_Estimates$outcome_predictive_covariates)
    coeff_addit <- rbind(coeff_addit, coeff_cov_op)
  }
  if(n_cov_cp > 0){
    coeff_cov_cp <- data.frame("Estimate" = x$Parameter_Estimates$class_predictive_covariates)
    rownames(coeff_cov_cp) <- paste0(rownames(coeff_cov_cp), " (in log-odds units)")
    coeff_log_int <- data.frame("Estimate" = x$Parameter_Estimates$logistic_intercept)
    rownames(coeff_log_int) <- "Logistic Intercept"
    coeff_addit <- rbind(coeff_addit, coeff_cov_cp, coeff_log_int)
  }

  ## return value
  out <- list("class_dep_params" = coeff_mat,
              "class_dep_varcov" = varcor_list,
              "class_ind_params" = coeff_addit,
              "msrf" = x$Convergence$multivariate_psrf,
              "mean_psrf" = x$Convergence$mean_psrf,
              "DIC" = x$Model_Fit$dic)

  cat("Bayesian piecewise random effects model\n")
  cat("Data:", x$Call$data, "\n")
  cat("Outcome:", x$Call$y_var, "\n")
  cat("Number of Classes:", x$Call$max_cp, "\n")
  cat("Maximum Number of Changepoints:", max_cp, "\n")
  if(n_cov_cp==1) cat("Class-Predictive Covariates:", x$Call$class_predictive_vars,"\n")
  if(n_cov_cp>1) cat("Class-Predictive Covariates:", as.character(x$Call$class_predictive_vars)[-1],"\n")
  if(n_cov_op==1) cat("Outcome-Predictive Covariates:", x$Call$outcome_predictive_vars,"\n")
  if(n_cov_op>1) cat("Outcome-Predictive Covariates:", as.character(x$Call$outcome_predictive_vars)[-1],"\n")
  cat("\n")
  cat("Class Dependent Parameters:\n")
  cat("Fixed Effects Parameters:\n")
  print(round(out$class_dep_params,3), na.print="")
  cat("\n")
  cat("Random Effects Parameters:\n")
  for(i in 1:length(out$class_dep_varcov)){
    cat(paste0("Class ",i,":\n"))
    print(round(out$class_dep_varcov[[i]],3), na.print="")
    cat("\n")
  }
  cat("Class Independent Parameters:\n")
  print(round(out$class_ind_params,3))
  cat("\n")
  cat("Gelman's msrf:", round(out$msrf, 3), "\n")
  cat("Mean psrf:", round(out$mean_psrf, 3), "\n")
  cat("DIC:", round(out$DIC,3))
  cat("\n")

  invisible(x)
}

#' Print the results of a bivariate piecewise random effects model (BPREM)
#'
#' @description
#' Provides a summary of a BPREM model, as returned by `Bayes_BPREM()`.
#'
#' @param x An object of class "BPREM" (returned by `Bayes_BPREM(...)`).
#' @param ... Additional arguments.
#'
#' @returns Returns a list of key parameter estimates.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_bprem)
#' # print results
#' print(results_bprem)
#'
#' @export
print.BPREM <- function(x, ...){

  # determine number of parameters
  n_param <- 4

  # determine number of covariances
  n_covar <- ((n_param*2)*((n_param*2)-1))/2

  # define parameters for labeling
  param_names <- c("Intercept", "Slope", "Change in Slope", "Changepoint")

  # FIXED EFFECTS -----
  fix_eff_est <- matrix(x$Parameter_Estimates[1:(n_param*2),"Mean"], ncol=2)
  colnames(fix_eff_est) <- c("Outcome 1", "Outcome 2")
  rownames(fix_eff_est) <- paste0(param_names, " Mean")

  # RANDOM EFFECTS -----

  ## COVARIANCE MATRIX -----
  ran_eff_cov_mat <- diag(n_param*2)
  diag(ran_eff_cov_mat) <- x$Parameter_Estimates[((n_param*2)+(1:(n_param*2))),"Mean"]

  ran_eff_cov_mat[upper.tri(ran_eff_cov_mat)] <- x$Parameter_Estimates[(2*(n_param*2)+(1:n_covar)),"Mean"] # covariances on upper triangle
  ran_eff_cov_mat[lower.tri(ran_eff_cov_mat)] <- t(ran_eff_cov_mat)[lower.tri(ran_eff_cov_mat)] # copy to lower triangle
  ran_eff_cov_mat[upper.tri(ran_eff_cov_mat)] <- NA

  colnames(ran_eff_cov_mat) <- rownames(ran_eff_cov_mat) <- paste0(rep(c("Outcome 1: ", "Outcome 2: "), e=4), rep(param_names,2))

  ## CORRELATION MATRIX -----
  ran_eff_corr_mat <- diag(n_param*2)

  ran_eff_corr_mat[upper.tri(ran_eff_corr_mat)] <- x$Parameter_Estimates[(2*(n_param*2)+n_covar+3+(1:n_covar)),"Mean"] # covariances on upper triangle
  ran_eff_corr_mat[lower.tri(ran_eff_corr_mat)] <- t(ran_eff_corr_mat)[lower.tri(ran_eff_corr_mat)] # copy to lower triangle
  ran_eff_corr_mat[upper.tri(ran_eff_corr_mat)] <- NA

  colnames(ran_eff_corr_mat) <- rownames(ran_eff_corr_mat) <- paste0(rep(c("Outcome 1: ", "Outcome 2: "), e=4), rep(param_names,2))

  # ERROR -----
  error_cov_mat <- diag(2)
  diag(error_cov_mat) <- x$Parameter_Estimates[(2*(n_param*2)+n_covar+(1:2)),"Mean"]

  error_cov_mat[upper.tri(error_cov_mat)] <- error_cov_mat[lower.tri(error_cov_mat)] <- x$Parameter_Estimates[(2*(n_param*2)+n_covar+(3)),"Mean"] # covariances on upper triangle
  error_cov_mat[upper.tri(error_cov_mat)] <- NA

  colnames(error_cov_mat) <- rownames(error_cov_mat) <- paste0(c("Outcome 1: ", "Outcome 2: "), "Error")

  ## return value
  out <- list("fix_eff_est" = fix_eff_est,
              "ran_eff_cov_mat" = ran_eff_cov_mat,
              "ran_eff_corr_mat" = ran_eff_corr_mat,
              "error_cov_mat" = error_cov_mat,
              "error_corr" = x$Parameter_Estimates[(2*(n_param*2)+2*n_covar+3+1),"Mean"],
              "msrf" = x$Convergence$multivariate_psrf,
              "mean_psrf" = x$Convergence$mean_psrf,
              "DIC" = x$Model_Fit$dic)

  cat("Bayesian bivariate piecewise random effects model\n")
  cat("Data:", x$Call$data, "\n")
  cat("Outcomes:", c(x$Call$y1_var, x$Call$y2_var), "\n")
  cat("\n")
  cat("Fixed Effect Parameters:\n")
  print(round(out$fix_eff_est,3))
  cat("\n")
  cat("Random Effect Parameters:\n")
  cat("Covariance Matrix:\n")
  print(round(out$ran_eff_cov_mat,3), na.print="")
  cat("\n")
  cat("Correlation Matrix:\n")
  print(round(out$ran_eff_corr_mat,3), na.print="")
  cat("\n")
  cat("Error:\n")
  cat("Covariance Matrix:\n")
  print(round(out$error_cov_mat,3), na.print="")
  cat("\n")
  cat("Error Corr:", round(out$error_corr,3), "\n")
  cat("Gelman's msrf:", round(out$msrf, 3), "\n")
  cat("Mean psrf:", round(out$mean_psrf, 3), "\n")
  cat("DIC:", round(out$DIC,3))
  cat("\n")

  invisible(x)
}

#' Print the results of a crossed random effects model (CREM)
#'
#' @description
#' Provides a summary of a CREM model, as returned by `Bayes_CREM()`.
#'
#' @param x An object of class "CREM" (returned by `Bayes_CREM(...)`).
#' @param ... Additional arguments.
#'
#' @returns Returns a list of key parameter estimates.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_pcrem)
#' # print results
#' print(results_pcrem)
#'
#' @export
print.CREM <- function(x, ...){

  # determine form
  form <- x$Functional_Form

  # determine number of parameters
  if(form=="linear")                           n_param <- 2
  if(form=="quadratic" | form=="exponential")  n_param <- 3
  if(form=="piecewise")                        n_param <- 4

  # determine number of covariances
  n_covar <- (n_param*(n_param-1))/2

  # define parameters for labeling
  if(form=="linear")       param_names <- c("Intercept", "Slope")
  if(form=="quadratic")    param_names <- c("Intercept", "Linear Slope", "Quadratic Slope")
  if(form=="exponential")  param_names <- c("Intercept", "Total Change", "Growth Rate")
  if(form=="piecewise")    param_names <- c("Intercept", "Slope", "Change in Slope", "Changepoint")

  # FIXED EFFECTS -----
  fix_eff_est <- matrix(x$Parameter_Estimates[1:n_param,"Mean"], ncol=1)
  colnames(fix_eff_est) <- "Estimate"
  rownames(fix_eff_est) <- paste0(param_names, " Mean")

  # RANDOM EFFECTS -----

  ## INDIVIDUALS -----
  ran_eff_b_mat <- diag(n_param)
  diag(ran_eff_b_mat) <- x$Parameter_Estimates[(n_param+(1:n_param)),"Mean"]

  ran_eff_b_mat[upper.tri(ran_eff_b_mat)] <- x$Parameter_Estimates[(2*n_param+(1:n_covar)),"Mean"] # covariances on upper triangle
  ran_eff_b_mat[lower.tri(ran_eff_b_mat)] <- t(ran_eff_b_mat)[lower.tri(ran_eff_b_mat)] # copy to lower triangle
  ran_eff_b_mat[upper.tri(ran_eff_b_mat)] <- NA

  colnames(ran_eff_b_mat) <- rownames(ran_eff_b_mat) <- param_names

  ## GROUPS -----
  ran_eff_g_mat <- diag(n_param)
  diag(ran_eff_g_mat) <- x$Parameter_Estimates[(2*n_param+n_covar+(1:n_param)),"Mean"]

  ran_eff_g_mat[upper.tri(ran_eff_g_mat)] <- x$Parameter_Estimates[(3*n_param+n_covar+(1:n_covar)),"Mean"] # covariances on upper triangle
  ran_eff_g_mat[lower.tri(ran_eff_g_mat)] <- t(ran_eff_g_mat)[lower.tri(ran_eff_g_mat)] # copy to lower triangle
  ran_eff_g_mat[upper.tri(ran_eff_g_mat)] <- NA

  colnames(ran_eff_g_mat) <- rownames(ran_eff_g_mat) <- param_names

  ## return value
  out <- list("fix_eff_est" = fix_eff_est,
              "ran_eff_b_mat" = ran_eff_b_mat,
              "ran_eff_g_mat" = ran_eff_g_mat,
              "error_var" = x$Parameter_Estimates[(3*n_param+2*n_covar+1),"Mean"],
              "msrf" = x$Convergence$multivariate_psrf,
              "mean_psrf" = x$Convergence$mean_psrf,
              "DIC" = x$Model_Fit$dic)

  cat("Bayesian crossed random effects model\n")
  cat("Data:", x$Call$data, "\n")
  cat("Outcome:", x$Call$y_var, "\n")
  cat("Individuals:", x$Call$ind_id_var, "\n")
  cat("Group:", x$Call$cross_id_var, "\n")
  cat("Functional Form:", x$Call$form, "\n")
  cat("\n")
  cat("Fixed Effect Parameters:\n")
  print(round(out$fix_eff_est,3))
  cat("\n")
  cat("Random Effect Parameters:\n")
  cat("Individual Random Effects Covariance Matrix:\n")
  print(round(out$ran_eff_b_mat,3), na.print="")
  cat("\n")
  cat("Group Random Effects Covariance Matrix:\n")
  print(round(out$ran_eff_g_mat,3), na.print="")
  cat("\n")
  cat("Error Var:", round(out$error_var,3), "\n")
  cat("Gelman's msrf:", round(out$msrf, 3), "\n")
  cat("Mean psrf:", round(out$mean_psrf, 3), "\n")
  cat("DIC:", round(out$DIC,3))
  cat("\n")

  invisible(x)
}

#' Extract class probabilities
#'
#' @description
#' Extracts the posterior class probabilities for each individual from a fitted model of class "PREM".
#'
#' @param x An object of class "PREM".
#' @param ... Additional arguments.
#'
#' @returns Returns a list of the posterior class probabilities for each individual.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_prem)
#' # get class probabilities
#' getClassProb(results_prem)
#'
#' @export
getClassProb <- function(x, ...) UseMethod(("getClassProb"))

#' @rdname getClassProb
#' @export
getClassProb.PREM <- function(x, ...){

  # determine number of classes
  n_class <- length(unique(x$Class_Information$class_membership))

  # detect covariates
  n_cov_op <- length(x$Parameter_Estimates$outcome_predictive_covariates)
  n_cov_cp <- length(x$Parameter_Estimates$class_predictive_covariates)

  if(n_class==1){
    individ_class_info <- cbind(names(x$Class_Information$class_membership),
                                x$Class_Information$class_membership,
                                x$Class_Information$individ_class_probability)
    individ_class_info <- as.data.frame(individ_class_info)
    names(individ_class_info) <- c(x$Call$id_var, "class_membership", "probability_class_1")
    class_prob <- x$Class_Information$unconditional_class_probability
  }
  if(n_class>1){
    if(n_cov_cp==0){
      individ_class_info <- cbind(names(x$Class_Information$class_membership),
                                  x$Class_Information$class_membership,
                                  x$Class_Information$individ_class_probability)
      individ_class_info <- as.data.frame(individ_class_info)
      names(individ_class_info) <- c(x$Call$id_var, "class_membership", paste0("probability_class_",1:n_class))
      class_prob <- x$Class_Information$unconditional_class_probability
    }
    if(n_cov_cp>0){
      individ_class_info <- cbind(names(x$Class_Information$class_membership),
                                  x$Class_Information$class_membership,
                                  x$Class_Information$individ_class_probability,
                                  x$Class_Information$unconditional_class_probability)
      individ_class_info <- as.data.frame(individ_class_info)
      names(individ_class_info) <- c(x$Call$id_var, "class_membership", paste0("probability_class_",1:n_class), "probability_class_ref")
      class_prob <- NULL
    }
  }

  out <- list(class_prob = class_prob,
              individ_class_info = individ_class_info)
  class(out) <- c("getClassProb.PREM", class(out))
  return(out)

}

#' @rdname getClassProb
#' @export
print.getClassProb.PREM <- function(x, ...){

  if(!is.null(x$class_prob)){
    cat("Population Class Probabilities\n")
    print(round(x$class_prob,2), na.print="")
    cat("\n")
  }

  cat("Individual Class Information\n")
  print(x$individ_class_info, na.print="")
  cat("\n")

  invisible(x)
}

#' Bayesian Bivariate Piecewise Random Effects Model (BPREM)
#'
#' @description Estimates a Bayesian bivariate piecewise random effects models (BPREM) for longitudinal data with two interrelated outcomes. See Peralta et al. (2022) for more details.
#'
#' @param data Data frame in long format, where each row describes a measurement occasion for a given individual. It is assumed that each individual has the same number of assigned timepoints (a.k.a., rows). There can be missingness in the outcomes (`y1_var` and ` y2_var`), but there cannot be missingness in time (`time_var`).
#' @param id_var Name of column that contains ids for individuals with repeated measures in a longitudinal dataset.
#' @param time_var Name of column that contains the time variable. This column cannot contain any missing values.
#' @param y1_var Name of column that contains the first outcome variable. Missing values should be denoted by NA.
#' @param y2_var Name of column that contains the second outcome variable. Missing values should be denoted by NA.
#' @param iters_adapt (optional) Number of iterations for adaptation of jags model (default = 5000).
#' @param iters_burn_in (optional) Number of iterations for burn-in (default = 100000).
#' @param iters_sampling (optional) Number of iterations for posterior sampling (default = 50000).
#' @param thin (optional) Thinning interval for posterior sampling (default = 15).
#' @param save_full_chains Logical indicating whether the MCMC chains from rjags should be saved (default = FALSE). Note, this should not be used regularly as it will result in an object with a large file size.
#' @param save_conv_chains Logical indicating whether the MCMC chains from rjags should be saved but only for the parameters monitored for convergence (default = FALSE). This would be useful for plotting traceplots for relevant model parameters to evaluate convergence behavior. Note, this should not be used regularly as it will result in an object with a large file size.
#' @param verbose Logical controlling whether progress messages/bars are generated (default = TRUE).
#'
#' @returns A list (an object of class `BPREM`) with elements:
#' \item{Call}{A list of all specified function arguments.}
#' \item{Sample Size}{Number of subjects (`n_subj`) and number of timepoints (`n_time`) in the dataset.}
#' \item{Data}{Dataset used for estimation.}
#' \item{Convergence}{Potential scale reduction factor (PSRF) for each parameter (`parameter_psrf`), Gelman multivariate scale reduction factor (`multivariate_psrf`), and mean PSRF (`mean_psrf`) to assess model convergence.}
#' \item{Model_Fit}{Deviance (`deviance`), effective number of parameters (`pD`), and Deviance information criterion (`dic`) to assess model fit.}
#' \item{Fitted_Values}{Data frame with the fitted values for y1 and y2 at each timepoint for each individual.}
#' \item{Parameter_Estimates}{Data frame with posterior mean and 95% credible intervals for each model parameter.}
#' \item{Random_Coefficients}{List object with data frames providing random effects and random coefficients for each individual.}
#' \item{Run_Time}{Total run time for model fitting.}
#' \item{Full_MCMC_Chains}{If save_full_chains=TRUE, raw MCMC chains from rjags.}
#' \item{Convergence_MCMC_Chains}{If save_conv_chains=TRUE, raw MCMC chains from rjags but only for the parameters monitored for convergence.}
#'
#' @details
#' For more information on the model equation and priors implemented in this function, see Peralta et al. (2022).
#'
#' @author Corissa T. Rohloff, Yadira Peralta
#'
#' @references Peralta, Y., Kohli, N., Lock, E. F., & Davison, M. L. (2022). Bayesian modeling of associations in bivariate piecewise linear mixed-effects models. Psychological Methods, 27(1), 44–64. https://doi.org/10.1037/met0000358
#'
#' @examples
#' \donttest{
#' # load simulated data
#' data(SimData_BPREM)
#' # fit Bayes_BPREM()
#' results_bprem <- Bayes_BPREM(data = SimData_BPREM,
#'                              id_var = "id",
#'                              time_var = "time",
#'                              y1_var = "y1",
#'                              y2_var = "y2")
#' # result summary
#' summary(results_bprem)
#' # plot fitted results
#' plot(results_bprem)
#' }
#'
#' @import stats
#'
#' @export
Bayes_BPREM <- function(data,
                        id_var, time_var, y1_var, y2_var,
                        iters_adapt=5000, iters_burn_in=100000, iters_sampling=50000, thin=15,
                        save_full_chains=FALSE, save_conv_chains=FALSE,
                        verbose=TRUE){

  # START OF SETUP ----

  ## Initial data check
  data <- as.data.frame(data)

  ## Check that number of rows is the same for each id_var
  rows_per_id <- table(data[,id_var])
  if(length(unique(rows_per_id)) != 1L) stop("The number of rows should be the same for each individual. \nSee ?Bayes_PREM for more information.")

  ## Control progress messages/bars
  if(verbose) progress_bar = "text"
  if(!verbose) progress_bar = "none"

  ## Start tracking run time
  run_time_total_start <- Sys.time()

  ## Load module to compute DIC
  suppressMessages(rjags::load.module('dic'))

  ## Set number of chains - will use 3 chains across all models
  n_chains <- 3

  # Reshape data for modeling ----

  ## outcome data - matrix form
  y1 <- reshape(data[,c(id_var, time_var, y1_var)],
                idvar=id_var,
                timevar=time_var,
                direction='wide')
  y1 <- unname(as.matrix(y1[,names(y1)!=id_var]))

  y2 <- reshape(data[,c(id_var, time_var, y2_var)],
                idvar=id_var,
                timevar=time_var,
                direction='wide')
  y2 <- unname(as.matrix(y2[,names(y2)!=id_var]))

  ## JAGS model assumes the two outcome variables are in an array
  y <- array(c(y1, y2), dim = c(nrow(y1),ncol(y1),2))

  ## time data - matrix form
  ## should be the same dimensions as y1 and y2
  x <- matrix(data[,c(time_var)],
              byrow=TRUE,
              nrow=dim(y1)[1],
              ncol=dim(y1)[2])

  # Define relevant variables ----
  n_subj <- dim(y)[1]
  n_time <- dim(y)[2]

  max_time <- max(x)
  min_time <- min(x)
  min_cp_mean <- unique(sort(x))[2]
  max_cp_mean <- unique(sort(x, decreasing=TRUE))[2]

  mean_cp <- (max_time-min_time)/2          # mean of changepoint
  prec_cp <- 1/((max_time-min_time)/4)^2    # precision of changepoint
  bound_e1 <- min(apply(y[,,1], 2, var, na.rm = TRUE))  # bound of error variance 1
  bound_e2 <- min(apply(y[,,2], 2, var, na.rm = TRUE))  # bound of error variance 2
  beta_mean_zero <- rep(0,8)  # Vector of zeros to center the distribution of raw random-effects
  omega_b <- diag(8)     # Scale matrix for the Wishart distribution

  # Error messages for input ----

  ## data format warnings
  if(sum(is.na(x)>0)) stop('Columns for t_var cannot have NA values (but y_var can)')
  if(min(x)!=0) warning('Prior assumes first time point measured at x=0')

  # FULL MODEL ----

  ## Specify full JAGS model
  full_spec <- textConnection(bivariate_pw)

  ## Variables to extract from full model
  param_recovery_full <- c('beta', 'beta_mean', 'beta_rand',
                           'var_b', 'cov_b',
                           'cor_b', 'rho',
                           'mu_y', 'sigma2_error',
                           'for_conv', 'pD', 'deviance')

  ## Create data list for JAGS model
  data_list <- list('y' = y, 'x' = x,
                    'n_subj' = n_subj, 'n_time' = n_time,
                    'mean_cp' = mean_cp, 'prec_cp' = prec_cp,
                    'min_time' = min_time, 'max_time' = max_time,
                    'bound_e1' = bound_e1, 'bound_e2' = bound_e2,
                    'beta_mean_zero' = beta_mean_zero,
                    'omega_b' = omega_b)

  ## Set Seeds for Reproducible Code
  initial_vals <- vector('list', n_chains)
  seeds <- sample(1:10, n_chains, replace = FALSE) # randomly generate non-repeating integers
  # sensitive to set.seed()
  for(i in 1:n_chains){
    initial_vals[[i]]$.RNG.name <- "base::Wichmann-Hill" # arbitrary
    initial_vals[[i]]$.RNG.seed <- seeds[i]
  }

  ## Create JAGS model
  if(verbose) cat('Calibrating MCMC...\n')
  full_model <- rjags::jags.model(full_spec,
                                  data = data_list,
                                  inits = initial_vals,
                                  n.chains = n_chains,
                                  n.adapt = iters_adapt,
                                  quiet = TRUE)

  # burn-in
  if(verbose) cat('Running burn-in...\n')
  update(full_model, iters_burn_in, progress.bar=progress_bar)

  # sampling
  if(verbose) cat('Collecting samples...\n')
  full_out <- rjags::jags.samples(full_model,
                                  variable.names=param_recovery_full,
                                  n.iter=iters_sampling,
                                  thin=thin,
                                  progress.bar=progress_bar)

  # Compiling Full Results -----

  ## Call Info
  call_info <- as.list(match.call())
  def_args <- formals(sys.function())
  for(arg in names(def_args)){
    if(!(arg %in% names(call_info))){
      call_info[[arg]] <- def_args[[arg]]
    }
  }

  ## Convergence

  # Setup the parameter vector
  beta_mean_param <- paste0('beta_', rep(1:2,e=4), c(0:2, 'cp'), '_mean')
  beta_var_param <- paste0('var_b_', rep(1:2,e=4), c(0:2, 'cp'))
  base_cov <- c('11_10',
                '12_10',  '12_11',
                '1cp_10', '1cp_11', '1cp_12',
                '20_10',  '20_11',  '20_12',  '20_1cp',
                '21_10',  '21_11',  '21_12',  '21_1cp',  '21_20',
                '22_10',  '22_11',  '22_12',  '22_1cp',  '22_20',  '22_21',
                '2cp_10', '2cp_11', '2cp_12', '2cp_1cp', '2cp_20', '2cp_21', '2cp_22')
  beta_cov_param <- paste0("cov_b_", base_cov)
  beta_corr_param <- paste0("corr_b_", base_cov)
  error_param <- c("error_var_11", "error_var_22", "error_cov_12")
  param_names <- c(beta_mean_param, beta_var_param, beta_cov_param, error_param,
                   beta_corr_param, "error_corr_12")

  mcmc_list <- coda::as.mcmc.list(full_out$for_conv)
  for(i in 1:3){colnames(mcmc_list[[i]]) <- param_names}

  gelman_msrf <- coda::gelman.diag(mcmc_list[,1:47]) # individual parameter psrf, and multivariate psrf
  # beta_corr_param and error_corr_12 are redundant parameters, thus they were removed when assessing convergence
  parameter_psrf <- data.frame(point_est = gelman_msrf$psrf[,1],
                               upper_ci = gelman_msrf$psrf[,2])
  multivariate_psrf <- gelman_msrf$mpsrf
  mean_psrf <- mean(parameter_psrf$point_est) # mean psrf across parameters
  convergence <- list(parameter_psrf=parameter_psrf, multivariate_psrf=multivariate_psrf, mean_psrf=mean_psrf)

  ## Model Fit
  deviance <- mean(full_out$deviance)
  pD <- mean(full_out$pD)
  dic <- deviance + pD
  model_fit <- list(deviance=deviance, pD=pD, dic=dic)

  ## Fitted Values
  y_mean <- summary(full_out$mu_y, FUN='mean')[[1]]
  fit_vals <- cbind(data[,id_var], data[,time_var], data[,y1_var], data[,y2_var],
                    as.vector(t(y_mean[,,1])),
                    as.vector(t(y_mean[,,2])))
  fit_vals <- as.data.frame(fit_vals)
  names(fit_vals) <- c(id_var, time_var, y1_var, y2_var, "y1_fitted", "y2_fitted")

  ## Random Effects and Coefficients
  ranef_b <- cbind(unique(data[,id_var]), summary(full_out$beta_rand, FUN='mean')[[1]])
  ranef_b <- as.data.frame(ranef_b)
  names(ranef_b) <- c(id_var, gsub("var_", "", beta_var_param))

  rancoef <- cbind(unique(data[,id_var]), summary(full_out$beta, FUN='mean')[[1]])
  rancoef <- as.data.frame(rancoef)
  names(rancoef) <- c(id_var, gsub("var_", "", beta_var_param))

  ran_ef_coef <- list("ranef_b" = ranef_b,
                      "rancoef" = rancoef)

  ## Parameter Estimates
  sum_mcmc <- summary(mcmc_list) # parameter estimates
  param_est <- data.frame(Mean = sum_mcmc$statistics[,1],
                          CI_Lower = sum_mcmc$quantiles[,1],
                          CI_Upper = sum_mcmc$quantiles[,5])

  ## Stop tracking run time
  run_time_total_end <- Sys.time()
  run_time_total <- run_time_total_end - run_time_total_start

  my_results <- list('Call' = call_info,
                     'Sample_Size' = list(n_subj=n_subj, n_time=n_time),
                     'Data' = data,
                     'Convergence' = convergence,
                     'Model_Fit' = model_fit,
                     'Fitted_Values' = fit_vals,
                     'Parameter_Estimates' = param_est,
                     'Random_Coefficients' = ran_ef_coef,
                     'Run_Time' = format(run_time_total))
  if(save_full_chains==TRUE){my_results$Full_MCMC_Chains=full_out}
  if(save_conv_chains==TRUE){my_results$Convergence_MCMC_Chains=mcmc_list[,1:47]}
  class(my_results) <- c('BPREM', 'BEND')
  return(my_results)
}

# Model ----

## Bivariate Piecewise ----

bivariate_pw <- "model{
  for(i in 1:n_subj){
    for(j in 1:n_time){
      y[i,j,1:2] ~ dmnorm(mu_y[i,j,1:2], tau_error[1:2,1:2])
      mu_y[i,j,1] <- beta[i,1] + beta[i,2]*x[i,j] + beta[i,3]*(max(0, x[i,j]-beta[i,4]))
      mu_y[i,j,2] <- beta[i,5] + beta[i,6]*x[i,j] + beta[i,7]*(max(0, x[i,j]-beta[i,8]))
    }
  }

  ## Level 1 precision and error variance
  tau_error[1:2,1:2] <- inverse(sigma2_e[,])
  sigma2_e[1,1] <- sigma2_e1
  sigma2_e[2,2] <- sigma2_e2
  sigma2_e[2,1] <- rho*sqrt(sigma2_e1)*sqrt(sigma2_e2)
  sigma2_e[1,2] <- sigma2_e[2,1]
  sigma2_e1 ~ dunif(0,bound_e1)
  sigma2_e2 ~ dunif(0,bound_e2)
  rho ~ dunif(-1,1)

  ## Distribution of random-effects and error variance
  for (i in 1:n_subj){
    for(k in 1:8){
      beta[i,k] <- beta_mean[k] + beta_rand[i,k]
      beta_rand[i,k] <- scale_c[k]*beta_raw[i,k]
    }
    beta_raw[i,1:8] ~ dmnorm(beta_mean_zero[1:8], tau_b_raw[1:8,1:8])
  }

  ## Priors for fixed-effects
  beta_mean[1] ~ dnorm(0, 0.0001)
  beta_mean[2] ~ dnorm(0, 0.0001)
  beta_mean[3] ~ dnorm(0, 0.0001)
  beta_mean[4] ~ dnorm(mean_cp, prec_cp)T(min_time,max_time)
  beta_mean[5] ~ dnorm(0, 0.0001)
  beta_mean[6] ~ dnorm(0, 0.0001)
  beta_mean[7] ~ dnorm(0, 0.0001)
  beta_mean[8] ~ dnorm(mean_cp, prec_cp)T(min_time,max_time)

  ## Priors for scaling constants
  for(k in 1:8){
    scale_c[k] ~ dunif(0, 100)
  }

  ## Prior for covariance matrix of random-effects
  # Inverse Wishart prior for the raw covariance matrix
  tau_b_raw[1:8,1:8] ~ dwish(omega_b[1:8,1:8], 9)
  sigma2_b_raw[1:8,1:8] <- inverse(tau_b_raw[,])

  ## Define elements of correlation and covariance matrices to recover
  for(k in 1:8){
    for(k_prime in 1:8){
      rho_b[k,k_prime] <- sigma2_b_raw[k,k_prime]/sqrt(sigma2_b_raw[k,k]*sigma2_b_raw[k_prime,k_prime])
      cov_b_mat[k,k_prime] <- scale_c[k]*scale_c[k_prime]*sigma2_b_raw[k,k_prime]
    }
  }

  # Variances:
  for(k in 1:8){
    var_b[k] <- cov_b_mat[k,k]
  }

  # Covariances
  cov_b[1] <- cov_b_mat[2,1]
  cov_b[2] <- cov_b_mat[3,1]
  cov_b[3] <- cov_b_mat[3,2]
  cov_b[4] <- cov_b_mat[4,1]
  cov_b[5] <- cov_b_mat[4,2]
  cov_b[6] <- cov_b_mat[4,3]
  cov_b[7] <- cov_b_mat[5,1]
  cov_b[8] <- cov_b_mat[5,2]
  cov_b[9] <- cov_b_mat[5,3]
  cov_b[10] <- cov_b_mat[5,4]
  cov_b[11] <- cov_b_mat[6,1]
  cov_b[12] <- cov_b_mat[6,2]
  cov_b[13] <- cov_b_mat[6,3]
  cov_b[14] <- cov_b_mat[6,4]
  cov_b[15] <- cov_b_mat[6,5]
  cov_b[16] <- cov_b_mat[7,1]
  cov_b[17] <- cov_b_mat[7,2]
  cov_b[18] <- cov_b_mat[7,3]
  cov_b[19] <- cov_b_mat[7,4]
  cov_b[20] <- cov_b_mat[7,5]
  cov_b[21] <- cov_b_mat[7,6]
  cov_b[22] <- cov_b_mat[8,1]
  cov_b[23] <- cov_b_mat[8,2]
  cov_b[24] <- cov_b_mat[8,3]
  cov_b[25] <- cov_b_mat[8,4]
  cov_b[26] <- cov_b_mat[8,5]
  cov_b[27] <- cov_b_mat[8,6]
  cov_b[28] <- cov_b_mat[8,7]

  # Correlations
  cor_b[1] <- rho_b[2,1]
  cor_b[2] <- rho_b[3,1]
  cor_b[3] <- rho_b[3,2]
  cor_b[4] <- rho_b[4,1]
  cor_b[5] <- rho_b[4,2]
  cor_b[6] <- rho_b[4,3]
  cor_b[7] <- rho_b[5,1]
  cor_b[8] <- rho_b[5,2]
  cor_b[9] <- rho_b[5,3]
  cor_b[10] <- rho_b[5,4]
  cor_b[11] <- rho_b[6,1]
  cor_b[12] <- rho_b[6,2]
  cor_b[13] <- rho_b[6,3]
  cor_b[14] <- rho_b[6,4]
  cor_b[15] <- rho_b[6,5]
  cor_b[16] <- rho_b[7,1]
  cor_b[17] <- rho_b[7,2]
  cor_b[18] <- rho_b[7,3]
  cor_b[19] <- rho_b[7,4]
  cor_b[20] <- rho_b[7,5]
  cor_b[21] <- rho_b[7,6]
  cor_b[22] <- rho_b[8,1]
  cor_b[23] <- rho_b[8,2]
  cor_b[24] <- rho_b[8,3]
  cor_b[25] <- rho_b[8,4]
  cor_b[26] <- rho_b[8,5]
  cor_b[27] <- rho_b[8,6]
  cor_b[28] <- rho_b[8,7]

  # Define elements of error covariance (level 1) to recover
  sigma2_error[1] <- sigma2_e[1,1]
  sigma2_error[2] <- sigma2_e[2,2]
  sigma2_error[3] <- sigma2_e[2,1]

  # Collect important parameters to assess convergence
  for_conv[1:8] <- beta_mean
  for_conv[9:16] <- var_b
  for_conv[17:44] <- cov_b
  for_conv[45:47] <- sigma2_error
  for_conv[48:75] <- cor_b
  for_conv[76] <- rho
  }"

#' Extract random coefficients
#'
#' @description
#' Extracts the random coefficients from a fitted model of class "BPREM", "CREM", or "PREM".
#'
#' @param x An object of class "BPREM", "CREM", or "PREM".
#' @param ... Additional arguments.
#'
#' @returns Returns a data frame of the random coefficients for each individual/group.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_prem)
#' # get random coefficients
#' getCoef(results_prem)
#'
#' @export
getCoef <- function(x, ...) UseMethod(("getCoef"))

#' @rdname getCoef
#' @export
getCoef.BEND <- function(x, ...){

  out <- x$Random_Coefficients$rancoef
  class(out) <- c("getCoef.BEND", class(out))
  return(out)

}

#' @rdname getCoef
#' @export
print.getCoef <- function(x, ...){

  print(x, na.print="")
  cat("\n")

  invisible(x)
}

#' Print the results of a crossed random effects model (CREM)
#'
#' @description
#' Provides a summary of a CREM model, as returned by `Bayes_CREM()`.
#'
#' @param x An object of class "CREM" (returned by `Bayes_CREM(...)`).
#' @param ... Additional arguments.
#'
#' @returns Returns a list of key parameter estimates.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_pcrem)
#' # print results
#' print(results_pcrem)
#'
#' @export
print.CREM <- function(x, ...){

  # determine form
  form <- x$Functional_Form

  # determine number of parameters
  if(form=="linear")                           n_param <- 2
  if(form=="quadratic" | form=="exponential")  n_param <- 3
  if(form=="piecewise")                        n_param <- 4

  # determine number of covariances
  n_covar <- (n_param*(n_param-1))/2

  # define parameters for labeling
  if(form=="linear")       param_names <- c("Intercept", "Slope")
  if(form=="quadratic")    param_names <- c("Intercept", "Linear Slope", "Quadratic Slope")
  if(form=="exponential")  param_names <- c("Intercept", "Total Change", "Growth Rate")
  if(form=="piecewise")    param_names <- c("Intercept", "Slope", "Change in Slope", "Changepoint")

  # FIXED EFFECTS -----
  fix_eff_est <- matrix(x$Parameter_Estimates[1:n_param,"Mean"], ncol=1)
  colnames(fix_eff_est) <- "Estimate"
  rownames(fix_eff_est) <- paste0(param_names, " Mean")

  # RANDOM EFFECTS -----

  ## INDIVIDUALS -----
  ran_eff_b_mat <- diag(n_param)
  diag(ran_eff_b_mat) <- x$Parameter_Estimates[(n_param+(1:n_param)),"Mean"]

  ran_eff_b_mat[upper.tri(ran_eff_b_mat)] <- x$Parameter_Estimates[(2*n_param+(1:n_covar)),"Mean"] # covariances on upper triangle
  ran_eff_b_mat[lower.tri(ran_eff_b_mat)] <- t(ran_eff_b_mat)[lower.tri(ran_eff_b_mat)] # copy to lower triangle
  ran_eff_b_mat[upper.tri(ran_eff_b_mat)] <- NA

  colnames(ran_eff_b_mat) <- rownames(ran_eff_b_mat) <- param_names

  ## GROUPS -----
  ran_eff_g_mat <- diag(n_param)
  diag(ran_eff_g_mat) <- x$Parameter_Estimates[(2*n_param+n_covar+(1:n_param)),"Mean"]

  ran_eff_g_mat[upper.tri(ran_eff_g_mat)] <- x$Parameter_Estimates[(3*n_param+n_covar+(1:n_covar)),"Mean"] # covariances on upper triangle
  ran_eff_g_mat[lower.tri(ran_eff_g_mat)] <- t(ran_eff_g_mat)[lower.tri(ran_eff_g_mat)] # copy to lower triangle
  ran_eff_g_mat[upper.tri(ran_eff_g_mat)] <- NA

  colnames(ran_eff_g_mat) <- rownames(ran_eff_g_mat) <- param_names

  ## return value
  out <- list("fix_eff_est" = fix_eff_est,
              "ran_eff_b_mat" = ran_eff_b_mat,
              "ran_eff_g_mat" = ran_eff_g_mat,
              "error_var" = x$Parameter_Estimates[(3*n_param+2*n_covar+1),"Mean"],
              "msrf" = x$Convergence$multivariate_psrf,
              "mean_psrf" = x$Convergence$mean_psrf,
              "DIC" = x$Model_Fit$dic)

  cat("Bayesian crossed random effects model\n")
  cat("Data:", x$Call$data, "\n")
  cat("Outcome:", x$Call$y_var, "\n")
  cat("Individuals:", x$Call$ind_id_var, "\n")
  cat("Group:", x$Call$cross_id_var, "\n")
  cat("\n")
  cat("Fixed Effect Parameters:\n")
  print(round(out$fix_eff_est,3))
  cat("\n")
  cat("Random Effect Parameters:\n")
  cat("Individual Random Effects Covariance Matrix:\n")
  print(round(out$ran_eff_b_mat,3), na.print="")
  cat("\n")
  cat("Group Random Effects Covariance Matrix:\n")
  print(round(out$ran_eff_g_mat,3), na.print="")
  cat("\n")
  cat("Error Var:", round(out$error_var,3), "\n")
  cat("Gelman's msrf:", round(out$msrf, 3), "\n")
  cat("Mean psrf:", round(out$mean_psrf, 3), "\n")
  cat("DIC:", round(out$DIC,3))
  cat("\n")

  invisible(x)
}

#' Bayesian Crossed Random Effects Model (CREM)
#'
#' @description
#' Estimates a Bayesian crossed random effects models (CREM) for longitudinal data with dynamic group membership. Four different choices for functional forms are provided: linear, quadratic, exponential, and piecewise. See Rohloff et al. (2024) for more details.
#'
#' @param data Data frame in long format, where each row describes a measurement occasion for a given individual. It is assumed that each individual has the same number of assigned timepoints (a.k.a., rows). There can be missingness in the outcome (`y_var`), but there cannot be missingness in time (`time_var`).
#' @param ind_id_var Name of column that contains ids for individuals with repeated measures in a longitudinal dataset (e.g., students).
#' @param cross_id_var Name of column that contains ids for the crossed factor (e.g., teachers).
#' @param time_var Name of column that contains the time variable. This column cannot contain any missing values.
#' @param y_var Name of column that contains the outcome variable. Missing values should be denoted by NA.
#' @param form Name of the functional form. Options include: ‘linear’ (default), ‘quadratic’, ‘exponential’, ‘piecewise’.
#' @param fixed_effects (optional) Starting values for the fixed effects parameters.
#' @param iters_adapt (optional) Number of iterations for adaptation of jags model (default = 5000).
#' @param iters_burn_in (optional) Number of iterations for burn-in (default = 50000).
#' @param iters_sampling (optional) Number of iterations for posterior sampling (default = 50000).
#' @param thin (optional) Thinning interval for posterior sampling (default = 15).
#' @param save_full_chains Logical indicating whether the MCMC chains from rjags should be saved (default = FALSE). Note, this should not be used regularly as it will result in an object with a large file size.
#' @param save_conv_chains Logical indicating whether the MCMC chains from rjags should be saved but only for the parameters monitored for convergence (default = FALSE). This would be useful for plotting traceplots for relevant model parameters to evaluate convergence behavior. Note, this should not be used regularly as it will result in an object with a large file size.
#' @param verbose Logical controlling whether progress messages/bars are generated (default = TRUE).
#'
#' @returns A list (an object of class `CREM`) with elements:
#' \item{Call}{A list of all specified function arguments.}
#' \item{Sample Size}{Number of subjects (`n_subj`), number of groups (`n_group`) and number of timepoints (`n_time`) in the dataset.}
#' \item{Data}{Dataset used for estimation.}
#' \item{Convergence}{Potential scale reduction factor (PSRF) for each parameter (`parameter_psrf`), Gelman multivariate scale reduction factor (`multivariate_psrf`), and mean PSRF (`mean_psrf`) to assess model convergence.}
#' \item{Model_Fit}{Deviance (`deviance`), effective number of parameters (`pD`), and Deviance information criterion (`dic`) to assess model fit.}
#' \item{Fitted_Values}{Data frame with the fitted values at each timepoint for each individual.}
#' \item{Functional_Form}{Functional form fitted.}
#' \item{Parameter_Estimates}{Data frame with posterior mean and 95% credible intervals for each model parameter.}
#' \item{Random_Coefficients}{List object with data frames providing individual random effects, group random effects, and random coefficients for each individual and group.}
#' \item{Run_Time}{Total run time for model fitting.}
#' \item{Full_MCMC_Chains}{If save_full_chains=TRUE, raw MCMC chains from rjags.}
#' \item{Convergence_MCMC_Chains}{If save_conv_chains=TRUE, raw MCMC chains from rjags but only for the parameters monitored for convergence.}
#'
#' @details
#' For more information on the model equation and priors implemented in this function, see Rohloff et al. (2024).
#'
#' Note, this function differs from the above reference by estimating the covariances between the random effects parameters. The variance-covariance matrices of the individual and group random effects have a scaled inverse-Wishart prior (see Peralta et al., 2022).
#'
#' @author Corissa T. Rohloff
#'
#' @references
#' Peralta, Y., Kohli, N., Lock, E. F., & Davison, M. L. (2022). Bayesian modeling of associations in bivariate piecewise linear mixed-effects models. Psychological Methods, 27(1), 44–64. https://doi.org/10.1037/met0000358
#'
#' Rohloff, C. T., Kohli, N., & Lock, E. F. (2024). Identifiability and estimability of Bayesian linear and nonlinear crossed random effects models. British Journal of Mathematical and Statistical Psychology. https://doi.org/10.1111/bmsp.12334
#'
#' @examples
#' \donttest{
#' # load simulated data
#' data(SimData_PCREM)
#' # fit Bayes_CREM()
#' results_pcrem <- Bayes_CREM(data = SimData_PCREM,
#'                             ind_id_var = "id",
#'                             cross_id_var = "teacherid",
#'                             time_var = "time",
#'                             y_var = "y",
#'                             form="piecewise")
#' # result summary
#' summary(results_pcrem)
#' # plot fitted results
#' plot(results_pcrem)
#' }
#'
#' @import stats
#'
#' @export
Bayes_CREM <- function(data,
                       ind_id_var, cross_id_var, time_var, y_var,
                       form="linear",
                       fixed_effects=NULL,
                       iters_adapt=5000, iters_burn_in=50000, iters_sampling=50000, thin=15,
                       save_full_chains=FALSE, save_conv_chains=FALSE,
                       verbose=TRUE){

  # START OF SETUP ----

  ## Initial data check
  data <- as.data.frame(data)

  ## Control progress messages/bars
  if(verbose) progress_bar = "text"
  if(!verbose) progress_bar = "none"

  ## Start tracking run time
  run_time_total_start <- Sys.time()

  ## Load module to compute DIC
  suppressMessages(rjags::load.module('dic'))

  ## Set number of chains - will use 3 chains across all models
  n_chains <- 3

  # Pull data for modeling ----
  y <- data[,paste0(y_var)] # vector of all outcome observations
  t <- data[,paste0(time_var)] # vector of all timepoints observations
  ind_id <- as.numeric(factor(data[,paste0(ind_id_var)])) # vector of individual ids
  cross_id <- as.numeric(factor(data[,paste0(cross_id_var)])) # vector of crossed factor ids

  # Define relevant variables ----
  n_subj <- length(unique(data[,paste0(ind_id_var)]))  # number of individuals (e.g., students)
  n_group <- length(unique(data[,paste0(cross_id_var)])) # number of groups in the crossed factor (e.g., teachers)
  n_obs <- length(y) # number of total observations

  # For piecewise model
  min_cp <- unique(sort(t))[2]
  max_cp <- unique(sort(t, decreasing = TRUE))[2]

  # Error messages for input ----

  ## data format warnings
  if(sum(length(y)==length(t))<1) stop('Columns for y_var and t_var must have the same dimensions')
  if(sum(is.na(t))>0) stop('Columns for t_var cannot have NA values (but y_var can)')
  if(min(t)!=0) warning('Prior assumes first time point measured at t=0')

  ## model specification warnings
  if(form!="linear"&&form!="quadratic"&&form!="exponential"&&form!="piecewise") stop("Specified functional form must be \"linear\", \"quadratic\", \"exponential\", or \"piecewise\"")

  # FULL MODEL ----

  ## Specify full JAGS model
  ## Linear
  if(form=="linear")      n_beta <- 2
  if(form=="linear")      full_spec <- textConnection(ln_crem)
  ## Quadratic
  if(form=="quadratic")   n_beta <- 3
  if(form=="quadratic")   full_spec <- textConnection(qd_crem)
  ## Exponential
  if(form=="exponential") n_beta <- 3
  if(form=="exponential") full_spec <- textConnection(ex_crem)
  ## Piecewise
  if(form=="piecewise")   n_beta <- 4
  if(form=="piecewise")   full_spec <- textConnection(pw_crem)

  ## Scale matrix for the Wishart distribution
  omega_b <- diag(n_beta)

  ## Variables to extract from full model
  param_recovery_full <- c('beta_ir', 'beta_mean', 'b0', 'g0',
                           'cov_b', 'var_b',
                           'cov_g', 'var_g',
                           'mu_y', 'sigma2_error',
                           'for_conv', 'pD', 'deviance')

  ## Create data list for JAGS model
  data_list <- list("t" = t, "y" = y,
                    "n_obs" = n_obs, "n_subj" = n_subj, 'n_group' = n_group,
                    "ind_id" = ind_id, 'cross_id' = cross_id,
                    "n_beta" = n_beta, "omega_b" = omega_b)
  if(form=="piecewise") data_list$min_cp <- min_cp
  if(form=="piecewise") data_list$max_cp <- max_cp

  ## Compile Info for Initial Values
  initial_vals <- vector('list', n_chains)
  if(!is.null(fixed_effects)){
    for(i in 1:n_chains){initial_vals[[i]]$beta_mean <- fixed_effects}
  }

  ## Set Seeds for Reproducible Code
  seeds <- sample(1:10, n_chains, replace = FALSE) # randomly generate non-repeating integers
                                                   # sensitive to set.seed()
  for(i in 1:n_chains){
    initial_vals[[i]]$.RNG.name <- "base::Wichmann-Hill" # arbitrary
    initial_vals[[i]]$.RNG.seed <- seeds[i]
  }

  ## Create JAGS model
  if(verbose) cat("Calibrating MCMC...\n")
  full_model <- rjags::jags.model(full_spec,
                                  data = data_list,
                                  inits = initial_vals,
                                  n.chains = n_chains,
                                  n.adapt = iters_adapt,
                                  quiet = TRUE)

  # burn-in
  if(verbose) cat("Burn in of jags model...\n")
  update(full_model, iters_burn_in, progress.bar=progress_bar)

  # sampling
  if(verbose) cat("Collecting samples...\n")
  full_out <- rjags::jags.samples(full_model,
                                  variable.names = param_recovery_full,
                                  n.iter = iters_sampling,
                                  thin = thin,
                                  progress.bar=progress_bar)

  # Compiling Full Results -----

  ## Call Info
  call_info <- as.list(match.call())
  def_args <- formals(sys.function())
  for(arg in names(def_args)){
    if(!(arg %in% names(call_info))){
      call_info[[arg]] <- def_args[[arg]]
    }
  }

  if(form=="linear")      param_names <- c("beta_0_mean", "beta_1_mean",
                                           "var_b_0", "var_b_1",
                                           "cov_b_01",
                                           "var_g_0", "var_g_1",
                                           "cov_g_01",
                                           "error_var")
  if(form=="quadratic" |
     form=="exponential") param_names <- c("beta_0_mean", "beta_1_mean", "beta_2_mean",
                                           "var_b_0", "var_b_1", "var_b_2",
                                           "cov_b_01", "cov_b_02", "cov_b_12",
                                           "var_g_0", "var_g_1", "var_g_2",
                                           "cov_g_01", "cov_g_02", "cov_g_12",
                                           "error_var")

  if(form=="piecewise")   param_names <- c("beta_0_mean", "beta_1_mean", "beta_2_mean", "beta_cp_mean",
                                           "var_b_0", "var_b_1", "var_b_2", "var_b_cp",
                                           "cov_b_01", "cov_b_02", "cov_b_12", "cov_b_0cp", "cov_b_1cp", "cov_b_2cp",
                                           "var_g_0", "var_g_1", "var_g_2", "var_g_cp",
                                           "cov_g_01", "cov_g_02", "cov_g_12", "cov_g_0cp", "cov_g_1cp", "cov_g_2cp",
                                           "error_var")

  ## Convergence
  mcmc_list <- coda::as.mcmc.list(full_out$for_conv)
  for(i in 1:3){colnames(mcmc_list[[i]]) <- param_names}

  gelman_msrf <- coda::gelman.diag(mcmc_list) # individual parameter psrf, and multivariate psrf
  parameter_psrf <- data.frame(point_est = gelman_msrf$psrf[,1],
                               upper_ci = gelman_msrf$psrf[,2])
  multivariate_psrf <- gelman_msrf$mpsrf
  mean_psrf <- mean(parameter_psrf$point_est) # mean psrf across parameters
  convergence <- list(parameter_psrf=parameter_psrf, multivariate_psrf=multivariate_psrf, mean_psrf=mean_psrf)

  ## Model Fit
  deviance <- mean(full_out$deviance)
  pD <- mean(full_out$pD)
  dic <- deviance + pD
  model_fit <- list(deviance=deviance, pD=pD, dic=dic)

  ## Fitted Values
  y_mean <- summary(full_out$mu_y, FUN='mean')[[1]]
  fit_vals <- cbind(ind_id, cross_id, t, y, y_mean)
  fit_vals <- as.data.frame(fit_vals)
  names(fit_vals) <- c(ind_id_var, cross_id_var, time_var, y_var, "fitted")

  ## Random Effects and Coefficients
  ranef_b <- cbind(unique(ind_id), summary(full_out$b0, FUN='mean')[[1]])
  ranef_b <- as.data.frame(ranef_b)
  names(ranef_b) <- c(ind_id_var, gsub("var_", "", grep("^var_b", param_names, value = TRUE)))

  ranef_g <- cbind(unique(cross_id), summary(full_out$g0, FUN='mean')[[1]])
  ranef_g <- as.data.frame(ranef_g)
  names(ranef_g) <- c(cross_id_var, gsub("var_", "", grep("^var_g", param_names, value = TRUE)))

  rancoef <- cbind(ind_id, cross_id, summary(full_out$beta_ir, FUN='mean')[[1]])
  rancoef <- as.data.frame(rancoef)
  names(rancoef) <- c(ind_id_var, cross_id_var, gsub("var_", "", grep("^var_b", param_names, value = TRUE)))

  ran_ef_coef <- list("ranef_b" = ranef_b,
                      "ranef_g" = ranef_g,
                      "rancoef" = rancoef)

  ## Parameter Estimates
  sum_mcmc <- summary(mcmc_list) # parameter estimates
  param_est <- data.frame(Mean = sum_mcmc$statistics[,1],
                          CI_Lower = sum_mcmc$quantiles[,1],
                          CI_Upper = sum_mcmc$quantiles[,5])

  ## Stop tracking run time
  run_time_total_end <- Sys.time()
  run_time_total <- run_time_total_end - run_time_total_start

  my_results <- list('Call' = call_info,
                     'Sample_Size' = list(n_subj=n_subj, n_group=n_group, n_time=length(t)),
                     'Data' = data,
                     'Convergence' = convergence,
                     'Model_Fit' = model_fit,
                     'Fitted_Values' = fit_vals,
                     'Functional_Form' = form,
                     'Parameter_Estimates' = param_est,
                     'Random_Coefficients' = ran_ef_coef,
                     'Run_Time' = format(run_time_total))
  if(save_full_chains==TRUE){my_results$Full_MCMC_Chains=full_out}
  if(save_conv_chains==TRUE){my_results$Convergence_MCMC_Chains=mcmc_list}
  class(my_results) <- c('CREM', 'BEND')
  return(my_results)
}

# Models ----

## Linear ----
ln_crem <- "model{
  ##### Level-1 Model #####
  for(j in 1:n_obs){
    y[j] ~ dnorm(mu_y[j], tau_y)
    mu_y[j] <- beta_ir[j,1] + beta_ir[j,2]*t[j]
    beta_ir[j,1] <- beta_mean[1] + b0[ind_id[j],1] + g0[cross_id[j],1]
    beta_ir[j,2] <- beta_mean[2] + b0[ind_id[j],2] + g0[cross_id[j],2]
  }

  ##### Loop over Individuals #####
  for(i in 1:n_subj){
    for(p in 1:n_beta){
      # scale variances
      b0[i,p] <- b_scale_c[p]*b0_raw[i,p]
    }
    # generate raw variance-covariance matrix
    b0_raw[i,1:n_beta] ~ dmnorm(rep(0,n_beta), tau_b_raw[1:n_beta,1:n_beta])
  }

  ##### Loop over Crossed Factor #####
  for(r in 1:n_group){
    for(p in 1:n_beta){
      # scale variances
      g0[r,p] <- g_scale_c[p]*g0_raw[r,p]
    }
    # generate raw variance-covariance matrix
    g0_raw[r,1:n_beta] ~ dmnorm(rep(0,n_beta), tau_g_raw[1:n_beta,1:n_beta])
  }

  ##### Priors #####
  ## Priors for Fixed Effects ##
  for(p in 1:n_beta){
   beta_mean[p] ~ dnorm(0, 0.00001)
  }

  ## Priors for Residual Variance Components ##
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  ## Priors for Individual Random Effects ##
  tau_b_raw[1:n_beta,1:n_beta] ~ dwish(omega_b[1:n_beta,1:n_beta], n_beta+1)
  sigma2_b_raw[1:n_beta,1:n_beta] <- inverse(tau_b_raw[,])

  ## Priors for Crossed Random Effects ##
  tau_g_raw[1:n_beta,1:n_beta] ~ dwish(omega_b[1:n_beta,1:n_beta], n_beta+1)
  sigma2_g_raw[1:n_beta,1:n_beta] <- inverse(tau_g_raw[,])

  ## Priors for Scaling Constants ##
  for(p in 1:n_beta){
    # For Individual Effects #
    b_scale_c[p] ~ dunif(0, 100)
    # For Crossed Effects #
    g_scale_c[p] ~ dunif(0, 100)
  }

  ##### Variance-Covariance Matrix Elements #####
  ## Individual Effects ##
  # Scale Covariance Matrices #
  for(p in 1:n_beta){
    for(p_prime in 1:n_beta){
      cov_b_mat[p,p_prime] <- b_scale_c[p]*b_scale_c[p_prime]*sigma2_b_raw[p,p_prime]
    }
  }

  # Variances #
  for(p in 1:n_beta){
    var_b[p] <- cov_b_mat[p,p]
  }

  # Covariances #
  cov_b[1] <- cov_b_mat[2,1] # 0,1

  ## Crossed Effects ##
  # Scale Covariance Matrices #
  for(p in 1:n_beta){
    for(p_prime in 1:n_beta){
      cov_g_mat[p,p_prime] <- g_scale_c[p]*g_scale_c[p_prime]*sigma2_g_raw[p,p_prime]
    }
  }

  # Variances #
  for(p in 1:n_beta){
    var_g[p] <- cov_g_mat[p,p]
  }

  # Covariances #
  cov_g[1] <- cov_g_mat[2,1] # 0,1

  ##### Collect important parameters to assess convergence #####
  for_conv[1:2] <- beta_mean
  for_conv[3:4] <- var_b
  for_conv[5] <- cov_b
  for_conv[6:7] <- var_g
  for_conv[8] <- cov_g
  for_conv[9] <- sigma2_error
  }
"

## Quadratic ----
qd_crem <- "model{
  ##### Level-1 Model #####
  for(j in 1:n_obs){
    y[j] ~ dnorm(mu_y[j], tau_y)
    mu_y[j] <- beta_ir[j,1] + beta_ir[j,2]*t[j] + beta_ir[j,3]*(t[j]^2)
    beta_ir[j,1] <- beta_mean[1] + b0[ind_id[j],1] + g0[cross_id[j],1]
    beta_ir[j,2] <- beta_mean[2] + b0[ind_id[j],2] + g0[cross_id[j],2]
    beta_ir[j,3] <- beta_mean[3] + b0[ind_id[j],3] + g0[cross_id[j],3]
  }

  ##### Loop over Individuals #####
  for(i in 1:n_subj){
    for(p in 1:n_beta){
      # scale variances
      b0[i,p] <- b_scale_c[p]*b0_raw[i,p]
    }
    # generate raw variance-covariance matrix
    b0_raw[i,1:n_beta] ~ dmnorm(rep(0,n_beta), tau_b_raw[1:n_beta,1:n_beta])
  }

  ##### Loop over Crossed Factor #####
  for(r in 1:n_group){
    for(p in 1:n_beta){
      # scale variances
      g0[r,p] <- g_scale_c[p]*g0_raw[r,p]
    }
    # generate raw variance-covariance matrix
    g0_raw[r,1:n_beta] ~ dmnorm(rep(0,n_beta), tau_g_raw[1:n_beta,1:n_beta])
  }

  ##### Priors #####
  ## Priors for Fixed Effects ##
  for(p in 1:n_beta){
   beta_mean[p] ~ dnorm(0, 0.00001)
  }

  ## Priors for Residual Variance Components ##
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  ## Priors for Individual Random Effects ##
  tau_b_raw[1:n_beta,1:n_beta] ~ dwish(omega_b[1:n_beta,1:n_beta], n_beta+1)
  sigma2_b_raw[1:n_beta,1:n_beta] <- inverse(tau_b_raw[,])

  ## Priors for Crossed Random Effects ##
  tau_g_raw[1:n_beta,1:n_beta] ~ dwish(omega_b[1:n_beta,1:n_beta], n_beta+1)
  sigma2_g_raw[1:n_beta,1:n_beta] <- inverse(tau_g_raw[,])

  ## Priors for Scaling Constants ##
  for(p in 1:n_beta){
    # For Individual Effects #
    b_scale_c[p] ~ dunif(0, 100)
    # For Crossed Effects #
    g_scale_c[p] ~ dunif(0, 100)
  }

  ##### Variance-Covariance Matrix Elements #####
  ## Individual Effects ##
  # Scale Covariance Matrices #
  for(p in 1:n_beta){
    for(p_prime in 1:n_beta){
      cov_b_mat[p,p_prime] <- b_scale_c[p]*b_scale_c[p_prime]*sigma2_b_raw[p,p_prime]
    }
  }

  # Variances #
  for(p in 1:n_beta){
    var_b[p] <- cov_b_mat[p,p]
  }

  # Covariances #
  cov_b[1] <- cov_b_mat[2,1] # 0,1
  cov_b[2] <- cov_b_mat[3,1] # 0,2
  cov_b[3] <- cov_b_mat[3,2] # 1,2

  ## Crossed Effects ##
  # Scale Covariance Matrices #
  for(p in 1:n_beta){
    for(p_prime in 1:n_beta){
      cov_g_mat[p,p_prime] <- g_scale_c[p]*g_scale_c[p_prime]*sigma2_g_raw[p,p_prime]
    }
  }

  # Variances #
  for(p in 1:n_beta){
    var_g[p] <- cov_g_mat[p,p]
  }

  # Covariances #
  cov_g[1] <- cov_g_mat[2,1] # 0,1
  cov_g[2] <- cov_g_mat[3,1] # 0,2
  cov_g[3] <- cov_g_mat[3,2] # 1,2

  ##### Collect important parameters to assess convergence #####
  for_conv[1:3] <- beta_mean
  for_conv[4:6] <- var_b
  for_conv[7:9] <- cov_b
  for_conv[10:12] <- var_g
  for_conv[13:15] <- cov_g
  for_conv[16] <- sigma2_error
  }
"

## Exponential ----
ex_crem <- "model{
  ##### Level-1 Model #####
  for(j in 1:n_obs){
    y[j] ~ dnorm(mu_y[j], tau_y)
    mu_y[j] <- beta_ir[j,1] + beta_ir[j,2]*(1-exp(-beta_ir[j,3]*t[j]))
    beta_ir[j,1] <- beta_mean[1] + b0[ind_id[j],1] + g0[cross_id[j],1]
    beta_ir[j,2] <- beta_mean[2] + b0[ind_id[j],2] + g0[cross_id[j],2]
    beta_ir[j,3] <- beta_mean[3] + b0[ind_id[j],3] + g0[cross_id[j],3]
  }

  ##### Loop over Individuals #####
  for(i in 1:n_subj){
    for(p in 1:n_beta){
      # scale variances
      b0[i,p] <- b_scale_c[p]*b0_raw[i,p]
    }
    # generate raw variance-covariance matrix
    b0_raw[i,1:n_beta] ~ dmnorm(rep(0,n_beta), tau_b_raw[1:n_beta,1:n_beta])
  }

  ##### Loop over Crossed Factor #####
  for(r in 1:n_group){
    for(p in 1:n_beta){
      # scale variances
      g0[r,p] <- g_scale_c[p]*g0_raw[r,p]
    }
    # generate raw variance-covariance matrix
    g0_raw[r,1:n_beta] ~ dmnorm(rep(0,n_beta), tau_g_raw[1:n_beta,1:n_beta])
  }

  ##### Priors #####
  ## Priors for Fixed Effects ##
  for(p in 1:n_beta){
   beta_mean[p] ~ dnorm(0, 0.00001)
  }

  ## Priors for Residual Variance Components ##
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  ## Priors for Individual Random Effects ##
  tau_b_raw[1:n_beta,1:n_beta] ~ dwish(omega_b[1:n_beta,1:n_beta], n_beta+1)
  sigma2_b_raw[1:n_beta,1:n_beta] <- inverse(tau_b_raw[,])

  ## Priors for Crossed Random Effects ##
  tau_g_raw[1:n_beta,1:n_beta] ~ dwish(omega_b[1:n_beta,1:n_beta], n_beta+1)
  sigma2_g_raw[1:n_beta,1:n_beta] <- inverse(tau_g_raw[,])

  ## Priors for Scaling Constants ##
  for(p in 1:n_beta){
    # For Individual Effects #
    b_scale_c[p] ~ dunif(0, 100)
    # For Crossed Effects #
    g_scale_c[p] ~ dunif(0, 100)
  }

  ##### Variance-Covariance Matrix Elements #####
  ## Individual Effects ##
  # Scale Covariance Matrices #
  for(p in 1:n_beta){
    for(p_prime in 1:n_beta){
      cov_b_mat[p,p_prime] <- b_scale_c[p]*b_scale_c[p_prime]*sigma2_b_raw[p,p_prime]
    }
  }

  # Variances #
  for(p in 1:n_beta){
    var_b[p] <- cov_b_mat[p,p]
  }

  # Covariances #
  cov_b[1] <- cov_b_mat[2,1] # 0,1
  cov_b[2] <- cov_b_mat[3,1] # 0,2
  cov_b[3] <- cov_b_mat[3,2] # 1,2

  ## Crossed Effects ##
  # Scale Covariance Matrices #
  for(p in 1:n_beta){
    for(p_prime in 1:n_beta){
      cov_g_mat[p,p_prime] <- g_scale_c[p]*g_scale_c[p_prime]*sigma2_g_raw[p,p_prime]
    }
  }

  # Variances #
  for(p in 1:n_beta){
    var_g[p] <- cov_g_mat[p,p]
  }

  # Covariances #
  cov_g[1] <- cov_g_mat[2,1] # 0,1
  cov_g[2] <- cov_g_mat[3,1] # 0,2
  cov_g[3] <- cov_g_mat[3,2] # 1,2

  ##### Collect important parameters to assess convergence #####
  for_conv[1:3] <- beta_mean
  for_conv[4:6] <- var_b
  for_conv[7:9] <- cov_b
  for_conv[10:12] <- var_g
  for_conv[13:15] <- cov_g
  for_conv[16] <- sigma2_error
  }
"

## Piecewise ----
pw_crem <- "model{
  ##### Level-1 Model #####
  for(j in 1:n_obs){
    y[j] ~ dnorm(mu_y[j], tau_y)
    mu_y[j] <- beta_ir[j,1] + beta_ir[j,2]*t[j] + beta_ir[j,3]*(max(0, t[j]-beta_ir[j,4]))
    beta_ir[j,1] <- beta_mean[1] + b0[ind_id[j],1] + g0[cross_id[j],1]
    beta_ir[j,2] <- beta_mean[2] + b0[ind_id[j],2] + g0[cross_id[j],2]
    beta_ir[j,3] <- beta_mean[3] + b0[ind_id[j],3] + g0[cross_id[j],3]
    beta_ir[j,4] <- beta_mean[4] + b0[ind_id[j],4] + g0[cross_id[j],4]
  }

  ##### Loop over Individuals #####
  for(i in 1:n_subj){
    for(p in 1:n_beta){
      # scale variances
      b0[i,p] <- b_scale_c[p]*b0_raw[i,p]
    }
    # generate raw variance-covariance matrix
    b0_raw[i,1:n_beta] ~ dmnorm(rep(0,n_beta), tau_b_raw[1:n_beta,1:n_beta])
  }

  ##### Loop over Crossed Factor #####
  for(r in 1:n_group){
    for(p in 1:n_beta){
      # scale variances
      g0[r,p] <- g_scale_c[p]*g0_raw[r,p]
    }
    # generate raw variance-covariance matrix
    g0_raw[r,1:n_beta] ~ dmnorm(rep(0,n_beta), tau_g_raw[1:n_beta,1:n_beta])
  }

  ##### Priors #####
  ## Priors for Fixed Effects ##
  for(p in 1:(n_beta-1)){
   beta_mean[p] ~ dnorm(0, 0.00001)
  }
  beta_mean[n_beta] ~ dunif(min_cp,max_cp)

  ## Priors for Residual Variance Components ##
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  ## Priors for Individual Random Effects ##
  tau_b_raw[1:n_beta,1:n_beta] ~ dwish(omega_b[1:n_beta,1:n_beta], n_beta+1)
  sigma2_b_raw[1:n_beta,1:n_beta] <- inverse(tau_b_raw[,])

  ## Priors for Crossed Random Effects ##
  tau_g_raw[1:n_beta,1:n_beta] ~ dwish(omega_b[1:n_beta,1:n_beta], n_beta+1)
  sigma2_g_raw[1:n_beta,1:n_beta] <- inverse(tau_g_raw[,])

  ## Priors for Scaling Constants ##
  for(p in 1:n_beta){
    # For Individual Effects #
    b_scale_c[p] ~ dunif(0, 100)
    # For Crossed Effects #
    g_scale_c[p] ~ dunif(0, 100)
  }

  ##### Variance-Covariance Matrix Elements #####
  ## Individual Effects ##
  # Scale Covariance Matrices #
  for(p in 1:n_beta){
    for(p_prime in 1:n_beta){
      cov_b_mat[p,p_prime] <- b_scale_c[p]*b_scale_c[p_prime]*sigma2_b_raw[p,p_prime]
    }
  }

  # Variances #
  for(p in 1:n_beta){
    var_b[p] <- cov_b_mat[p,p]
  }

  # Covariances #
  cov_b[1] <- cov_b_mat[2,1] # 0,1
  cov_b[2] <- cov_b_mat[3,1] # 0,2
  cov_b[3] <- cov_b_mat[3,2] # 1,2
  cov_b[4] <- cov_b_mat[4,1] # 0,3
  cov_b[5] <- cov_b_mat[4,2] # 1,3
  cov_b[6] <- cov_b_mat[4,3] # 2,3

  ## Crossed Effects ##
  # Scale Covariance Matrices #
  for(p in 1:n_beta){
    for(p_prime in 1:n_beta){
      cov_g_mat[p,p_prime] <- g_scale_c[p]*g_scale_c[p_prime]*sigma2_g_raw[p,p_prime]
    }
  }

  # Variances #
  for(p in 1:n_beta){
    var_g[p] <- cov_g_mat[p,p]
  }

  # Covariances #
  cov_g[1] <- cov_g_mat[2,1] # 0,1
  cov_g[2] <- cov_g_mat[3,1] # 0,2
  cov_g[3] <- cov_g_mat[3,2] # 1,2
  cov_g[4] <- cov_g_mat[4,1] # 0,3
  cov_g[5] <- cov_g_mat[4,2] # 1,3
  cov_g[6] <- cov_g_mat[4,3] # 2,3

  ##### Collect important parameters to assess convergence #####
  for_conv[1:4] <- beta_mean
  for_conv[5:8] <- var_b
  for_conv[9:14] <- cov_b
  for_conv[15:18] <- var_g
  for_conv[19:24] <- cov_g
  for_conv[25] <- sigma2_error
  }
"

#' Extract fitted values
#'
#' @description
#' Extracts the individual fitted values from a fitted model of class "BPREM", "CREM", or "PREM".
#'
#' @param x An object of class "BPREM", "CREM", or "PREM".
#' @param ... Additional arguments.
#'
#' @returns Returns a data frame of the fitted values for each individual and timepoint.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_prem)
#' # get fitted values
#' getFitted(results_prem)
#'
#' @export
getFitted <- function(x, ...) UseMethod(("getFitted"))

#' @rdname getFitted
#' @export
getFitted.BEND <- function(x, ...){

  out <- x$Fitted_Values
  class(out) <- c("getFitted.BEND", class(out))
  return(out)

}

#' @rdname getFitted
#' @export
print.getFitted <- function(x, ...){

  print(x, na.print="")
  cat("\n")

  invisible(x)
}

#' Bayesian Piecewise Random Effects Model (PREM) + Extensions
#'
#' @description Estimates a Bayesian piecewise random effects model (PREM), with some useful extensions. There are three model options included in this function:
#' * `PREM` estimates a Bayesian piecewise random effects model with a latent number of changepoints (default). Allows the inclusion of outcome-predictive covariates (`CI-PREM`).
#' * `PREMM` estimates a piecewise random effects mixture model for a given number of latent classes and a latent number of possible changepoints in each class.
#' * `CI-PREMM` estimates a covariate influenced piecewise random effects mixture model for a given number of latent classes and a latent number of possible changepoints in each class. Allows the inclusion of outcome- and/or class-predictive covariates.
#' See Lock et al. (2018) and Lamm (2022) for more details.
#'
#' @param data Data frame in long format, where each row describes a measurement occasion for a given individual. It is assumed that each individual has the same number of assigned timepoints (a.k.a., rows). There can be missingness in the outcome (`y_var`), but there cannot be missingness in time (`time_var`).
#' @param id_var Name of column that contains ids for individuals with repeated measures in a longitudinal dataset.
#' @param time_var Name of column that contains the time variable. This column cannot contain any missing values.
#' @param y_var Name of column that contains the outcome variable. Missing values should be denoted by NA.
#' @param n_class Number of latent classes (default = 1). Note, CI-PREMM only allows for two classes.
#' @param max_cp Maximum number of changepoints in each latent class (default = 2).
#' @param class_predictive_vars Name(s) of column(s) that contain class-predictive covariates (time-invariant only). Give a vector of names if multiple covariates. Note, there cannot be any missingness in the covariates.
#' @param outcome_predictive_vars Name(s) of column(s) that contain outcome-predictive covariates (time-varying or -invariant). Give a vector of names if multiple covariates. Note, there cannot be any missingness in the covariates.
#' @param scale_prior Prior for the scale parameter for the hierarchical random effects. Options include: ‘uniform’ (scaled uniform prior; default) or ‘hc’ (scaled half-cauchy prior).
#' @param alpha Concentration parameter for Dirichlet prior for latent classes (default = 1). This can be a vector of values corresponding to the number of classes (specified by n_class). Note, this is not used for CI-PGMM.
#' @param cp_prior Prior for the number of changepoints in each class. Options include: 'binomial' (default) or 'uniform'.
#' @param binom_prob Probability for binomial prior, if specified (default = 0.5).
#' @param iters_adapt (optional) Number of iterations for adaptation of jags model (default = 1000).
#' @param iters_burn_in (optional) Number of iterations for burn-in (default = 20000).
#' @param iters_sampling (optional) Number of iterations for posterior sampling (default = 30000).
#' @param thin (optional) Thinning interval for posterior sampling (default = 15).
#' @param save_full_chains Logical indicating whether the MCMC chains from rjags should be saved (default = FALSE). Note, this should not be used regularly as it will result in an object with a large file size.
#' @param save_conv_chains Logical indicating whether the MCMC chains from rjags should be saved but only for the parameters monitored for convergence (default = FALSE). This would be useful for plotting traceplots for relevant model parameters to evaluate convergence behavior. Note, this should not be used regularly as it will result in an object with a large file size.
#' @param verbose Logical controlling whether progress messages/bars are generated (default = TRUE).
#'
#' @returns A list (an object of class `PREM`) with elements:
#' \item{Call}{A list of all specified function arguments.}
#' \item{Sample Size}{Number of subjects (`n_subj`) and number of timepoints (`n_time`) in the dataset.}
#' \item{Data}{Dataset used for estimation.}
#' \item{Convergence}{Potential scale reduction factor (PSRF) for each parameter (`parameter_psrf`), Gelman multivariate scale reduction factor (`multivariate_psrf`), and mean PSRF (`mean_psrf`) to assess model convergence.}
#' \item{Model_Fit}{Deviance (`deviance`), effective number of parameters (`pD`), and Deviance information criterion (`dic`) to assess model fit.}
#' \item{Fitted_Values}{Data frame with the fitted values at each timepoint for each individual.}
#' \item{Parameter_Estimates}{Data frame with posterior mean and 95% credible intervals for each model parameter.}
#' \item{Random_Coefficients}{List object with data frames providing random coefficients for each individual.}
#' \item{Run_Time}{Total run time for model fitting.}
#' \item{Full_MCMC_Chains}{If save_full_chains=TRUE, raw MCMC chains from rjags.}
#' \item{Convergence_MCMC_Chains}{If save_conv_chains=TRUE, raw MCMC chains from rjags but only for the parameters monitored for convergence.}
#' `Class_Information` contains a list with elements:
#' \item{class_membership}{Vector of length n with class membership assignments for each individual.}
#' \item{individ_class_probability}{nxC matrix with each individual’s probabilities of belonging to each class conditional on their class-predictive covariates (when applicable) and growth curve.}
#' \item{unconditional_class_probability}{This output will differ based on which model was fit. For a PREM or CI-PREM, this will equal 1 as there is only one class. For a PREMM or CI-PREMM with only outcome-predictive covariates, this will be a vector of length C denoting the population probability of belonging to each class. For a CI-PREMM with class-predictive covariates, this will be a vector of length n denoting the probability of each individual belonging to the non-reference class (Class 2) based on their class-predictive covariates only.}
#'
#' @details
#' For more information on the model equation and priors implemented in this function, see Lamm et al. (2022; CI-PREMM) and Lock et al. (2018; PREMM).
#'
#' @author Corissa T. Rohloff, Rik Lamm, Eric F. Lock
#'
#' @references Lamm, R. (2022). Incorporation of covariates in Bayesian piecewise growth mixture models. https://hdl.handle.net/11299/252533
#'
#' Lock, E. F., Kohli, N., & Bose, M. (2018). Detecting multiple random changepoints in Bayesian piecewise growth mixture models. Psychometrika, 83(3), 733–750. https://doi.org/10.1007/s11336-017-9594-5
#'
#' @examples
#' \donttest{
#' # load simulated data
#' data(SimData_PREM)
#'
#' # PREM ---------------------------------------------------------------------------------
#' # fit Bayes_PREM()
#' results_prem <- Bayes_PREM(data = SimData_PREM,
#'                            id_var = "id",
#'                            time_var = "time",
#'                            y_var = "y")
#' # result summary
#' summary(results_prem)
#' # plot fitted results
#' plot(results_prem)
#'
#' # CI-PREM ---------------------------------------------------------------------------------
#' # fit Bayes_PREM()
#' results_ciprem <- Bayes_PREM(data = SimData_PREM,
#'                              id_var = "id",
#'                              time_var = "time",
#'                              y_var = "y",
#'                              outcome_predictive_vars = "outcome_pred_1")
#' # result summary
#' summary(results_ciprem)
#' # plot fitted results
#' plot(results_ciprem)
#'
#' # PREMM ---------------------------------------------------------------------------------
#' # fit Bayes_PREM()
#' results_premm <- Bayes_PREM(data = SimData_PREM,
#'                             id_var = "id",
#'                             time_var = "time",
#'                             y_var = "y",
#'                             n_class = 2)
#' # result summary
#' summary(results_premm)
#' # plot fitted results
#' plot(results_premm)
#'
#'
#' # CI-PREMM ---------------------------------------------------------------------------------
#' # fit Bayes_PREM()
#' results_cipremm <- Bayes_PREM(data = SimData_PREM,
#'                               id_var = "id",
#'                               time_var = "time",
#'                               y_var = "y",
#'                               n_class = 2,
#'                               class_predictive_vars = c("class_pred_1", "class_pred_2"),
#'                               outcome_predictive_vars = "outcome_pred_1")
#' # result summary
#' summary(results_cipremm)
#' # plot fitted results
#' plot(results_cipremm)
#' }
#'
#' @import stats
#'
#' @export
Bayes_PREM <- function(data,
                       id_var, time_var, y_var,
                       n_class=1, max_cp=2,
                       class_predictive_vars=NULL, outcome_predictive_vars=NULL,
                       scale_prior='uniform', alpha=1,
                       cp_prior='binomial', binom_prob=0.5,
                       iters_adapt=1000, iters_burn_in=20000, iters_sampling=30000, thin=15,
                       save_full_chains=FALSE, save_conv_chains=FALSE,
                       verbose=TRUE){

  # START OF SETUP ----

  ## Initial data check
  data <- as.data.frame(data)

  ## Check that number of rows is the same for each id_var
  rows_per_id <- table(data[,id_var])
  if(length(unique(rows_per_id)) != 1L) stop("The number of rows should be the same for each individual. \nSee ?Bayes_PREM for more information.")

  ## Control progress messages/bars
  if(verbose) progress_bar = "text"
  if(!verbose) progress_bar = "none"

  ## Start tracking run time
  run_time_total_start <- Sys.time()

  ## Load module to compute DIC
  suppressMessages(rjags::load.module('dic'))

  ## Set number of chains - will use 3 chains across all models
  n_chains <- 3

  # Reshape data for modeling ----

  ## outcome data - matrix form
  y <- reshape(data[,c(id_var, time_var, y_var)],
               idvar=id_var,
               timevar=time_var,
               direction='wide')
  y <- unname(as.matrix(y[,names(y)!=id_var]))

  ## time data - matrix form
  ## should be the same dimensions as y
  x <- matrix(data[,c(time_var)],
              byrow=TRUE,
              nrow=dim(y)[1],
              ncol=dim(y)[2])

  # Define relevant variables ----
  n_subj <- nrow(y)
  n_time <- ncol(y)
  max_beta <- max_cp + 2

  max_time <- max(x)
  min_time <- min(x)
  min_cp_mean <- unique(sort(x))[2]
  max_cp_mean <- unique(sort(x, decreasing=TRUE))[2]

  n_cov_class_predictive <- if(is.null(class_predictive_vars)) 0 else length(class_predictive_vars)
  n_cov_outcome_predictive <- if(is.null(outcome_predictive_vars)) 0 else length(outcome_predictive_vars)

  ## Define number of parameters (used to assess convergence later)
  n_params <- 1 + n_class*(2*max_beta) + n_class*(2*max_cp) + n_cov_outcome_predictive + n_cov_class_predictive
  if(n_cov_class_predictive>0) n_params <- n_params + 1 # only for class_predictive models (logistic intercept)

  ## for CI-PREMM only
  time_vec <- x[1,]

  # Reshape covariates (if applicable) ----

  ## class predictive covariates - matrix form
  ## assumed to be time invariant
  if(n_cov_class_predictive>0){
    class_predictive_covariates <- unique(data[,c(id_var, class_predictive_vars)])
    class_predictive_covariates <- unname(as.matrix(class_predictive_covariates[,names(class_predictive_covariates)!=id_var]))
  }

  ## outcome predictive covariates - array of matrices
  ## assumed to be time varying or invariant
  if(n_cov_outcome_predictive>0){
    outcome_predictive_covariates_list <- list()
    for(i in 1:n_cov_outcome_predictive){
      cov_matrix <- reshape(data[,c(id_var, time_var, outcome_predictive_vars[i])],
                            idvar=id_var,
                            timevar=time_var,
                            direction='wide')
      outcome_predictive_covariates_list[[i]] <- unname(as.matrix(cov_matrix[,names(cov_matrix)!=id_var]))
    }
    outcome_predictive_covariates <- simplify2array(outcome_predictive_covariates_list) # dimensions = (n_subj, n_time, n_cov_outcome_predictive)
  }

  # Error messages for input ----

  ## data format warnings
  if(sum(is.na(x)>0)) stop('Columns for t_var cannot have NA values (but y_var can)')
  if(min(x)!=0) warning('Prior assumes first time point measured at x=0')

  ## model specification warnings
  if(cp_prior!='binomial' && cp_prior!='uniform') stop('Input for cp_prior must be \'binomial\' or \'uniform\'')
  if(scale_prior!='uniform' && scale_prior!='hc') stop('Input for scale_prior must be \'uniform\' or \'hc\'')
  if(!is.null(class_predictive_vars) && !is.null(outcome_predictive_vars) && n_class>2) stop('CI-PREMM only allows for two classes')

  # Define mean & precision for random coefficients ----
  mean <- c(mean(y[,1], na.rm=TRUE), 0, rep(0,max_cp))
  prec_param_int <- 1/var(y[,1], na.rm=TRUE)
  prec_param <- (1/(max_cp*sd(y, na.rm=TRUE)/(sd(x))))^2
  prec_vec <- c(prec_param_int, rep(prec_param, max_cp+1))
  prec <- diag(max_beta)
  diag(prec) <- prec_vec

  ## Bernoulli indicator prior for uniform number of changepoints
  aux_prob <- c(max_cp:1)/c((max_cp+1):2)

  # Model specification ----
  mod_type <- if(n_class==1) ('PREM') else
    if(!is.null(class_predictive_vars) && !is.null(outcome_predictive_vars)) ('CI_PREMM_Full') else
      if(!is.null(class_predictive_vars) && is.null(outcome_predictive_vars)) ('CI_PREMM_Class_Predictive') else
        if(is.null(class_predictive_vars) && !is.null(outcome_predictive_vars)) ('CI_PREMM_Outcome_Predictive') else 'PREMM'

  # END OF SETUP ----

  if(mod_type=='PREM'){

    # PREM ----

    ## GENERATING INITIAL VALUES ----

    ## Specify fixed JAGS model
    if(cp_prior=='binomial' && n_cov_outcome_predictive==0) init_fixed_spec <- textConnection(model_prem_binomial_fixed)
    if(cp_prior=='binomial' && n_cov_outcome_predictive>0) init_fixed_spec <- textConnection(model_prem_cov_binomial_fixed)
    if(cp_prior=='uniform'  && n_cov_outcome_predictive==0) init_fixed_spec <- textConnection(model_prem_uniform_fixed)
    if(cp_prior=='uniform'  && n_cov_outcome_predictive>0) init_fixed_spec <- textConnection(model_prem_cov_uniform_fixed)

    ## Variables to extract for initialization of full model
    param_recovery_init <- c('beta', 'cp', 'n_cp', 'cp_indicator', 'sigma2_error')
    if(n_cov_outcome_predictive>0) param_recovery_init <- c(param_recovery_init, 'outcome_predictive_covariate_alpha')

    ## Create data list for JAGS model
    data_list <- list('x'=x, 'y'=y,
                      'n_subj'=n_subj, 'n_time'=n_time,
                      'max_cp'=max_cp, 'max_beta'=max_beta,
                      'mean'=mean, 'prec'=prec,
                      'min_cp_mean'=min_cp_mean, 'max_cp_mean'=max_cp_mean,
                      'n_class'=n_class)
    if(cp_prior=='binomial') data_list$binom_prob <- binom_prob
    if(cp_prior=='uniform') data_list$aux_prob <- aux_prob
    if(n_cov_outcome_predictive>0) data_list$n_cov_outcome_predictive <- n_cov_outcome_predictive
    if(n_cov_outcome_predictive>0) data_list$outcome_predictive_covariates <- outcome_predictive_covariates

    ## Set Seeds for Reproducible Code
    initial_vals <- vector('list', n_chains)
    seeds <- sample(1:10, n_chains, replace = FALSE) # randomly generate non-repeating integers
    # sensitive to set.seed()
    for(i in 1:n_chains){
      initial_vals[[i]]$.RNG.name <- "base::Wichmann-Hill" # arbitrary
      initial_vals[[i]]$.RNG.seed <- seeds[i]
    }

    ## Create JAGS model
    if(verbose) cat('Computing initial values...\n')
    init_fixed_model <- rjags::jags.model(init_fixed_spec,
                                          data=data_list,
                                          inits = initial_vals,
                                          n.chains=n_chains,
                                          n.adapt=500,
                                          quiet=TRUE)
    ## burn-in
    update(init_fixed_model, 2000, progress.bar=progress_bar)

    ## sampling
    init_fixed_out <- rjags::jags.samples(init_fixed_model,
                                          variable.names=param_recovery_init,
                                          n.iter=2000,
                                          thin=4,
                                          progress.bar=progress_bar)

    ### Compiling Initialization Results ----

    ## Permute changepoint labels, if necessary, so they are ordered correctly
    init_fixed_out <- Permute_CP(output=init_fixed_out,
                                 n_class=n_class,
                                 max_cp=max_cp,
                                 max_beta=max_beta)

    ## Extract initial values for full model parameters
    initial_vals <- vector('list', n_chains)
    for(i in 1:n_chains){initial_vals[[i]]$beta_mean <- matrix(nrow=n_class, ncol=max_beta)}
    for(i in 1:n_chains){initial_vals[[i]]$beta_sd <- matrix(nrow=n_class, ncol=max_beta)}
    for(i in 1:n_chains){initial_vals[[i]]$cp_mean <- matrix(nrow=n_class, ncol=max_cp)}
    for(i in 1:n_chains){initial_vals[[i]]$cp_sd <- matrix(nrow=n_class, ncol=max_cp)}

    for(i in 1:n_chains){
      for(j in 1:n_class){
        initial_vals[[i]]$beta_mean[j,] <- rowMeans(init_fixed_out$beta[j,,init_fixed_out$n_cp[j,,i]==getmode(init_fixed_out$n_cp[j,,i]),i])
        if(length(initial_vals[[i]]$cp_mean)<2){
          initial_vals[[i]]$cp_mean[j,] <- mean(init_fixed_out$cp[j,,init_fixed_out$n_cp[j,,i]==getmode(init_fixed_out$n_cp[j,,i]),i])
        }
        if(length(initial_vals[[i]]$cp_mean)>1){
          initial_vals[[i]]$cp_mean[j,] <- rowMeans(init_fixed_out$cp[j,,init_fixed_out$n_cp[j,,i]==getmode(init_fixed_out$n_cp[j,,i]),i])
        }
        for(k in 1:max_cp){
          initial_vals[[i]]$cp_sd[j,k] <- (max_time-min_time)/(4*max_cp)
        }
        initial_vals[[i]]$beta_sd[j,1] <- sqrt(1/prec_param_int)
        for(k in 2:max_beta){
          initial_vals[[i]]$beta_sd[j,k] <- sqrt(2/prec_param)
        }
      }
    }

    ## only when cp_prior=='binomial'
    if(cp_prior=='binomial'){
      for(i in 1:n_chains){initial_vals[[i]]$cp_indicator <- matrix(nrow=n_class, ncol=max_cp)}
      for(i in 1:n_chains){
        for(j in 1:n_class){
          for(k in 1:max_cp){
            initial_vals[[i]]$cp_indicator[j,k] <- getmode(init_fixed_out$cp_indicator[j,k,,i])
          }
        }
      }
    }

    ## only when cp_prior=='uniform'
    if(cp_prior=='uniform'){
      for(i in 1:n_chains){initial_vals[[i]]$Temp <- matrix(nrow=n_class, ncol=max_cp)}
      for(i in 1:n_chains){
        for(j in 1:n_class){
          for(k in 1:max_cp){
            initial_vals[[i]]$Temp[j,k] <- getmode(init_fixed_out$cp_indicator[j,k,,i])
          }
        }
      }
    }

    ## only when n_cov_outcome_predictive>0
    if(n_cov_outcome_predictive>0){
      for(i in 1:n_chains){
        if(n_cov_outcome_predictive==1){
          initial_vals[[i]]$outcome_predictive_covariate_alpha <- mean(init_fixed_out$outcome_predictive_covariate_alpha[,,i])
        }
        if(n_cov_outcome_predictive>1){
          initial_vals[[i]]$outcome_predictive_covariate_alpha <- rowMeans(init_fixed_out$outcome_predictive_covariate_alpha[,,i])
        }
      }
    }

    ## FULL MODEL -----

    ## Specify full JAGS model
    # without outcome predictive covariates
    if(n_cov_outcome_predictive==0 && cp_prior=='binomial' && scale_prior=='uniform') full_spec <- textConnection(model_prem_binomial_scaleunif)
    if(n_cov_outcome_predictive==0 && cp_prior=='binomial' && scale_prior=='hc') full_spec <- textConnection(model_prem_binomial_scalehc)
    if(n_cov_outcome_predictive==0 && cp_prior=='uniform' && scale_prior=='uniform') full_spec <- textConnection(model_prem_uniform_scaleunif)
    if(n_cov_outcome_predictive==0 && cp_prior=='uniform' && scale_prior=='hc') full_spec <- textConnection(model_prem_uniform_scalehc)
    # with outcome predictive covariates
    if(n_cov_outcome_predictive>0 && cp_prior=='binomial' && scale_prior=='uniform') full_spec <- textConnection(model_prem_cov_binomial_scaleunif)
    if(n_cov_outcome_predictive>0 && cp_prior=='binomial' && scale_prior=='hc') full_spec <- textConnection(model_prem_cov_binomial_scalehc)
    if(n_cov_outcome_predictive>0 && cp_prior=='uniform' && scale_prior=='uniform') full_spec <- textConnection(model_prem_cov_uniform_scaleunif)
    if(n_cov_outcome_predictive>0 && cp_prior=='uniform' && scale_prior=='hc') full_spec <- textConnection(model_prem_cov_uniform_scalehc)

    ## Variables to extract from full model
    param_recovery_full <- c('beta', 'beta_mean', 'beta_sd',
                             'cp', 'cp_mean', 'cp_sd',
                             'n_cp', 'cp_indicator',
                             'mu_y', 'sigma2_error',
                             'for_conv','pD','deviance')
    if(n_cov_outcome_predictive>0) param_recovery_full <- c(param_recovery_full, 'outcome_predictive_covariate_alpha')

    ## Create data list for JAGS model
    data_list <- list('x'=x, 'y'=y,
                      'n_subj'=n_subj, 'n_time'=n_time,
                      'min_time'=min_time, 'max_time'=max_time,
                      'max_cp'=max_cp, 'max_beta'=max_beta,
                      'mean'=mean, 'prec'=prec,
                      'min_cp_mean'=min_cp_mean, 'max_cp_mean'=max_cp_mean,
                      'n_class'=n_class,
                      'n_params'=n_params)
    if(cp_prior=='binomial') data_list$binom_prob <- binom_prob
    if(cp_prior=='uniform') data_list$aux_prob <- aux_prob
    if(n_cov_outcome_predictive>0) data_list$n_cov_outcome_predictive <- n_cov_outcome_predictive
    if(n_cov_outcome_predictive>0) data_list$outcome_predictive_covariates <- outcome_predictive_covariates

    ## Set Seeds for Reproducible Code
    seeds <- sample(1:10, n_chains, replace = FALSE) # randomly generate non-repeating integers
    # sensitive to set.seed()
    for(i in 1:n_chains){
      initial_vals[[i]]$.RNG.name <- "base::Wichmann-Hill" # arbitrary
      initial_vals[[i]]$.RNG.seed <- seeds[i]
    }

    ## Create JAGS model
    if(verbose) cat('Calibrating MCMC...\n')
    full_model <- rjags::jags.model(full_spec,
                                    data=data_list,
                                    inits=initial_vals,
                                    n.chains=n_chains,
                                    n.adapt=iters_adapt,
                                    quiet=TRUE)

    # burn-in
    if(verbose) cat('Running burn-in...\n')
    update(full_model, iters_burn_in, progress.bar=progress_bar)

    # sampling
    if(verbose) cat('Collecting samples...\n')
    full_out <- rjags::jags.samples(full_model,
                                    variable.names=param_recovery_full,
                                    n.iter=iters_sampling,
                                    thin=thin,
                                    progress.bar=progress_bar)

    ## Permute changepoint labels, if necessary, so they are ordered correctly
    full_out <- Permute_CP(output=full_out,
                           n_class=n_class,
                           max_cp=max_cp,
                           max_beta=max_beta)
    #####

  } else
    if(mod_type=='PREMM'){

      # PREMM ----

      ## GENERATING INITIAL VALUES ----

      ## Specify fixed JAGS model
      if(cp_prior=='binomial') init_fixed_spec <- textConnection(model_premm_binomial_fixed)
      if(cp_prior=='uniform')  init_fixed_spec <- textConnection(model_premm_uniform_fixed)

      ## Variables to extract for initialization of full model
      param_recovery_init <- c('beta', 'cp', 'n_cp', 'cp_indicator', 'sigma2_error', 'class')

      ## Create data list for JAGS model
      data_list <- list('x'=x, 'y'=y,
                        'n_subj'=n_subj, 'n_time'=n_time,
                        'max_cp'=max_cp, 'max_beta'=max_beta,
                        'mean'=mean, 'prec'=prec,
                        'min_cp_mean'=min_cp_mean, 'max_cp_mean'=max_cp_mean,
                        'n_class'=n_class, 'alpha'=rep(alpha, n_class))
      if(cp_prior=='binomial') data_list$binom_prob <- binom_prob
      if(cp_prior=='uniform') data_list$aux_prob <- aux_prob

      ## Set Seeds for Reproducible Code
      initial_vals <- vector('list', n_chains)
      seeds <- sample(1:10, n_chains, replace = FALSE) # randomly generate non-repeating integers
      # sensitive to set.seed()
      for(i in 1:n_chains){
        initial_vals[[i]]$.RNG.name <- "base::Wichmann-Hill" # arbitrary
        initial_vals[[i]]$.RNG.seed <- seeds[i]
      }

      ## Create JAGS model
      if(verbose) cat('Computing initial values...\n')
      init_fixed_model <- rjags::jags.model(init_fixed_spec,
                                            data=data_list,
                                            inits = initial_vals,
                                            n.chains=n_chains,
                                            n.adapt=500,
                                            quiet=TRUE)
      ## burn-in
      update(init_fixed_model, 2000, progress.bar=progress_bar)

      ## sampling
      init_fixed_out <- rjags::jags.samples(init_fixed_model,
                                            variable.names=param_recovery_init,
                                            n.iter=2000,
                                            thin=4,
                                            progress.bar=progress_bar)

      ### Compiling Initialization Results ----

      ## Permute changepoint labels, if necessary, so they are ordered correctly
      init_fixed_out <- Permute_CP(output=init_fixed_out,
                                   n_class=n_class,
                                   max_cp=max_cp,
                                   max_beta=max_beta)

      ## Realign Classes - checking if the classes are properly ordered
      init_fixed_out <- Realign_ECR(output=init_fixed_out,
                                    n_class=n_class,
                                    model=mod_type)

      ## Extract initial values for full model parameters
      initial_vals <- vector('list', n_chains)
      for(i in 1:n_chains){initial_vals[[i]]$beta_mean <- matrix(nrow=n_class, ncol=max_beta)}
      for(i in 1:n_chains){initial_vals[[i]]$beta_sd <- matrix(nrow=n_class, ncol=max_beta)}
      for(i in 1:n_chains){initial_vals[[i]]$cp_mean <- matrix(nrow=n_class, ncol=max_cp)}
      for(i in 1:n_chains){initial_vals[[i]]$cp_sd <- matrix(nrow=n_class, ncol=max_cp)}
      for(i in 1:n_chains){initial_vals[[i]]$class <- NA}

      for(i in 1:n_chains){
        initial_vals[[i]]$class <- apply(init_fixed_out$class[,,i],1,getmode)
        for(j in 1:n_class){
          initial_vals[[i]]$beta_mean[j,] <- rowMeans(init_fixed_out$beta[j,,init_fixed_out$n_cp[j,,i]==getmode(init_fixed_out$n_cp[j,,i]),i])
          if(length(initial_vals[[i]]$cp_mean)<2){
            initial_vals[[i]]$cp_mean[j,] <- mean(init_fixed_out$cp[j,,init_fixed_out$n_cp[j,,i]==getmode(init_fixed_out$n_cp[j,,i]),i])
          }
          if(length(initial_vals[[i]]$cp_mean)>1){
            initial_vals[[i]]$cp_mean[j,] <- rowMeans(init_fixed_out$cp[j,,init_fixed_out$n_cp[j,,i]==getmode(init_fixed_out$n_cp[j,,i]),i])
          }
          for(k in 1:max_cp){
            initial_vals[[i]]$cp_sd[j,k] <- (max_time-min_time)/(4*max_cp)
          }
          initial_vals[[i]]$beta_sd[j,1] <- sqrt(1/prec_param_int)
          for(k in 2:max_beta){
            initial_vals[[i]]$beta_sd[j,k] <- sqrt(2/prec_param)
          }
        }
      }

      ## only when cp_prior=='binomial'
      if(cp_prior=='binomial'){
        for(i in 1:n_chains){initial_vals[[i]]$cp_indicator <- matrix(nrow=n_class, ncol=max_cp)}
        for(i in 1:n_chains){
          for(j in 1:n_class){
            for(k in 1:max_cp){
              initial_vals[[i]]$cp_indicator[j,k] <- getmode(init_fixed_out$cp_indicator[j,k,,i])
            }
          }
        }
      }

      ## only when cp_prior=='uniform'
      if(cp_prior=='uniform'){
        for(i in 1:n_chains){initial_vals[[i]]$Temp <- matrix(nrow=n_class, ncol=max_cp)}
        for(i in 1:n_chains){
          for(j in 1:n_class){
            for(k in 1:max_cp){
              initial_vals[[i]]$Temp[j,k] <- getmode(init_fixed_out$cp_indicator[j,k,,i])
            }
          }
        }
      }

      ## FULL MODEL -----

      ## Specify full JAGS model
      if(cp_prior=='binomial' && scale_prior=='uniform') full_spec <- textConnection(model_premm_binomial_scaleunif)
      if(cp_prior=='binomial' && scale_prior=='hc') full_spec <- textConnection(model_premm_binomial_scalehc)
      if(cp_prior=='uniform' && scale_prior=='uniform') full_spec <- textConnection(model_premm_uniform_scaleunif)
      if(cp_prior=='uniform' && scale_prior=='hc') full_spec <- textConnection(model_premm_uniform_scalehc)

      ## Variables to extract from full model
      param_recovery_full <- c('beta', 'beta_mean', 'beta_sd',
                               'cp', 'cp_mean', 'cp_sd',
                               'n_cp', 'cp_indicator',
                               'class_prob', 'class',
                               'mu_y', 'sigma2_error',
                               'for_conv','pD','deviance')

      ## Create data list for JAGS model
      data_list <- list('x'=x, 'y'=y,
                        'n_subj'=n_subj, 'n_time'=n_time,
                        'min_time'=min_time, 'max_time'=max_time,
                        'max_cp'=max_cp, 'max_beta'=max_beta,
                        'mean'=mean, 'prec'=prec,
                        'min_cp_mean'=min_cp_mean, 'max_cp_mean'=max_cp_mean,
                        'n_class'=n_class, 'alpha'=rep(alpha, n_class),
                        'n_params'=n_params)
      if(cp_prior=='binomial') data_list$binom_prob <- binom_prob
      if(cp_prior=='uniform') data_list$aux_prob <- aux_prob

      ## Set Seeds for Reproducible Code
      seeds <- sample(1:10, n_chains, replace = FALSE) # randomly generate non-repeating integers
      # sensitive to set.seed()
      for(i in 1:n_chains){
        initial_vals[[i]]$.RNG.name <- "base::Wichmann-Hill" # arbitrary
        initial_vals[[i]]$.RNG.seed <- seeds[i]
      }

      ## Create JAGS model
      if(verbose) cat('Calibrating MCMC...\n')
      full_model <- rjags::jags.model(full_spec,
                                      data=data_list,
                                      inits=initial_vals,
                                      n.chains=n_chains,
                                      n.adapt=iters_adapt,
                                      quiet=TRUE)

      # burn-in
      if(verbose) cat('Running burn-in...\n')
      update(full_model, iters_burn_in, progress.bar=progress_bar)

      # sampling
      if(verbose) cat('Collecting samples...\n')
      full_out <- rjags::jags.samples(full_model,
                                      variable.names=param_recovery_full,
                                      n.iter=iters_sampling,
                                      thin=thin,
                                      progress.bar=progress_bar)

      ## Permute changepoint labels, if necessary, so they are ordered correctly
      full_out <- Permute_CP(output=full_out,
                             n_class=n_class,
                             max_cp=max_cp,
                             max_beta=max_beta)

      ## Realign Classes - checking if the classes are properly ordered
      full_out <- Realign_ECR(output=full_out,
                              n_class=n_class,
                              model=mod_type)
      #####

    } else
    {

      # CI-PREMM ----

      ## GENERATE INITIAL VALUES ----

      ## Specify fixed JAGS model
      if(mod_type=='CI_PREMM_Full' && cp_prior=='binomial') init_fixed_spec <- textConnection(model_cipremm_full_binomial_fixed)
      if(mod_type=='CI_PREMM_Full' && cp_prior=='uniform') init_fixed_spec <- textConnection(model_cipremm_full_uniform_fixed)
      if(mod_type=='CI_PREMM_Class_Predictive' && cp_prior=='binomial') init_fixed_spec <- textConnection(model_cipremm_cpo_binomial_fixed)
      if(mod_type=='CI_PREMM_Class_Predictive' && cp_prior=='uniform') init_fixed_spec <- textConnection(model_cipremm_cpo_uniform_fixed)
      if(mod_type=='CI_PREMM_Outcome_Predictive' && cp_prior=='binomial') init_fixed_spec <- textConnection(model_cipremm_opo_binomial_fixed)
      if(mod_type=='CI_PREMM_Outcome_Predictive' && cp_prior=='uniform') init_fixed_spec <- textConnection(model_cipremm_opo_uniform_fixed)

      ## Variables to extract for initialization of full model
      param_recovery_init <- c('beta', 'cp', 'n_cp', 'cp_indicator', 'sigma2_error', 'class')
      if(mod_type=='CI_PREMM_Full') param_recovery_init <- c(param_recovery_init, c('logistic_intercept', 'class_predictive_covariate_lambda', 'outcome_predictive_covariate_alpha'))
      if(mod_type=='CI_PREMM_Class_Predictive') param_recovery_init <- c(param_recovery_init, c('logistic_intercept', 'class_predictive_covariate_lambda'))
      if(mod_type=='CI_PREMM_Outcome_Predictive') param_recovery_init <- c(param_recovery_init, 'outcome_predictive_covariate_alpha')

      ## Create data list for JAGS model
      data_list <- list('x'=x, 'y'=y,
                        'n_subj'=n_subj, 'n_time'=n_time,
                        'max_cp'=max_cp, 'max_beta'=max_beta,
                        'mean'=mean, 'prec'=prec,
                        'min_cp_mean'=min_cp_mean, 'max_cp_mean'=max_cp_mean,
                        'n_class'=n_class)
      if(cp_prior=='binomial') data_list$binom_prob <- binom_prob
      if(cp_prior=='uniform') data_list$aux_prob <- aux_prob
      if(mod_type=='CI_PREMM_Outcome_Predictive') data_list$alpha <- rep(alpha, n_class)
      if(mod_type=='CI_PREMM_Full' | mod_type=='CI_PREMM_Class_Predictive') data_list$class_predictive_covariates <- class_predictive_covariates
      if(mod_type=='CI_PREMM_Full' | mod_type=='CI_PREMM_Class_Predictive') data_list$n_cov_class_predictive <- n_cov_class_predictive
      if(mod_type=='CI_PREMM_Full' | mod_type=='CI_PREMM_Outcome_Predictive') data_list$outcome_predictive_covariates <- outcome_predictive_covariates
      if(mod_type=='CI_PREMM_Full' | mod_type=='CI_PREMM_Outcome_Predictive') data_list$n_cov_outcome_predictive <- n_cov_outcome_predictive

      ## Set Seeds for Reproducible Code
      initial_vals <- vector('list', n_chains)
      seeds <- sample(1:10, n_chains, replace = FALSE) # randomly generate non-repeating integers
      # sensitive to set.seed()
      for(i in 1:n_chains){
        initial_vals[[i]]$.RNG.name <- "base::Wichmann-Hill" # arbitrary
        initial_vals[[i]]$.RNG.seed <- seeds[i]
      }

      ## Create JAGS model
      if(verbose) cat('Computing initial values...\n')
      init_fixed_model <- rjags::jags.model(init_fixed_spec,
                                            data=data_list,
                                            inits = initial_vals,
                                            n.chains=n_chains,
                                            n.adapt=500,
                                            quiet=TRUE)
      ## burn-in
      update(init_fixed_model, 2000, progress.bar=progress_bar)

      ## sampling
      init_fixed_out <- rjags::jags.samples(init_fixed_model,
                                            variable.names=param_recovery_init,
                                            n.iter=2000,
                                            thin=4,
                                            progress.bar=progress_bar)

      ### Compiling Initialization Results ----

      ## Permute changepoint labels, if necessary, so they are ordered correctly
      init_fixed_out <- Permute_CP(output=init_fixed_out,
                                   n_class=n_class,
                                   max_cp=max_cp,
                                   max_beta=max_beta)

      ## Realign Classes - checking if the classes are properly ordered
      init_fixed_out <- Realign_ECR(output=init_fixed_out,
                                    n_class=n_class,
                                    model=mod_type)

      ## Extract initial values for full model parameters
      initial_vals <- vector('list', n_chains)
      for(i in 1:n_chains){initial_vals[[i]]$beta_mean <- matrix(nrow=n_class, ncol=max_beta)}
      for(i in 1:n_chains){initial_vals[[i]]$beta_sd <- matrix(nrow=n_class, ncol=max_beta)}
      for(i in 1:n_chains){initial_vals[[i]]$cp_mean <- matrix(nrow=n_class, ncol=max_cp)}
      for(i in 1:n_chains){initial_vals[[i]]$cp_sd <- matrix(nrow=n_class, ncol=max_cp)}

      for(i in 1:n_chains){
        for(j in 1:n_class){
          initial_vals[[i]]$beta_mean[j,] <- rowMeans(init_fixed_out$beta[j,,init_fixed_out$n_cp[j,,i]==getmode(init_fixed_out$n_cp[j,,i]),i])
          if(length(initial_vals[[i]]$cp_mean)<2){
            initial_vals[[i]]$cp_mean[j,] <- mean(init_fixed_out$cp[j,,init_fixed_out$n_cp[j,,i]==getmode(init_fixed_out$n_cp[j,,i]),i])
          }
          if(length(initial_vals[[i]]$cp_mean)>1){
            initial_vals[[i]]$cp_mean[j,] <- rowMeans(init_fixed_out$cp[j,,init_fixed_out$n_cp[j,,i]==getmode(init_fixed_out$n_cp[j,,i]),i])
          }
          for(k in 1:max_cp){
            initial_vals[[i]]$cp_sd[j,k] <- (max_time-min_time)/(4*max_cp)
          }
          initial_vals[[i]]$beta_sd[j,1] <- sqrt(1/prec_param_int)
          for(k in 2:max_beta){
            initial_vals[[i]]$beta_sd[j,k] <- sqrt(2/prec_param)
          }
        }
      }

      ## only for CI_PREMM_Outcome_Predictive
      if(mod_type=='CI_PREMM_Outcome_Predictive'){
        for(i in 1:n_chains){
          initial_vals[[i]]$class <- apply(init_fixed_out$class[,,i],1,getmode)
        }
      }

      ## only when cp_prior=='binomial'
      if(cp_prior=='binomial'){
        for(i in 1:n_chains){initial_vals[[i]]$cp_indicator <- matrix(nrow=n_class, ncol=max_cp)}
        for(i in 1:n_chains){
          for(j in 1:n_class){
            for(k in 1:max_cp){
              initial_vals[[i]]$cp_indicator[j,k] <- getmode(init_fixed_out$cp_indicator[j,k,,i])
            }
          }
        }
      }

      ## only when cp_prior=='uniform'
      if(cp_prior=='uniform'){
        for(i in 1:n_chains){initial_vals[[i]]$Temp <- matrix(nrow=n_class, ncol=max_cp)}
        for(i in 1:n_chains){
          for(j in 1:n_class){
            for(k in 1:max_cp){
              initial_vals[[i]]$Temp[j,k] <- getmode(init_fixed_out$cp_indicator[j,k,,i])
            }
          }
        }
      }

      ## only when n_cov_outcome_predictive>0
      if(n_cov_outcome_predictive>0){
        for(i in 1:n_chains){
          if(n_cov_outcome_predictive==1){
            initial_vals[[i]]$outcome_predictive_covariate_alpha <- mean(init_fixed_out$outcome_predictive_covariate_alpha[,,i])
          }
          if(n_cov_outcome_predictive>1){
            initial_vals[[i]]$outcome_predictive_covariate_alpha <- rowMeans(init_fixed_out$outcome_predictive_covariate_alpha[,,i])
          }
        }
      }

      ## only when n_cov_class_predictive>0
      if(n_cov_class_predictive>0){
        for(i in 1:n_chains){
          if(n_cov_class_predictive==1){
            initial_vals[[i]]$class_predictive_covariate_lambda <- mean(init_fixed_out$class_predictive_covariate_lambda[,,i])
          }
          if(n_cov_class_predictive>1){
            initial_vals[[i]]$class_predictive_covariate_lambda <- rowMeans(init_fixed_out$class_predictive_covariate_lambda[,,i])
          }
          initial_vals[[i]]$logistic_intercept <- mean(init_fixed_out$logistic_intercept[,,i])
        }
      }

      ## FULL MODEL -----

      ## Specify full JAGS model
      if(mod_type=='CI_PREMM_Full') model_list <- cipremm_mods_full
      if(mod_type=='CI_PREMM_Class_Predictive') model_list <- cipremm_mods_cpo
      if(mod_type=='CI_PREMM_Outcome_Predictive') model_list <- cipremm_mods_opo

      if(cp_prior=='binomial' && scale_prior=='uniform') full_spec <- textConnection(model_list[[1]])
      if(cp_prior=='binomial' && scale_prior=='hc') full_spec <- textConnection(model_list[[2]])
      if(cp_prior=='uniform' && scale_prior=='uniform') full_spec <- textConnection(model_list[[3]])
      if(cp_prior=='uniform' && scale_prior=='hc') full_spec <- textConnection(model_list[[4]])

      ## Variables to extract from full model
      param_recovery_full <- c('beta', 'beta_mean', 'beta_sd',
                               'cp', 'cp_mean', 'cp_sd',
                               'n_cp', 'cp_indicator',
                               'class',
                               'mu_y', 'sigma2_error',
                               'for_conv','pD','deviance')
      if(mod_type=='CI_PREMM_Full') param_recovery_full <- c(param_recovery_full, c('logistic_intercept', 'class_predictive_covariate_lambda', 'outcome_predictive_covariate_alpha', 'logistic_class_prob'))
      if(mod_type=='CI_PREMM_Class_Predictive') param_recovery_full <- c(param_recovery_full, c('logistic_intercept', 'class_predictive_covariate_lambda', 'logistic_class_prob'))
      if(mod_type=='CI_PREMM_Outcome_Predictive') param_recovery_full <- c(param_recovery_full, 'outcome_predictive_covariate_alpha', 'class_prob')

      ## Create data list for JAGS model
      data_list <- list('x'=x, 'y'=y,
                        'n_subj'=n_subj, 'n_time'=n_time,
                        'min_time'=min_time, 'max_time'=max_time,
                        'max_cp'=max_cp, 'max_beta'=max_beta,
                        'mean'=mean, 'prec'=prec,
                        'min_cp_mean'=min_cp_mean, 'max_cp_mean'=max_cp_mean,
                        'n_class'=n_class,
                        'n_params'=n_params)
      if(cp_prior=='binomial') data_list$binom_prob <- binom_prob
      if(cp_prior=='uniform') data_list$aux_prob <- aux_prob
      if(mod_type=='CI_PREMM_Outcome_Predictive') data_list$alpha <- rep(alpha, n_class)
      if(mod_type=='CI_PREMM_Full' | mod_type=='CI_PREMM_Class_Predictive') data_list$class_predictive_covariates <- class_predictive_covariates
      if(mod_type=='CI_PREMM_Full' | mod_type=='CI_PREMM_Class_Predictive') data_list$n_cov_class_predictive <- n_cov_class_predictive
      if(mod_type=='CI_PREMM_Full' | mod_type=='CI_PREMM_Outcome_Predictive') data_list$outcome_predictive_covariates <- outcome_predictive_covariates
      if(mod_type=='CI_PREMM_Full' | mod_type=='CI_PREMM_Outcome_Predictive') data_list$n_cov_outcome_predictive <- n_cov_outcome_predictive

      ## Set Seeds for Reproducible Code
      seeds <- sample(1:10, n_chains, replace = FALSE) # randomly generate non-repeating integers
      # sensitive to set.seed()
      for(i in 1:n_chains){
        initial_vals[[i]]$.RNG.name <- "base::Wichmann-Hill" # arbitrary
        initial_vals[[i]]$.RNG.seed <- seeds[i]
      }

      ## Create JAGS model
      if(verbose) cat('Calibrating MCMC...\n')
      full_model <- rjags::jags.model(full_spec,
                                      data=data_list,
                                      inits=initial_vals,
                                      n.chains=n_chains,
                                      n.adapt=iters_adapt,
                                      quiet=TRUE)

      # burn-in
      if(verbose) cat('Running burn-in...\n')
      update(full_model, iters_burn_in, progress.bar=progress_bar)

      # sampling
      if(verbose) cat('Collecting samples...\n')
      full_out <- rjags::jags.samples(full_model,
                                      variable.names=param_recovery_full,
                                      n.iter=iters_sampling,
                                      thin=thin,
                                      progress.bar=progress_bar)

      ## Permute changepoint labels, if necessary, so they are ordered correctly
      full_out <- Permute_CP(output=full_out,
                             n_class=n_class,
                             max_cp=max_cp,
                             max_beta=max_beta)

      ## Realign Classes - checking if the classes are properly ordered
      full_out <- Realign_ECR(output=full_out,
                              n_class=n_class,
                              model=mod_type)
    }

  # Compiling Full Results -----

  ## Call Info
  call_info <- as.list(match.call())
  def_args <- formals(sys.function())
  for(arg in names(def_args)){
    if(!(arg %in% names(call_info))){
      call_info[[arg]] <- def_args[[arg]]
    }
  }

  ## Convergence
  # growth parameters (class dependent)
  for(c in 1:n_class){
    full_out$for_conv[1:max_beta+(4+4*max_cp)*(c-1),,] <- full_out$beta_mean[c,,,]
    full_out$for_conv[(1+max_beta):(2*max_beta)+(4+4*max_cp)*(c-1),,] <- full_out$beta_sd[c,,,]
    full_out$for_conv[(1+2*max_beta):(2*max_beta+max_cp)+(4+4*max_cp)*(c-1),,] <- full_out$cp_mean[c,,,]
    full_out$for_conv[(1+2*max_beta+max_cp):(2*max_beta+2*max_cp)+(4+4*max_cp)*(c-1),,] <- full_out$cp_sd[c,,,]
  }
  # covariates (if applicable)
  if(n_cov_outcome_predictive>0){
    full_out$for_conv[(1+2*max_beta+2*max_cp):(2*max_beta+2*max_cp+n_cov_outcome_predictive)+(4+4*max_cp)*(n_class-1),,] <- full_out$outcome_predictive_covariate_alpha[,,]
  }
  if(n_cov_class_predictive>0){
    full_out$for_conv[(1+2*max_beta+2*max_cp+n_cov_outcome_predictive):(2*max_beta+2*max_cp+n_cov_outcome_predictive+n_cov_class_predictive)+(4+4*max_cp)*(n_class-1),,] <- full_out$class_predictive_covariate_lambda[,,]
  }
  if(n_cov_class_predictive>0){
    full_out$for_conv[(1+2*max_beta+2*max_cp+n_cov_outcome_predictive+n_cov_class_predictive)+(4+4*max_cp)*(n_class-1),,] <- full_out$logistic_intercept[,,]
  }
  # error variance
  full_out$for_conv[n_params,,] <- full_out$sigma2_error

  mcmc_list <- coda::as.mcmc.list(full_out$for_conv)
  gelman_msrf <- coda::gelman.diag(mcmc_list)

  # Initialize the parameter vector
  param_names <- c('NA')
  for(c in 1:n_class){
    param_names <- c(param_names,
                     paste0('Class ', c, ': beta_', 0:(max_beta-1),'_mean'),
                     paste0('Class ', c, ': beta_', 0:(max_beta-1),'_sd'),
                     paste0('Class ', c, ': cp_', 1:max_cp,'_mean'),
                     paste0('Class ', c, ': cp_', 1:max_cp,'_sd'))
  }
  param_names <- c(param_names[-1],
                   if(n_cov_outcome_predictive>0) paste0('outcome_predictive_covariate_alpha_', 1:n_cov_outcome_predictive),
                   if(n_cov_class_predictive>0) paste0('class_predictive_covariate_lambda_', 1:n_cov_class_predictive),
                   if(n_cov_class_predictive>0) 'logistic_intercept',
                   'error_var')
  row.names(gelman_msrf$psrf) <- param_names

  parameter_psrf <- data.frame(point_est = gelman_msrf$psrf[,1],
                               upper_ci = gelman_msrf$psrf[,2])
  multivariate_psrf <- gelman_msrf$mpsrf
  mean_psrf <- mean(parameter_psrf$point_est) # mean psrf across parameters
  convergence <- list(parameter_psrf=parameter_psrf, multivariate_psrf=multivariate_psrf, mean_psrf=mean_psrf)

  ## Model Fit
  deviance <- mean(full_out$deviance)
  pD <- mean(full_out$pD)
  dic <- deviance + pD
  model_fit <- list(deviance=deviance, pD=pD, dic=dic)

  ## Fitted Values
  y_mean <- summary(full_out$mu_y, FUN='mean')[[1]]
  fit_vals <- cbind(data[,id_var], data[,time_var], data[,y_var], as.vector(t(y_mean)))
  fit_vals <- as.data.frame(fit_vals)
  names(fit_vals) <- c(id_var, time_var, y_var, "fitted")

  ## Random Coefficients
  rancoef <- cbind(unique(data[,id_var]),
                   summary(full_out$beta, FUN='mean')[[1]],
                   summary(full_out$cp, FUN='mean')[[1]])
  rancoef <- as.data.frame(rancoef)
  names(rancoef) <- c(id_var,
                      paste0("beta_", 0:(max_beta-1)),
                      paste0("cp_", 1:max_cp))
  ran_ef_coef <- list("rancoef" = rancoef)

  ## Class Information
  # each of these is fixed to 1 for PREM
  class_membership <- rep(1, n_subj)
  names(class_membership) <- unique(data[,id_var])
  individ_class_probability <- matrix(1, nrow=n_subj, ncol=n_class) # conditional on logistic model (when applicable) and growth curve
  rownames(individ_class_probability) <- unique(data[,id_var])
  unconditional_class_probability <- 1
  names(unconditional_class_probability) <- "Class 1"
  if(n_class>1){
    for(i in 1:n_subj){class_membership[i] <- as.numeric(names(which.max(table(full_out$class[i,,]))))}
    for(i in 1:n_subj){individ_class_probability[i,1:n_class] <- prop.table(table(factor(full_out$class[i,,], levels = 1:n_class)))}
    if(n_cov_class_predictive==0){
      unconditional_class_probability <- apply(full_out$class_prob, c(1), FUN='mean') # not based on growth curve part of the model
      names(unconditional_class_probability) <- paste0("Class ", 1:n_class)
    }
    if(n_cov_class_predictive>0){
      unconditional_class_probability <- apply(full_out$logistic_class_prob, c(1), FUN='mean')
      names(unconditional_class_probability) <- unique(data[,id_var])
    }
  }
  class_info <- list('class_membership'=class_membership,
                     'individ_class_probability'=individ_class_probability,
                     'unconditional_class_probability'=unconditional_class_probability)

  ## Parameter Estimates
  param_est <- list()
  for(c in 1:n_class){

    # Probability for number of changepoints
    K_prob <- matrix(nrow=(max_cp+1), ncol=1)
    for(k in 0:max_cp){
      K_prob[k+1] <- sum(full_out$n_cp[c,,]==k)/length(full_out$n_cp[c,,])
    }
    colnames(K_prob) <- 'Probability'
    rownames(K_prob) <- c(paste0('K=',0:max_cp))

    # Parameter estimates for each result for the number of changepoints
    K=list()
    for(k in 0:max_cp){
      K[[k+1]] <- list()
      if(K_prob[k+1]>0.01){
        beta_array <- array(full_out$beta, dim=c(dim(full_out$beta)[1],dim(full_out$beta)[2],dim(full_out$beta)[3]*dim(full_out$beta)[4]))
        beta_array <- beta_array[,,full_out$n_cp[c,,]==k]
        beta <- apply(beta_array, c(1,2), 'mean')[,1:(2+k)]
        beta_mean_array <- array(full_out$beta_mean[c,,,], dim=c(dim(full_out$beta_mean)[2],dim(full_out$beta_mean)[3]*dim(full_out$beta_mean)[4]))
        beta_mean_array <- beta_mean_array[1:(2+k),full_out$n_cp[c,,]==k]
        b_mean <- apply(beta_mean_array, c(1), 'mean')
        b_mean_CI <- apply(beta_mean_array,c(1), 'quantile', probs=c(0.025,0.975))
        beta_sd_array <- array(full_out$beta_sd[c,,,], dim=c(dim(full_out$beta_sd)[2],dim(full_out$beta_sd)[3]*dim(full_out$beta_sd)[4]))
        beta_sd_array <- beta_sd_array[1:(2+k),full_out$n_cp[c,,]==k]
        beta_sd <- apply(beta_sd_array, c(1), 'mean')
        beta_sd_CI <- apply(beta_sd_array, c(1), 'quantile', probs=c(0.025,0.975))
        cp <- cp_mean <- cp_mean_CI <- cp_sd <- cp_sd_CI <- NULL
        if(k>0){
          cp_array <- array(full_out$cp, dim=c(dim(full_out$cp)[1],dim(full_out$cp)[2],dim(full_out$cp)[3]*dim(full_out$cp)[4]))
          cp_array <- cp_array[,,full_out$n_cp[c,,]==k]
          cp <- apply(cp_array,c(1,2),'mean')[,1:k]
          cp_mean_array <- array(full_out$cp_mean[c,,,], dim=c(dim(full_out$cp_mean)[2],dim(full_out$cp_mean)[3]*dim(full_out$cp_mean)[4]))
          cp_mean_array <- cp_mean_array[1:k,full_out$n_cp[c,,]==k,drop=FALSE]
          cp_mean <- apply(cp_mean_array,c(1),'mean')
          cp_mean_CI <- apply(cp_mean_array,c(1),'quantile',probs=c(0.025,0.975))
          cp_sd_array <- array(full_out$cp_sd[c,,,], dim=c(dim(full_out$cp_sd)[2],dim(full_out$cp_sd)[3]*dim(full_out$cp_sd)[4]))
          cp_sd_array <- cp_sd_array[1:k,full_out$n_cp[c,,]==k,drop=FALSE]
          cp_sd <- apply(cp_sd_array,c(1),'mean')
          cp_sd_CI <- apply(cp_sd_array,c(1),'quantile',probs=c(0.025,0.975))
        }
        K[[k+1]] <- list('beta'=beta,
                         'beta_mean'=b_mean, 'beta_mean_CI'=b_mean_CI,
                         'beta_var'=beta_sd^2, 'beta_var_CI'=beta_sd_CI^2, # provide variance estimates as output
                         'cp'=cp,
                         'cp_mean'=cp_mean, 'cp_mean_CI'=cp_mean_CI,
                         'cp_var'=cp_sd^2,'cp_var_CI'=cp_sd_CI^2) # provide variance estimates as output
      }
      names(K)[k+1] <- paste0('K_', k)
      param_est[[c]] <- list('K_prob'=K_prob,'K'=K)
    }
  }
  names(param_est) <- paste0('Class_', 1:n_class)

  param_est$error_var <- apply(full_out$sigma2_error, c(1), 'mean')
  param_est$error_var_CI <- apply(full_out$sigma2_error, c(1), 'quantile', probs=c(0.025,0.975))

  if(n_cov_outcome_predictive>0){
    param_est$outcome_predictive_covariates <- apply(full_out$outcome_predictive_covariate_alpha, c(1), 'mean')
    param_est$outcome_predictive_covariates_CI <- apply(full_out$outcome_predictive_covariate_alpha, c(1), 'quantile', probs=c(0.025,0.975))
    names(param_est$outcome_predictive_covariates) <- outcome_predictive_vars
  }

  if(n_cov_class_predictive>0){
    param_est$class_predictive_covariates <- apply(full_out$class_predictive_covariate_lambda, c(1), 'mean')
    param_est$class_predictive_covariates_CI <- apply(full_out$class_predictive_covariate_lambda, c(1), 'quantile', probs=c(0.025,0.975))
    names(param_est$class_predictive_covariates) <- class_predictive_vars

    param_est$logistic_intercept <- apply(full_out$logistic_intercept, c(1), 'mean')
    param_est$logistic_intercept_CI <- apply(full_out$logistic_intercept, c(1), 'quantile', probs=c(0.025,0.975))
  }

  ## Stop tracking run time
  run_time_total_end <- Sys.time()
  run_time_total <- run_time_total_end - run_time_total_start

  my_results <- list('Call' = call_info,
                     'Sample_Size' = list(n_subj=n_subj, n_time=n_time),
                     'Data' = data,
                     'Convergence' = convergence,
                     'Model_Fit' = model_fit,
                     'Fitted_Values' = fit_vals,
                     'Class_Information' = class_info,
                     'Parameter_Estimates' = param_est,
                     'Random_Coefficients' = ran_ef_coef,
                     'Run_Time' = format(run_time_total))
  if(save_full_chains==TRUE){my_results$Full_MCMC_Chains=full_out}
  if(save_conv_chains==TRUE){my_results$Convergence_MCMC_Chains=mcmc_list}
  class(my_results) <- c('PREM', 'BEND')
  return(my_results)
}

#' Summarize the results of a bivariate piecewise random effects model (BPREM)
#'
#' @description
#' Provides a summary of a BPREM model, as returned by `Bayes_BPREM()`.
#'
#' @param object An object of class "BPREM" (returned by `Bayes_BPREM(...)`).
#' @param ... Additional arguments.
#'
#' @returns Returns a list of key parameter estimates.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_bprem)
#' # result summary
#' summary(results_bprem)
#'
#' @export
summary.BPREM <- function(object, ...){

  x <- object

  # determine number of parameters
  n_param <- 4

  # determine number of covariances
  n_covar <- ((n_param*2)*((n_param*2)-1))/2

  # define parameters for labeling
  param_names <- c("Intercept", "Slope", "Change in Slope", "Changepoint")
  names(param_names) <- c("0", "1", "2", "g")

  # FIXED EFFECTS -----
  fix_eff_est <- x$Parameter_Estimates[1:(n_param*2),]
  names(fix_eff_est) <- c("Estimate", "Lower CI", "Upper CI")
  rownames(fix_eff_est) <- paste0(rep(c("Outcome 1: ", "Outcome 2: "), e=4), rep(param_names,2), " Mean")

  # RANDOM EFFECTS -----

  ## COVARIANCE MATRIX -----
  ran_eff_var_mat <- x$Parameter_Estimates[((n_param*2)+(1:(n_param*2))),]
  names(ran_eff_var_mat) <- c("Estimate", "Lower CI", "Upper CI")
  rownames(ran_eff_var_mat) <- paste0(rep(c("Outcome 1: ", "Outcome 2: "), e=4), rep(param_names,2), " Var")

  ran_eff_cov_mat <- x$Parameter_Estimates[(2*(n_param*2)+(1:n_covar)),]
  names(ran_eff_cov_mat) <- c("Estimate", "Lower CI", "Upper CI")
  # get rownames
  ss <- sub("cp", "g", gsub("cov_b_", "", rownames(ran_eff_cov_mat))) ## replace cp with g to make following code simpler
  rownames(ran_eff_cov_mat) <- paste0("Cov(Outcome ", substr(ss,1,1), ": ", param_names[substr(ss,2,2)],
                                      ", Outcome ", substr(ss,4,4), ": ", param_names[substr(ss,5,5)], ")")

  ## CORRELATION MATRIX -----
  ran_eff_corr_mat <- x$Parameter_Estimates[(2*(n_param*2)+n_covar+3+(1:n_covar)),]
  names(ran_eff_corr_mat) <- c("Estimate", "Lower CI", "Upper CI")
  # get rownames
  ss <- sub("cp", "g", gsub("corr_b_", "", rownames(ran_eff_corr_mat))) ## replace cp with g to make following code simpler
  rownames(ran_eff_corr_mat) <- paste0("Corr(Outcome ", substr(ss,1,1), ": ", param_names[substr(ss,2,2)],
                                      ", Outcome ", substr(ss,4,4), ": ", param_names[substr(ss,5,5)], ")")

  # ERROR -----
  error_cov_mat <- x$Parameter_Estimates[(2*(n_param*2)+n_covar+(1:3)),]
  names(error_cov_mat) <- c("Estimate", "Lower CI", "Upper CI")
  rownames(error_cov_mat) <- c("Outcome 1: Error Var", "Outcome 2: Error Var", "Cov(Outcome 1: Error, Outcome 2: Error)")

  error_cor_mat <- x$Parameter_Estimates[(2*(n_param*2)+2*n_covar+3+1),]
  names(error_cor_mat) <- c("Estimate", "Lower CI", "Upper CI")
  rownames(error_cor_mat) <- c("Corr(Outcome 1: Error, Outcome 2: Error)")

  ## return value
  out <- list("data" = x$Call$data,
              "y1_var" = x$Call$y1_var,
              "y2_var" = x$Call$y2_var,
              "fix_eff_est" = fix_eff_est,
              "ran_eff_var_mat" = ran_eff_var_mat,
              "ran_eff_cov_mat" = ran_eff_cov_mat,
              "ran_eff_corr_mat" = ran_eff_corr_mat,
              "error_cov_mat" = error_cov_mat,
              "error_corr" = error_cor_mat,
              "msrf" = x$Convergence$multivariate_psrf,
              "mean_psrf" = x$Convergence$mean_psrf,
              "DIC" = x$Model_Fit$dic)
  class(out) <- c("summary.BPREM", class(out))
  return(out)
}

#' @rdname summary.BPREM
#' @param x An object of class "summary.BPREM".
#' @export
print.summary.BPREM <- function(x, ...){

  cat("Bayesian bivariate piecewise random effects model\n")
  cat("Data:", x$data, "\n")
  cat("Outcomes:", c(x$y1_var, x$y2_var), "\n")
  cat("\n")
  cat("Fixed Effect Parameters:\n")
  print(round(x$fix_eff_est,3))
  cat("\n")
  cat("Random Effect Parameters:\n")
  cat("Variances:\n")
  print(round(x$ran_eff_var_mat,3), na.print="")
  cat("\n")
  cat("Covariances:\n")
  print(round(x$ran_eff_cov_mat,3), na.print="")
  cat("\n")
  cat("Correlations:\n")
  print(round(x$ran_eff_corr_mat,3), na.print="")
  cat("\n")
  cat("Error:\n")
  cat("Variance-Covariance:\n")
  print(round(x$error_cov_mat,3), na.print="")
  cat("\n")
  cat("Correlation:\n")
  print(round(x$error_corr,3), na.print="")
  cat("\n")
  cat("Gelman's msrf:", round(x$msrf, 3), "\n")
  cat("Mean psrf:", round(x$mean_psrf, 3), "\n")
  cat("DIC:", round(x$DIC,3))
  cat("\n")

  invisible(x)
}

# Bayes_PREM JAGS Models

# PREM (One class Models) ----

## Binomial CP ----

### Fixed ----

model_prem_binomial_fixed <- "model{
  for(i in 1:n_subj){
  class[i] <- 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[class[i],1] + beta[class[i],2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[class[i],3:max_beta]*x_cp[class[i],1:max_cp,i,j])
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    cp_indicator[c,k] ~ dbern(binom_prob)
    cp[c,k] ~ dunif(min_cp_mean, max_cp_mean)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[c,k])
      }
    }
  }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])

  # prior distribution of the fixed parameters
  beta[c,1:max_beta] ~ dmnorm(mean, prec)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y
  }
"

### Scale Unif ----

model_prem_binomial_scaleunif <- "model{
  for(i in 1:n_subj){
  class[i] <- 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    cp_indicator[c,k] ~ dbern(binom_prob)

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dunif(0, (max_time-min_time)/4)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dunif(0, (2/prec[q,q])^0.5)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

### Scale HC ----

model_prem_binomial_scalehc <- "model{
  for(i in 1:n_subj){
  class[i] <- 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    cp_indicator[c,k] ~ dbern(binom_prob)

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dt(0, 1/(((max_time-min_time)/4)/tan(0.45*3.1416))^2, 1) T(0,)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dt(0, 1/(((2/prec[q,q])^0.5)/tan(0.45*3.1416))^2, 1) T(0,)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

## Uniform CP ----

### Fixed ----

model_prem_uniform_fixed <- "model{
  for(i in 1:n_subj){
  class[i] <- 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[class[i],1] + beta[class[i],2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[class[i],3:max_beta]*x_cp[class[i],1:max_cp,i,j])
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    Temp[c,k] ~ dbern(aux_prob[k])
    cp_indicator[c,k] <- prod(Temp[c,1:k])
    # prior distribution of fixed effect of changepont
    cp[c,k] ~ dunif(min_cp_mean, max_cp_mean)
      for(j in 1:n_time){
        for(i in 1:n_subj){
       x_cp[c,k,i,j] <- max(0, x[i,j]-cp[c,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])

  # prior distribution of the fixed parameters
  beta[c,1:max_beta] ~ dmnorm(mean, prec)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y
  }
"

### Scale Unif ----

model_prem_uniform_scaleunif <- "model{
  for(i in 1:n_subj){
  class[i] <- 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    Temp[c,k] ~ dbern(aux_prob[k])
    cp_indicator[c,k] <- prod(Temp[c,1:k])

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dunif(0, (max_time-min_time)/4)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dunif(0, (2/prec[q,q])^0.5)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

### Scale HC ----

model_prem_uniform_scalehc <- "model{
  for(i in 1:n_subj){
  class[i] <- 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    Temp[c,k] ~ dbern(aux_prob[k])
    cp_indicator[c,k] <- prod(Temp[c,1:k])

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dt(0, 1/(((max_time-min_time)/4)/tan(0.45*3.1416))^2, 1) T(0,)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dt(0, 1/(((2/prec[q,q])^0.5)/tan(0.45*3.1416))^2, 1) T(0,)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

## Binomial CP (Covariates) ----

### Fixed ----

model_prem_cov_binomial_fixed <- "model{
  for(i in 1:n_subj){
  class[i] <- 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[class[i],1] + beta[class[i],2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[class[i],3:max_beta]*x_cp[class[i],1:max_cp,i,j]) +
                 inprod(outcome_predictive_covariate_alpha[], outcome_predictive_covariates[i,j,])
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    cp_indicator[c,k] ~ dbern(binom_prob)
    # prior distribution of fixed effect of changepont
    cp[c,k] ~ dunif(min_cp_mean, max_cp_mean)
      for(j in 1:n_time){
        for(i in 1:n_subj){
       x_cp[c,k,i,j] <- max(0, x[i,j]-cp[c,k])
      }
    }
  }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])

  # prior distribution of the fixed parameters
  beta[c,1:max_beta] ~ dmnorm(mean, prec)
  }

  # prior distribution of the outcome predictive covariates
  for(l in 1:n_cov_outcome_predictive){
  outcome_predictive_covariate_alpha[l] ~ dnorm(0, 0.001)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y
  }
"

### Scale Unif ----

model_prem_cov_binomial_scaleunif <- "model{
  for(i in 1:n_subj){
  class[i] <- 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j]) +
                 inprod(outcome_predictive_covariate_alpha[], outcome_predictive_covariates[i,j,])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    cp_indicator[c,k] ~ dbern(binom_prob)

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dunif(0, (max_time-min_time)/4)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dunif(0, (2/prec[q,q])^0.5)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the outcome predictive covariates
  for(l in 1:n_cov_outcome_predictive){
  outcome_predictive_covariate_alpha[l] ~ dnorm(0, 0.001)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

### Scale HC ----

model_prem_cov_binomial_scalehc <- "model{
  for(i in 1:n_subj){
  class[i] <- 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j]) +
                 inprod(outcome_predictive_covariate_alpha[], outcome_predictive_covariates[i,j,])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    cp_indicator[c,k] ~ dbern(binom_prob)

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dt(0, 1/(((max_time-min_time)/4)/tan(0.45*3.1416))^2, 1) T(0,)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dt(0, 1/(((2/prec[q,q])^0.5)/tan(0.45*3.1416))^2, 1) T(0,)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the outcome predictive covariates
  for(l in 1:n_cov_outcome_predictive){
  outcome_predictive_covariate_alpha[l] ~ dnorm(0, 0.001)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"


## Uniform CP (Covariates) ----

### Fixed ----

model_prem_cov_uniform_fixed <- "model{
  for(i in 1:n_subj){
  class[i] <- 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[class[i],1] + beta[class[i],2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[class[i],3:max_beta]*x_cp[class[i],1:max_cp,i,j]) +
                 inprod(outcome_predictive_covariate_alpha[], outcome_predictive_covariates[i,j,])
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    Temp[c,k] ~ dbern(aux_prob[k])
    cp_indicator[c,k] <- prod(Temp[c,1:k])
    # prior distribution of fixed effect of changepont
    cp[c,k] ~ dunif(min_cp_mean, max_cp_mean)
      for(j in 1:n_time){
        for(i in 1:n_subj){
       x_cp[c,k,i,j] <- max(0, x[i,j]-cp[c,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])

  # prior distribution of the fixed parameters
  beta[c,1:max_beta] ~ dmnorm(mean, prec)
  }

  # prior distribution of the outcome predictive covariates
  for(l in 1:n_cov_outcome_predictive){
  outcome_predictive_covariate_alpha[l] ~ dnorm(0, 0.001)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y
  }
"

### Scale Unif ----

model_prem_cov_uniform_scaleunif <- "model{
  for(i in 1:n_subj){
  class[i] <- 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j]) +
                 inprod(outcome_predictive_covariate_alpha[], outcome_predictive_covariates[i,j,])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    Temp[c,k] ~ dbern(aux_prob[k])
    cp_indicator[c,k] <- prod(Temp[c,1:k])

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dunif(0, (max_time-min_time)/4)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dunif(0, (2/prec[q,q])^0.5)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the outcome predictive covariates
  for(l in 1:n_cov_outcome_predictive){
  outcome_predictive_covariate_alpha[l] ~ dnorm(0, 0.001)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

### Scale HC ----

model_prem_cov_uniform_scalehc <- "model{
  for(i in 1:n_subj){
  class[i] <- 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j]) +
                 inprod(outcome_predictive_covariate_alpha[], outcome_predictive_covariates[i,j,])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    Temp[c,k] ~ dbern(aux_prob[k])
    cp_indicator[c,k] <- prod(Temp[c,1:k])

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dt(0, 1/(((max_time-min_time)/4)/tan(0.45*3.1416))^2, 1) T(0,)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dt(0, 1/(((2/prec[q,q])^0.5)/tan(0.45*3.1416))^2, 1) T(0,)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the outcome predictive covariates
  for(l in 1:n_cov_outcome_predictive){
  outcome_predictive_covariate_alpha[l] ~ dnorm(0, 0.001)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

# PREMM (No Covariates) ----

## Binomial CP ----

### Fixed ----

model_premm_binomial_fixed <- "model{
  for(i in 1:n_subj){
  class[i] ~ dcat(class_prob)
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[class[i],1] + beta[class[i],2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[class[i],3:max_beta]*x_cp[class[i],1:max_cp,i,j])
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    cp_indicator[c,k] ~ dbern(binom_prob)
    # prior distribution of fixed effect of changepont
    cp[c,k] ~ dunif(min_cp_mean, max_cp_mean)
      for(j in 1:n_time){
        for(i in 1:n_subj){
       x_cp[c,k,i,j] <- max(0, x[i,j]-cp[c,k])
      }
    }
  }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])

  # prior distribution of the fixed parameters
  beta[c,1:max_beta] ~ dmnorm(mean, prec)
  }

  # prior distribution of the class probability
  class_prob ~ ddirch(alpha)

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y
  }
"

### Scale Unif ----

model_premm_binomial_scaleunif <- "model{
  for(i in 1:n_subj){
  class[i] ~ dcat(class_prob)
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    cp_indicator[c,k] ~ dbern(binom_prob)

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dunif(0, (max_time-min_time)/4)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dunif(0, (2/prec[q,q])^0.5)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the class probability
  class_prob ~ ddirch(alpha)

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

### Scale HC ----

model_premm_binomial_scalehc <- "model{
  for(i in 1:n_subj){
  class[i] ~ dcat(class_prob)
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    cp_indicator[c,k] ~ dbern(binom_prob)

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dt(0, 1/(((max_time-min_time)/4)/tan(0.45*3.1416))^2, 1) T(0,)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dt(0, 1/(((2/prec[q,q])^0.5)/tan(0.45*3.1416))^2, 1) T(0,)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the class probability
  class_prob ~ ddirch(alpha)

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

## Uniform CP ----

### Fixed ----

model_premm_uniform_fixed <- "model{
  for(i in 1:n_subj){
  class[i] ~ dcat(class_prob)
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[class[i],1] + beta[class[i],2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[class[i],3:max_beta]*x_cp[class[i],1:max_cp,i,j])
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    Temp[c,k] ~ dbern(aux_prob[k])
    cp_indicator[c,k] <- prod(Temp[c,1:k])
    # prior distribution of fixed effect of changepont
    cp[c,k] ~ dunif(min_cp_mean, max_cp_mean)
      for(j in 1:n_time){
        for(i in 1:n_subj){
       x_cp[c,k,i,j] <- max(0, x[i,j]-cp[c,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])

  # prior distribution of the fixed parameters
  beta[c,1:max_beta] ~ dmnorm(mean, prec)
  }

  # prior distribution of the class probability
  class_prob ~ ddirch(alpha)

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y
 }
"

### Scale Unif ----

model_premm_uniform_scaleunif <- "model{
  for(i in 1:n_subj){
  class[i] ~ dcat(class_prob)
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    Temp[c,k] ~ dbern(aux_prob[k])
    cp_indicator[c,k] <- prod(Temp[c,1:k])

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dunif(0, (max_time-min_time)/4)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dunif(0, (2/prec[q,q])^0.5)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the class probability
  class_prob ~ ddirch(alpha)

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

### Scale HC ----

model_premm_uniform_scalehc <- "model{
  for(i in 1:n_subj){
  class[i] ~ dcat(class_prob)
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    Temp[c,k] ~ dbern(aux_prob[k])
    cp_indicator[c,k] <- prod(Temp[c,1:k])

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dt(0, 1/(((max_time-min_time)/4)/tan(0.45*3.1416))^2, 1) T(0,)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dt(0, 1/(((2/prec[q,q])^0.5)/tan(0.45*3.1416))^2, 1) T(0,)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the class probability
  class_prob ~ ddirch(alpha)

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

# CI-PREMM ----

## Full ----

### Binomial CP ----

#### Fixed ----

model_cipremm_full_binomial_fixed <- "model{
  for(i in 1:n_subj){
  logit(logistic_class_prob[i]) <- logistic_intercept +
                              sum(class_predictive_covariate_lambda[1:n_cov_class_predictive]*class_predictive_covariates[i,1:n_cov_class_predictive])
  class_bern[i] ~ dbern(logistic_class_prob[i])
  class[i] <- class_bern[i] + 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[class[i],1] + beta[class[i],2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[class[i],3:max_beta]*x_cp[class[i],1:max_cp,i,j]) +
                 inprod(outcome_predictive_covariate_alpha[], outcome_predictive_covariates[i,j,])
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    cp_indicator[c,k] ~ dbern(binom_prob)
    # prior distribution of fixed effect of changepont
    cp[c,k] ~ dunif(min_cp_mean, max_cp_mean)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[c,k])
      }
    }
  }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])

  # prior distribution of the fixed parameters
  beta[c,1:max_beta] ~ dmnorm(mean, prec)
  }

  # prior distribution of the logistic intercept
  logistic_intercept ~ dnorm(0, 0.1)

  # prior distribution of the class predictive covariates
  for(p in 1:n_cov_class_predictive){
  class_predictive_covariate_lambda[p] ~ dnorm(0, 0.1)
  }

  # prior distribution of the outcome predictive covariates
  for(l in 1:n_cov_outcome_predictive){
  outcome_predictive_covariate_alpha[l] ~ dnorm(0, 0.001)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y
  }
"

#### Scale Unif ----

model_cipremm_full_binomial_scaleunif <- "model{
  for(i in 1:n_subj){
  logit(logistic_class_prob[i]) <- logistic_intercept +
                              sum(class_predictive_covariate_lambda[1:n_cov_class_predictive]*class_predictive_covariates[i,1:n_cov_class_predictive])
  class_bern[i] ~ dbern(logistic_class_prob[i])
  class[i] <- class_bern[i] + 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j]) +
                 inprod(outcome_predictive_covariate_alpha[], outcome_predictive_covariates[i,j,])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    cp_indicator[c,k] ~ dbern(binom_prob)

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dunif(0, (max_time-min_time)/4)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dunif(0, (2/prec[q,q])^0.5)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the logistic intercept
  logistic_intercept ~ dnorm(0, 0.1)

  # prior distribution of the class predictive covariates
  for(p in 1:n_cov_class_predictive){
  class_predictive_covariate_lambda[p] ~ dnorm(0, 0.1)
  }

  # prior distribution of the outcome predictive covariates
  for(l in 1:n_cov_outcome_predictive){
  outcome_predictive_covariate_alpha[l] ~ dnorm(0, 0.001)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

#### Scale HC ----

model_cipremm_full_binomial_scalehc <- "model{
  for(i in 1:n_subj){
  logit(logistic_class_prob[i]) <- logistic_intercept +
                              sum(class_predictive_covariate_lambda[1:n_cov_class_predictive]*class_predictive_covariates[i,1:n_cov_class_predictive])
  class_bern[i] ~ dbern(logistic_class_prob[i])
  class[i] <- class_bern[i] + 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j]) +
                 inprod(outcome_predictive_covariate_alpha[], outcome_predictive_covariates[i,j,])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    cp_indicator[c,k] ~ dbern(binom_prob)

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dt(0, 1/(((max_time-min_time)/4)/tan(0.45*3.1416))^2, 1) T(0,)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dt(0, 1/(((2/prec[q,q])^0.5)/tan(0.45*3.1416))^2, 1) T(0,)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the logistic intercept
  logistic_intercept ~ dnorm(0, 0.1)

  # prior distribution of the class predictive covariates
  for(p in 1:n_cov_class_predictive){
  class_predictive_covariate_lambda[p] ~ dnorm(0, 0.1)
  }

  # prior distribution of the outcome predictive covariates
  for(l in 1:n_cov_outcome_predictive){
  outcome_predictive_covariate_alpha[l] ~ dnorm(0, 0.001)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

### Uniform CP ----

#### Fixed ----

model_cipremm_full_uniform_fixed <- "model{
  for(i in 1:n_subj){
  logit(logistic_class_prob[i]) <- logistic_intercept +
                              sum(class_predictive_covariate_lambda[1:n_cov_class_predictive]*class_predictive_covariates[i,1:n_cov_class_predictive])
  class_bern[i] ~ dbern(logistic_class_prob[i])
  class[i] <- class_bern[i] + 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[class[i],1] + beta[class[i],2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[class[i],3:max_beta]*x_cp[class[i],1:max_cp,i,j]) +
                 inprod(outcome_predictive_covariate_alpha[], outcome_predictive_covariates[i,j,])
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    Temp[c,k] ~ dbern(aux_prob[k])
    cp_indicator[c,k] <- prod(Temp[c,1:k])
    # prior distribution of fixed effect of changepont
    cp[c,k] ~ dunif(min_cp_mean, max_cp_mean)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[c,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])

  # prior distribution of the fixed parameters
  beta[c,1:max_beta] ~ dmnorm(mean, prec)
  }

  # prior distribution of the logistic intercept
  logistic_intercept ~ dnorm(0, 0.1)

  # prior distribution of the class predictive covariates
  for(p in 1:n_cov_class_predictive){
  class_predictive_covariate_lambda[p] ~ dnorm(0, 0.1)
  }

  # prior distribution of the outcome predictive covariates
  for(l in 1:n_cov_outcome_predictive){
  outcome_predictive_covariate_alpha[l] ~ dnorm(0, 0.001)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y
  }
"

#### Scale Unif ----

model_cipremm_full_uniform_scaleunif <- "model{
  for(i in 1:n_subj){
  logit(logistic_class_prob[i]) <- logistic_intercept +
                              sum(class_predictive_covariate_lambda[1:n_cov_class_predictive]*class_predictive_covariates[i,1:n_cov_class_predictive])
  class_bern[i] ~ dbern(logistic_class_prob[i])
  class[i] <- class_bern[i] + 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j]) +
                 inprod(outcome_predictive_covariate_alpha[], outcome_predictive_covariates[i,j,])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    Temp[c,k] ~ dbern(aux_prob[k])
    cp_indicator[c,k] <- prod(Temp[c,1:k])

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dunif(0, (max_time-min_time)/4)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dunif(0, (2/prec[q,q])^0.5)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the logistic intercept
  logistic_intercept ~ dnorm(0, 0.1)

  # prior distribution of the class predictive covariates
  for(p in 1:n_cov_class_predictive){
  class_predictive_covariate_lambda[p] ~ dnorm(0, 0.1)
  }

  # prior distribution of the outcome predictive covariates
  for(l in 1:n_cov_outcome_predictive){
  outcome_predictive_covariate_alpha[l] ~ dnorm(0, 0.001)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

#### Scale HC ----

model_cipremm_full_uniform_scalehc <- "model{
  for(i in 1:n_subj){
  logit(logistic_class_prob[i]) <- logistic_intercept +
                              sum(class_predictive_covariate_lambda[1:n_cov_class_predictive]*class_predictive_covariates[i,1:n_cov_class_predictive])
  class_bern[i] ~ dbern(logistic_class_prob[i])
  class[i] <- class_bern[i] + 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j]) +
                 inprod(outcome_predictive_covariate_alpha[], outcome_predictive_covariates[i,j,])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    Temp[c,k] ~ dbern(aux_prob[k])
    cp_indicator[c,k] <- prod(Temp[c,1:k])

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dt(0, 1/(((max_time-min_time)/4)/tan(0.45*3.1416))^2, 1) T(0,)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dt(0, 1/(((2/prec[q,q])^0.5)/tan(0.45*3.1416))^2, 1) T(0,)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the logistic intercept
  logistic_intercept ~ dnorm(0, 0.1)

  # prior distribution of the class predictive covariates
  for(p in 1:n_cov_class_predictive){
  class_predictive_covariate_lambda[p] ~ dnorm(0, 0.1)
  }

  # prior distribution of the outcome predictive covariates
  for(l in 1:n_cov_outcome_predictive){
  outcome_predictive_covariate_alpha[l] ~ dnorm(0, 0.001)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

## Class Predictive - Only (CPO) ----

### Binomial CP ----

#### Fixed ----

model_cipremm_cpo_binomial_fixed <- "model{
  for(i in 1:n_subj){
  logit(logistic_class_prob[i]) <- logistic_intercept +
                              sum(class_predictive_covariate_lambda[1:n_cov_class_predictive]*class_predictive_covariates[i,1:n_cov_class_predictive])
  class_bern[i] ~ dbern(logistic_class_prob[i])
  class[i] <- class_bern[i] + 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[class[i],1] + beta[class[i],2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[class[i],3:max_beta]*x_cp[class[i],1:max_cp,i,j])
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    cp_indicator[c,k] ~ dbern(binom_prob)
    # prior distribution of fixed effect of changepont
    cp[c,k] ~ dunif(min_cp_mean, max_cp_mean)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[c,k])
      }
    }
  }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])

  # prior distribution of the fixed parameters
  beta[c,1:max_beta] ~ dmnorm(mean, prec)
  }

  # prior distribution of the logistic intercept
  logistic_intercept ~ dnorm(0, 0.1)

  # prior distribution of the class predictive covariates
  for(p in 1:n_cov_class_predictive){
  class_predictive_covariate_lambda[p] ~ dnorm(0, 0.1)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y
  }
"

#### Scale Unif ----

model_cipremm_cpo_binomial_scaleunif <- "model{
  for(i in 1:n_subj){
  logit(logistic_class_prob[i]) <- logistic_intercept +
                              sum(class_predictive_covariate_lambda[1:n_cov_class_predictive]*class_predictive_covariates[i,1:n_cov_class_predictive])
  class_bern[i] ~ dbern(logistic_class_prob[i])
  class[i] <- class_bern[i] + 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    cp_indicator[c,k] ~ dbern(binom_prob)

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dunif(0, (max_time-min_time)/4)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dunif(0, (2/prec[q,q])^0.5)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the logistic intercept
  logistic_intercept ~ dnorm(0, 0.1)

  # prior distribution of the class predictive covariates
  for(p in 1:n_cov_class_predictive){
  class_predictive_covariate_lambda[p] ~ dnorm(0, 0.1)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

#### Scale HC ----

model_cipremm_cpo_binomial_scalehc <- "model{
  for(i in 1:n_subj){
    logit(logistic_class_prob[i]) <- logistic_intercept +
      sum(class_predictive_covariate_lambda[1:n_cov_class_predictive]*class_predictive_covariates[i,1:n_cov_class_predictive])
    class_bern[i] ~ dbern(logistic_class_prob[i])
    class[i] <- class_bern[i] + 1
    for(j in 1:n_time){
      y[i,j] ~ dnorm(mu_y[i,j], tau_y)
      mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                   sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
      beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
      cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    cp_indicator[c,k] ~ dbern(binom_prob)

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dt(0, 1/(((max_time-min_time)/4)/tan(0.45*3.1416))^2, 1) T(0,)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dt(0, 1/(((2/prec[q,q])^0.5)/tan(0.45*3.1416))^2, 1) T(0,)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the logistic intercept
  logistic_intercept ~ dnorm(0, 0.1)

  # prior distribution of the class predictive covariates
  for(p in 1:n_cov_class_predictive){
  class_predictive_covariate_lambda[p] ~ dnorm(0, 0.1)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

### Uniform CP ----

#### Fixed ----

model_cipremm_cpo_uniform_fixed <- "model{
  for(i in 1:n_subj){
  logit(logistic_class_prob[i]) <- logistic_intercept +
                              sum(class_predictive_covariate_lambda[1:n_cov_class_predictive]*class_predictive_covariates[i,1:n_cov_class_predictive])
  class_bern[i] ~ dbern(logistic_class_prob[i])
  class[i] <- class_bern[i] + 1
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[class[i],1] + beta[class[i],2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[class[i],3:max_beta]*x_cp[class[i],1:max_cp,i,j])
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    Temp[c,k] ~ dbern(aux_prob[k])
    cp_indicator[c,k] <- prod(Temp[c,1:k])
    # prior distribution of fixed effect of changepont
    cp[c,k] ~ dunif(min_cp_mean, max_cp_mean)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[c,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])

  # prior distribution of the fixed parameters
  beta[c,1:max_beta] ~ dmnorm(mean, prec)
  }

  # prior distribution of the logistic intercept
  logistic_intercept ~ dnorm(0, 0.1)

  # prior distribution of the class predictive covariates
  for(p in 1:n_cov_class_predictive){
  class_predictive_covariate_lambda[p] ~ dnorm(0, 0.1)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y
  }
"

#### Scale Unif ----

model_cipremm_cpo_uniform_scaleunif <- "model{
  for(i in 1:n_subj){
    logit(logistic_class_prob[i]) <- logistic_intercept +
      sum(class_predictive_covariate_lambda[1:n_cov_class_predictive]*class_predictive_covariates[i,1:n_cov_class_predictive])
    class_bern[i] ~ dbern(logistic_class_prob[i])
    class[i] <- class_bern[i] + 1
    for(j in 1:n_time){
      y[i,j] ~ dnorm(mu_y[i,j], tau_y)
      mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                   sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
      beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
      cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    Temp[c,k] ~ dbern(aux_prob[k])
    cp_indicator[c,k] <- prod(Temp[c,1:k])

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dunif(0, (max_time-min_time)/4)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dunif(0, (2/prec[q,q])^0.5)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the logistic intercept
  logistic_intercept ~ dnorm(0, 0.1)

  # prior distribution of the class predictive covariates
  for(p in 1:n_cov_class_predictive){
  class_predictive_covariate_lambda[p] ~ dnorm(0, 0.1)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

#### Scale HC ----

model_cipremm_cpo_uniform_scalehc <- "model{
  for(i in 1:n_subj){
    logit(logistic_class_prob[i]) <- logistic_intercept +
      sum(class_predictive_covariate_lambda[1:n_cov_class_predictive]*class_predictive_covariates[i,1:n_cov_class_predictive])
    class_bern[i] ~ dbern(logistic_class_prob[i])
    class[i] <- class_bern[i] + 1
    for(j in 1:n_time){
      y[i,j] ~ dnorm(mu_y[i,j], tau_y)
      mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
        sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
      beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
      cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    Temp[c,k] ~ dbern(aux_prob[k])
    cp_indicator[c,k] <- prod(Temp[c,1:k])

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dt(0, 1/(((max_time-min_time)/4)/tan(0.45*3.1416))^2, 1) T(0,)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dt(0, 1/(((2/prec[q,q])^0.5)/tan(0.45*3.1416))^2, 1) T(0,)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the logistic intercept
  logistic_intercept ~ dnorm(0, 0.1)

  # prior distribution of the class predictive covariates
  for(p in 1:n_cov_class_predictive){
  class_predictive_covariate_lambda[p] ~ dnorm(0, 0.1)
  }

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

## Outcome Predictive - Only (OPO) ----

### Binomial CP ----

#### Fixed ----

model_cipremm_opo_binomial_fixed <- "model{
  for(i in 1:n_subj){
  class[i] ~ dcat(class_prob)
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[class[i],1] + beta[class[i],2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[class[i],3:max_beta]*x_cp[class[i],1:max_cp,i,j]) +
                 inprod(outcome_predictive_covariate_alpha[], outcome_predictive_covariates[i,j,])
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    cp_indicator[c,k] ~ dbern(binom_prob)
    # prior distribution of fixed effect of changepont
    cp[c,k] ~ dunif(min_cp_mean, max_cp_mean)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[c,k])
      }
    }
  }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])

  # prior distribution of the fixed parameters
  beta[c,1:max_beta] ~ dmnorm(mean, prec)
  }

  # prior distribution of the outcome predictive covariates
  for(l in 1:n_cov_outcome_predictive){
  outcome_predictive_covariate_alpha[l] ~ dnorm(0, 0.001)
  }

  # prior distribution of the class probability
  class_prob ~ ddirch(alpha)

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y
  }
"

#### Scale Unif ----

model_cipremm_opo_binomial_scaleunif <- "model{
  for(i in 1:n_subj){
  class[i] ~ dcat(class_prob)
   for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j]) +
                 inprod(outcome_predictive_covariate_alpha[], outcome_predictive_covariates[i,j,])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    cp_indicator[c,k] ~ dbern(binom_prob)

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dunif(0, (max_time-min_time)/4)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dunif(0, (2/prec[q,q])^0.5)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the outcome predictive covariates
  for(l in 1:n_cov_outcome_predictive){
  outcome_predictive_covariate_alpha[l] ~ dnorm(0, 0.001)
  }

  # prior distribution of the class probability
  class_prob ~ ddirch(alpha)

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

#### Scale HC ----

model_cipremm_opo_binomial_scalehc <- "model{
  for(i in 1:n_subj){
  class[i] ~ dcat(class_prob)
   for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j]) +
                 inprod(outcome_predictive_covariate_alpha[], outcome_predictive_covariates[i,j,])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    cp_indicator[c,k] ~ dbern(binom_prob)

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dt(0, 1/(((max_time-min_time)/4)/tan(0.45*3.1416))^2, 1) T(0,)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dt(0, 1/(((2/prec[q,q])^0.5)/tan(0.45*3.1416))^2, 1) T(0,)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the outcome predictive covariates
  for(l in 1:n_cov_outcome_predictive){
  outcome_predictive_covariate_alpha[l] ~ dnorm(0, 0.001)
  }

  # prior distribution of the class probability
  class_prob ~ ddirch(alpha)

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

### Uniform CP ----

#### Fixed ----

model_cipremm_opo_uniform_fixed <- "model{
  for(i in 1:n_subj){
  class[i] ~ dcat(class_prob)
    for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[class[i],1] + beta[class[i],2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[class[i],3:max_beta]*x_cp[class[i],1:max_cp,i,j]) +
                 inprod(outcome_predictive_covariate_alpha[], outcome_predictive_covariates[i,j,])
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    Temp[c,k] ~ dbern(aux_prob[k])
    cp_indicator[c,k] <- prod(Temp[c,1:k])
    # prior distribution of fixed effect of changepont
    cp[c,k] ~ dunif(min_cp_mean, max_cp_mean)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[c,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])

  # prior distribution of the fixed parameters
  beta[c,1:max_beta] ~ dmnorm(mean, prec)
  }

  # prior distribution of the outcome predictive covariates
  for(l in 1:n_cov_outcome_predictive){
  outcome_predictive_covariate_alpha[l] ~ dnorm(0, 0.001)
  }

  # prior distribution of the class probability
  class_prob ~ ddirch(alpha)

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y
  }
"

#### Scale Unif ----

model_cipremm_opo_uniform_scaleunif <- "model{
  for(i in 1:n_subj){
  class[i] ~ dcat(class_prob)
   for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j]) +
                 inprod(outcome_predictive_covariate_alpha[], outcome_predictive_covariates[i,j,])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    Temp[c,k] ~ dbern(aux_prob[k])
    cp_indicator[c,k] <- prod(Temp[c,1:k])

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dunif(0, (max_time-min_time)/4)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dunif(0, (2/prec[q,q])^0.5)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the outcome predictive covariates
  for(l in 1:n_cov_outcome_predictive){
  outcome_predictive_covariate_alpha[l] ~ dnorm(0, 0.001)
  }

  # prior distribution of the class probability
  class_prob ~ ddirch(alpha)

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

#### Scale HC ----

model_cipremm_opo_uniform_scalehc <- "model{
  for(i in 1:n_subj){
  class[i] ~ dcat(class_prob)
   for(j in 1:n_time){
    y[i,j] ~ dnorm(mu_y[i,j], tau_y)
    mu_y[i,j] <- beta[i,1] + beta[i,2]*x[i,j] +
                 sum(cp_indicator[class[i],1:max_cp]*beta[i,3:max_beta]*x_cp[class[i],1:max_cp,i,j]) +
                 inprod(outcome_predictive_covariate_alpha[], outcome_predictive_covariates[i,j,])
    }
  }

  # prior distribution of the random coefficients
  for(i in 1:n_subj){
    for(q in 1:max_beta){
    beta[i,q] ~ dnorm(beta_mean[class[i],q], beta_tau[class[i],q])
    }
    for(k in 1:max_cp){
    cp[i,k] ~ dnorm(cp_mean[class[i],k], cp_prec[class[i],k]) T(min_time,max_time)
    }
  }

  # prior distribution of the changepoint
  for(c in 1:n_class){
    for(k in 1:max_cp){
    # prior distribution of number of changepoints
    Temp[c,k] ~ dbern(aux_prob[k])
    cp_indicator[c,k] <- prod(Temp[c,1:k])

    cp_mean[c,k] ~ dunif(min_cp_mean, max_cp_mean)
    cp_prec[c,k] <- 1/(cp_sd[c,k]^2)
    # prior distribution of the changepoint variance
    cp_sd[c,k] ~ dt(0, 1/(((max_time-min_time)/4)/tan(0.45*3.1416))^2, 1) T(0,)
      for(j in 1:n_time){
        for(i in 1:n_subj){
        x_cp[c,k,i,j] <- max(0, x[i,j]-cp[i,k])
        }
      }
    }

  n_cp[c] <- sum(cp_indicator[c,1:max_cp])
  beta_mean[c,1:max_beta] ~ dmnorm(mean, prec)

  # prior distribution of random-effects variances
    for(q in 1:max_beta){
    beta_sd[c,q] ~ dt(0, 1/(((2/prec[q,q])^0.5)/tan(0.45*3.1416))^2, 1) T(0,)
    beta_tau[c,q] <- 1/(beta_sd[c,q]^2)
    }
  }

  # prior distribution of the outcome predictive covariates
  for(l in 1:n_cov_outcome_predictive){
  outcome_predictive_covariate_alpha[l] ~ dnorm(0, 0.001)
  }

  # prior distribution of the class probability
  class_prob ~ ddirch(alpha)

  # prior distribution of the precision for y
  tau_y ~ dgamma(0.001, 0.001)
  sigma2_error <- 1/tau_y

  for_conv[1:n_params] <- c(rep(0, n_params))
  }
"

# CI-PREMM Model Lists ----
## Full ----
cipremm_mods_full = list(model_cipremm_full_binomial_scaleunif,
                         model_cipremm_full_binomial_scalehc,
                         model_cipremm_full_uniform_scaleunif,
                         model_cipremm_full_uniform_scalehc)

## CPO ----
cipremm_mods_cpo = list(model_cipremm_cpo_binomial_scaleunif,
                        model_cipremm_cpo_binomial_scalehc,
                        model_cipremm_cpo_uniform_scaleunif,
                        model_cipremm_cpo_uniform_scalehc)

## OPO ----
cipremm_mods_opo = list(model_cipremm_opo_binomial_scaleunif,
                        model_cipremm_opo_binomial_scalehc,
                        model_cipremm_opo_uniform_scaleunif,
                        model_cipremm_opo_uniform_scalehc)

#' Plot a BEND Model (PREM, CREM, BPREM)
#'
#' @description
#' Generates a "spaghetti plot" of observed longitudinal trajectories for each individual. If the results from a `BEND` function are supplied, the trajectory defined by the mean parameters is shown in bold. If fitting a mixture (`PREMM` or `CI-PREMM`) or bivariate model (`BPREM`), the mean trajectories for classes or outcomes will be distinguished by color.
#'
#' @param data Data frame in long format, where each row describes a measurement occasion for a given individual. It is assumed that each individual has the same number of assigned timepoints (a.k.a., rows).
#' @param id_var Name of column that contains ids for individuals with repeated measures in a longitudinal dataset.
#' @param time_var Name of column that contains the time variable.
#' @param y_var Name of column that contains the outcome variable.
#' @param y2_var (for `BPREM` only) Name of column that contains the second outcome variable.
#' @param results The output of `BEND` model to the data. If results=NULL, only a spaghetti plot of the data will be generated.
#' @param xlab X-axis label for the generated plot.
#' @param ylab Y-axis label for the generated plot.
#' @param colors Colors for each class (`PREMM` or `CI-PREMM`) or outcome (`BPREM`). By default, up to 5 colors are provided in the following order: “blue” (class 1 and outcome 1), “red” (class 2 and outcome 2), “green” (class 3), “gold” (class 4), “gray” (class 5).
#' @param mean_colors Colors for the trajectory defined by the mean parameters for each class (`PREMM` or `CI-PREMM`) or outcome (`BPREM`). By default, up to 5 colors are provided in the following order: “darkblue” (class 1 and outcome 1), “darkred” (class 2 and outcome 2), “darkgreen” (class 3), “gold4” (class 4), “darkgray” (class 5).
#' @param legend_pos (optional) Option to change legend position (default = "topright").
#' @param ... (optional) Other parameters to pass to the `plot()` function.
#'
#' @returns No return value.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load simulated data
#' data(SimData_PREM)
#' # plot observed data
#' plot_BEND(data = SimData_PREM,
#'           id_var = "id",
#'           time_var = "time",
#'           y_var = "y")
#' # load fitted model results
#' data(results_prem)
#' # plot fitted results
#' plot_BEND(data = SimData_PREM,
#'           id_var = "id",
#'           time_var = "time",
#'           y_var = "y",
#'           results = results_prem)
#'
#' @import graphics
#'
#' @export
plot_BEND <- function(data,
                      id_var, time_var, y_var,
                      y2_var=NULL,
                      results=NULL,
                      xlab='X', ylab='Y',
                      colors=NULL, mean_colors=NULL,
                      legend_pos="topright", ...){
  # Setup ----

  if(is.null(colors)) colors <- c('blue','red','green','gold','gray')
  if(is.null(mean_colors)) mean_colors = c('darkblue','darkred','darkgreen','gold4','darkgray')

  ## outcome data - matrix form
  y <- reshape(data[,c(id_var, time_var, y_var)],
               idvar=id_var,
               timevar=time_var,
               direction='wide')
  y <- unname(as.matrix(y[,names(y)!=id_var]))

  # For BPREM only
  if(!is.null(y2_var)){
    y2 <- reshape(data[,c(id_var, time_var, y2_var)],
                  idvar=id_var,
                  timevar=time_var,
                  direction='wide')
    y2 <- unname(as.matrix(y2[,names(y2)!=id_var]))
  }

  ## time data - matrix form
  ## should be the same dimensions as y
  x <- matrix(data[,c(time_var)],
              byrow=TRUE,
              nrow=dim(y)[1],
              ncol=dim(y)[2])

  ## Define relevant variables
  n_subj <- nrow(y)
  n_time <- ncol(y)
  xvec <- seq(min(x), max(x), length.out=100)

  # Observed Plot -----
  if(is.null(results)){
    plot(x[1, !is.na(y[1,])],
         y[1, !is.na(y[1,])],
         type = "l",
         ylim = c(min(y,na.rm=TRUE), max(y,na.rm=TRUE)),
         xlim= c(min(x,na.rm=TRUE), max(x,na.rm=TRUE)),
         xlab = xlab,
         ylab = ylab, ...)
    for(i in 2:n_subj){
      lines(x[i, !is.na(y[i,])],
            y[i, !is.na(y[i,])],
            type = "l")
    }

    # BPREM only
    if(!is.null(y2_var)){
      for(i in 2:n_subj){
        lines(x[i, !is.na(y[i,])],
              y[i, !is.na(y[i,])],
              type = "l",
              col=colors[1])
      }
      for(i in 1:n_subj){
        lines(x[i, !is.na(y2[i,])],
              y2[i, !is.na(y2[i,])],
              type = "l",
              col = colors[2])
      }
      legend(legend_pos, lty=1, col=colors[1:2], legend=c("Outcome 1", "Outcome 2"))
    }

    return('Observed trajectories')
  }

  # Fitted Plot -----
  if(!is.null(results)){

    ## PREM -----
    if(inherits(results, "PREM")){

      # determine number of classes
      n_class <- length(unique(results$Class_Information$class_membership))

      # determine number of changepoints in each class (based on final model results)
      changepoints <- c()
      class_data <- data.frame()
      for(i in 1:n_class){
        class_num <- paste0("Class_", i)
        changepoints[i] <- which.max(results$Parameter_Estimates[[class_num]]$K_prob)-1
      }
      max_cp <- max(changepoints)

      # determine who is in each class
      class_list <- list()
      for(i in 1:n_class){
        class_list[[i]] <- c(1:n_subj)[results$Class_Information$class_membership==i]
      }

      # class mean estimates
      class_means <- list()
      for(i in 1:n_class){
        class_num <- paste0("Class_", i)
        cp_num <- paste0("K_", changepoints)[i]
        k <- changepoints[i]

        if(k==0) I <- rep(0, max_cp)
        if(k>0) I <- c(rep(1,k), rep(0,max_cp-k+2))
        # i = k+1
        int <- results$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[1]
        slope1 <- results$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[2]

        slope_cp <- rep(0,max_cp)
        cp <- rep(0,max_cp)
        if(k>0){
          slope_cp[1:k] <- results$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[3:(k+2)]
          cp[1:k] <- results$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_mean[1:k]
        }

        class_means[[i]] <- rep(0,100)
        for(j in 1:100){
          temp <- int + slope1*xvec[j]
          if(k>0){
            for(l in 1:k){
              temp <- temp + slope_cp[l]*(max(0, xvec[j]-cp[l]))}
          }
          class_means[[i]][j] <- temp
        }
      }

      plot(NULL, NULL,
           xlim = c(min(x), max(x)),
           ylim = c(min(y,na.rm=TRUE), max(y,na.rm=TRUE)),
           ylab = ylab,
           xlab = xlab, ...)
      for(i in 1:n_class){
        for(j in class_list[[i]]){
          points(x[j,!is.na(y[j,])],
                 y[j,!is.na(y[j,])],
                 type = "l",
                 col=colors[i])
        }
        points(xvec,
               class_means[[i]],
               type='l',
               col=mean_colors[i],
               lwd=4)
      }
      legend(legend_pos, lty=1, col=colors[1:n_class], legend=paste0("Class ", 1:n_class))
    }

    ## CREM -----
    if(inherits(results, "CREM")){

      # determine form
      form <- results$Functional_Form

      # determine number of parameters
      if(form=="linear")                           n_param <- 2
      if(form=="quadratic" | form=="exponential")  n_param <- 3
      if(form=="piecewise")                        n_param <- 4

      # pull fixed effect estimates
      mean_est <- results$Parameter_Estimates$Mean[1:n_param]

      # define functional form equation

      fit_form_eq <- function(x,est){
        if(form=="linear")       return(est[1] + est[2]*x)
        if(form=="quadratic")    return(est[1] + est[2]*x + est[3]*(x^2))
        if(form=="exponential")  return(est[1] + est[2]*(1-exp(-est[3]*x)))
        if(form=="piecewise")    return(est[1] + est[2]*x + est[3]*(max(0, x-est[4])))
      }

      mean_traj <- rep(0,100)
      for(i in 1:100){
        mean_traj[i] <- fit_form_eq(xvec[i], mean_est)
      }

      plot(x[1, !is.na(y[1,])],
           y[1, !is.na(y[1,])],
           type = "l",
           col = "grey",
           ylim = c(min(y,na.rm=TRUE), max(y,na.rm=TRUE)),
           xlim= c(min(x,na.rm=TRUE), max(x,na.rm=TRUE)),
           xlab = xlab,
           ylab = ylab, ...)
      for(i in 2:n_subj){
        lines(x[i, !is.na(y[i,])],
              y[i, !is.na(y[i,])],
              type = "l",
              col = "grey")
      }
      points(xvec,
             mean_traj,
             type='l',
             col="black",
             lwd=4)
    }

    ## BPREM -----
    if(inherits(results, "BPREM")){

      # pull fixed effect estimates
      mean_est1 <- results$Parameter_Estimates$Mean[1:4]
      mean_est2 <- results$Parameter_Estimates$Mean[5:8]

      # define piecewise function
      pw_eq <- function(x,est){
        return(est[1] + est[2]*x + est[3]*(max(0, x-est[4])))
      }

      mean_traj1 <- rep(0,100)
      mean_traj2 <- rep(0,100)
      for(i in 1:100){
        mean_traj1[i] <- pw_eq(xvec[i], mean_est1)
        mean_traj2[i] <- pw_eq(xvec[i], mean_est2)
      }

      plot(x[1, !is.na(y[1,])],
           y[1, !is.na(y[1,])],
           type = "l",
           col = "grey",
           ylim = c(min(y,na.rm=TRUE), max(y,na.rm=TRUE)),
           xlim= c(min(x,na.rm=TRUE), max(x,na.rm=TRUE)),
           xlab = xlab,
           ylab = ylab, ...)
      for(i in 2:n_subj){
        lines(x[i, !is.na(y[i,])],
              y[i, !is.na(y[i,])],
              type = "l",
              col = colors[1])
      }
      for(i in 1:n_subj){
        lines(x[i, !is.na(y2[i,])],
              y2[i, !is.na(y2[i,])],
              type = "l",
              col = colors[2])
      }
      points(xvec,
             mean_traj1,
             type='l',
             col=mean_colors[1],
             lwd=4)
      points(xvec,
             mean_traj2,
             type='l',
             col=mean_colors[2],
             lwd=4)
      legend(legend_pos, lty=1, col=colors[1:2], legend=c("Outcome 1", "Outcome 2"))
    }

  }
}

#' Print the results of a piecewise random effects model (PREM)
#'
#' @description
#' Provides a summary of a PREM model, as returned by `Bayes_PREM()`.
#'
#' @param x An object of class "PREM" (returned by `Bayes_PREM(...)`).
#' @param ... Additional arguments.
#'
#' @returns Returns a list of key parameter estimates.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_prem)
#' # print results
#' print(results_prem)
#'
#' @export
print.PREM <- function(x, ...){

  # determine number of classes
  n_class <- length(unique(x$Class_Information$class_membership))

  # determine number of changepoints in each class (based on final model results)
  changepoints <- c()
  class_data <- data.frame()
  for(i in 1:n_class){
    class_num <- paste0("Class_", i)
    changepoints[i] <- which.max(x$Parameter_Estimates[[class_num]]$K_prob)-1
  }
  max_cp <- max(changepoints)

  # detect covariates
  n_cov_op <- length(x$Parameter_Estimates$outcome_predictive_covariates)
  n_cov_cp <- length(x$Parameter_Estimates$class_predictive_covariates)

  # CLASS DEPENDENT PARAMETERS -----
  # pull fixed effects for each class
  coeff_mat <- matrix(NA, nrow=2+2*max_cp, ncol=n_class)
  for(i in 1:n_class){
    class_num <- paste0("Class_", i)
    cp_num <- paste0("K_", changepoints)[i]
    # fixed effects
    coeff_mat[1,i] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[1]
    coeff_mat[2,i] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[2]
    # if there are changepoints (beyond linear)
    if(changepoints[i]>0){
      for(k in 1:changepoints[i]){
        # fixed effects
        coeff_mat[2*k+1,i] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_mean[k]
        coeff_mat[2*k+2,i] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[k+2]
      }
    }
  }
  colnames(coeff_mat) <- paste0("Class ", 1:n_class)

  my_rownames <- rep(NA, 2+2*max_cp)
  my_rownames[1] = "Intercept"
  my_rownames[2] = "Slope"
  for(k in 1:max_cp){
    my_rownames[2*k+1] = paste0("Changepoint ", k)
    my_rownames[2*k+2] = paste0("Change in Slope ", k)
  }
  rownames(coeff_mat) <- paste0(my_rownames, " Mean")

  # pull variances for each class
  varcor_list <- vector("list", n_class)
  for(i in 1:n_class){
    class_num <- paste0("Class_", i)
    cp_num <- paste0("K_", changepoints)[i]
    varcor_vec <- c()
    # random effects
    varcor_vec[1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var[1]
    varcor_vec[2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var[2]
    # if there are changepoints (beyond linear)
    if(changepoints[i]>0){
      for(k in 1:changepoints[i]){
        # random effects
        varcor_vec[2*k+1] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_var[k]
        varcor_vec[2*k+2] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_var[k+2]
      }
    }
    n_var <- length(varcor_vec)

    # param_names
    param_names <- rep(NA, n_var)
    param_names[1] = "Intercept"
    param_names[2] = "Slope"
    if(changepoints[i]>0){
      for(k in 1:changepoints[i]){
        param_names[2*k+1] = paste0("Changepoint ", k)
        param_names[2*k+2] = paste0("Change in Slope ", k)
      }
    }
    # format matrix
    varcor_mat <- diag(n_var)
    diag(varcor_mat) <- varcor_vec
    varcor_mat[row(varcor_mat) != col(varcor_mat)] <- NA
    colnames(varcor_mat) <- rownames(varcor_mat) <- param_names

    # save to list
    varcor_list[[i]] <- varcor_mat
  }
  names(varcor_list) <- paste0("class_", 1:n_class)

  # number of changepoints info
  coeff_n_cp <- matrix(changepoints, nrow = 1)
  rownames(coeff_n_cp) <- "Number of Changepoints"
  coeff_mat <- rbind(coeff_n_cp, coeff_mat)

  # class probability info
  class_probs <- matrix(prop.table(table(x$Class_Information$class_membership)), nrow=1)
  rownames(class_probs) <- "Empirical Class Probabilities"
  coeff_mat <- rbind(class_probs, coeff_mat)

  # CLASS INDEPENDENT PARAMETERS -----
  coeff_addit <- data.frame("Estimate" = x$Parameter_Estimates$error_var)
  rownames(coeff_addit) <- "Error Var"

  # Covariates
  if(n_cov_op > 0){
    coeff_cov_op <- data.frame("Estimate" = x$Parameter_Estimates$outcome_predictive_covariates)
    coeff_addit <- rbind(coeff_addit, coeff_cov_op)
  }
  if(n_cov_cp > 0){
    coeff_cov_cp <- data.frame("Estimate" = x$Parameter_Estimates$class_predictive_covariates)
    rownames(coeff_cov_cp) <- paste0(rownames(coeff_cov_cp), " (in log-odds units)")
    coeff_log_int <- data.frame("Estimate" = x$Parameter_Estimates$logistic_intercept)
    rownames(coeff_log_int) <- "Logistic Intercept"
    coeff_addit <- rbind(coeff_addit, coeff_cov_cp, coeff_log_int)
  }

  ## return value
  out <- list("class_dep_params" = coeff_mat,
              "class_dep_varcov" = varcor_list,
              "class_ind_params" = coeff_addit,
              "msrf" = x$Convergence$multivariate_psrf,
              "mean_psrf" = x$Convergence$mean_psrf,
              "DIC" = x$Model_Fit$dic)

  cat("Bayesian piecewise random effects model\n")
  cat("Data:", x$Call$data, "\n")
  cat("Outcome:", x$Call$y_var, "\n")
  cat("Number of Classes:", n_class, "\n")
  cat("Maximum Number of Changepoints:", n_class, "\n")
  if(n_cov_cp==1) cat("Class-Predictive Covariates:", x$Call$class_predictive_vars,"\n")
  if(n_cov_cp>1) cat("Class-Predictive Covariates:", as.character(x$Call$class_predictive_vars)[-1],"\n")
  if(n_cov_op==1) cat("Outcome-Predictive Covariates:", x$Call$outcome_predictive_vars,"\n")
  if(n_cov_op>1) cat("Outcome-Predictive Covariates:", as.character(x$Call$outcome_predictive_vars)[-1],"\n")
  cat("\n")
  cat("Class Dependent Parameters:\n")
  cat("Fixed Effects Parameters:\n")
  print(round(out$class_dep_params,3), na.print="")
  cat("\n")
  cat("Random Effects Parameters:\n")
  for(i in 1:length(out$class_dep_varcov)){
    cat(paste0("Class ",i,":\n"))
    print(round(out$class_dep_varcov[[i]],3), na.print="")
    cat("\n")
  }
  cat("Class Independent Parameters:\n")
  print(round(out$class_ind_params,3))
  cat("\n")
  cat("Gelman's msrf:", round(out$msrf, 3), "\n")
  cat("Mean psrf:", round(out$mean_psrf, 3), "\n")
  cat("DIC:", round(out$DIC,3))
  cat("\n")

  invisible(x)
}

#' Plot the results of a piecewise random effects model (PREM)
#'
#' @description
#' Provides a fitted plot of a PREM model, as returned by `Bayes_PREM()`.
#'
#' @param x An object of class "PREM" (returned by `Bayes_PREM(...)`).
#' @param xlab X-axis label for the generated plot.
#' @param ylab Y-axis label for the generated plot.
#' @param colors Colors for each class (`PREMM` or `CI-PREMM`). By default, up to 5 colors are provided in the following order: "blue" (class 1), "red" (class 2), "green" (class 3), "gold" (class 4), "gray" (class 5).
#' @param mean_colors Colors for the trajectory defined by the mean parameters for each class (`PREMM` or `CI-PREMM`). By default, up to 5 colors are provided in the following order: "darkblue" (class 1), "darkred" (class 2), "darkgreen" (class 3), "gold4" (class 4), "darkgray" (class 5).
#' @param legend_pos (optional) Option to change legend position (default = "topright").
#' @param ... (optional) Other parameters to pass to the `plot()` function.
#'
#' @returns No return value.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_prem)
#' # plot fitted results
#' plot(results_prem)
#'
#' @import graphics
#'
#' @export
plot.PREM <- function(x,
                      xlab='X', ylab='Y',
                      colors=NULL, mean_colors=NULL,
                      legend_pos="topright", ...){
  # Setup ----

  object <- x

  data <- object$Data
  id_var <- object$Call$id_var
  time_var <- object$Call$time_var
  y_var <- object$Call$y_var

  if(is.null(colors)) colors <- c('blue','red','green','gold','gray')
  if(is.null(mean_colors)) mean_colors = c('darkblue','darkred','darkgreen','gold4','darkgray')

  ## outcome data - matrix form
  y <- reshape(data[,c(id_var, time_var, y_var)],
               idvar=id_var,
               timevar=time_var,
               direction='wide')
  y <- unname(as.matrix(y[,names(y)!=id_var]))

  ## time data - matrix form
  ## should be the same dimensions as y
  x <- matrix(data[,c(time_var)],
              byrow=TRUE,
              nrow=dim(y)[1],
              ncol=dim(y)[2])

  ## Define relevant variables
  n_subj <- nrow(y)
  n_time <- ncol(y)
  xvec <- seq(min(x), max(x), length.out=100)

  # Fitted Plot -----

  # determine number of classes
  n_class <- length(unique(object$Class_Information$class_membership))

  # determine number of changepoints in each class (based on final model results)
  changepoints <- c()
  class_data <- data.frame()
  for(i in 1:n_class){
    class_num <- paste0("Class_", i)
    changepoints[i] <- which.max(object$Parameter_Estimates[[class_num]]$K_prob)-1
  }
  max_cp <- max(changepoints)

  # determine who is in each class
  class_list <- list()
  for(i in 1:n_class){
    class_list[[i]] <- c(1:n_subj)[object$Class_Information$class_membership==i]
  }

  # class mean estimates
  class_means <- list()
  for(i in 1:n_class){
    class_num <- paste0("Class_", i)
    cp_num <- paste0("K_", changepoints)[i]
    k <- changepoints[i]

    if(k==0) I <- rep(0, max_cp)
    if(k>0) I <- c(rep(1,k), rep(0,max_cp-k+2))
    # i = k+1
    int <- object$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[1]
    slope1 <- object$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[2]

    slope_cp <- rep(0,max_cp)
    cp <- rep(0,max_cp)
    if(k>0){
      slope_cp[1:k] <- object$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[3:(k+2)]
      cp[1:k] <- object$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_mean[1:k]
    }

    class_means[[i]] <- rep(0,100)
    for(j in 1:100){
      temp <- int + slope1*xvec[j]
      if(k>0){
        for(l in 1:k){
          temp <- temp + slope_cp[l]*(max(0, xvec[j]-cp[l]))}
      }
      class_means[[i]][j] <- temp
    }
  }

  plot(NULL, NULL,
       xlim = c(min(x), max(x)),
       ylim = c(min(y,na.rm=TRUE), max(y,na.rm=TRUE)),
       ylab = ylab,
       xlab = xlab, ...)
  for(i in 1:n_class){
    for(j in class_list[[i]]){
      points(x[j,!is.na(y[j,])],
             y[j,!is.na(y[j,])],
             type = "l",
             col=colors[i])
    }
    points(xvec,
           class_means[[i]],
           type='l',
           col=mean_colors[i],
           lwd=4)
  }
  legend(legend_pos, lty=1, col=colors[1:n_class], legend=paste0("Class ", 1:n_class))

}

#' Plot the results of a bivariate piecewise random effects model (BPREM)
#'
#' @description
#' Provides a fitted plot of a BPREM model, as returned by `Bayes_BPREM()`.
#'
#' @param x An object of class "BPREM" (returned by `Bayes_BPREM(...)`).
#' @param xlab X-axis label for the generated plot.
#' @param ylab Y-axis label for the generated plot.
#' @param colors Colors for each class outcome. By default, up to 2 colors are provided in the following order: "blue" (outcome 1), "red" (outcome 2).
#' @param mean_colors Colors for the trajectory defined by the mean parameters for each outcome. By default, up to 2 colors are provided in the following order: "darkblue" (outcome 1), "darkred" (outcome 2).
#' @param legend_pos (optional) Option to change legend position (default = "topright").
#' @param ... (optional) Other parameters to pass to the `plot()` function.
#'
#' @returns No return value.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_bprem)
#' # plot fitted results
#' plot(results_bprem)
#'
#' @import graphics
#'
#' @export
plot.BPREM <- function(x,
                       xlab='X', ylab='Y',
                       colors=NULL, mean_colors=NULL,
                       legend_pos="topright", ...){
  # Setup ----

  object <- x

  data <- object$Data
  id_var <- object$Call$id_var
  time_var <- object$Call$time_var
  y_var <- object$Call$y1_var
  y2_var <- object$Call$y2_var

  if(is.null(colors)) colors <- c('blue','red')
  if(is.null(mean_colors)) mean_colors = c('darkblue','darkred')

  ## outcome data - matrix form
  y <- reshape(data[,c(id_var, time_var, y_var)],
               idvar=id_var,
               timevar=time_var,
               direction='wide')
  y <- unname(as.matrix(y[,names(y)!=id_var]))

  y2 <- reshape(data[,c(id_var, time_var, y2_var)],
                idvar=id_var,
                timevar=time_var,
                direction='wide')
  y2 <- unname(as.matrix(y2[,names(y2)!=id_var]))

  ## time data - matrix form
  ## should be the same dimensions as y
  x <- matrix(data[,c(time_var)],
              byrow=TRUE,
              nrow=dim(y)[1],
              ncol=dim(y)[2])

  ## Define relevant variables
  n_subj <- nrow(y)
  n_time <- ncol(y)
  xvec <- seq(min(x), max(x), length.out=100)

  # Fitted Plot -----

  # pull fixed effect estimates
  mean_est1 <- object$Parameter_Estimates$Mean[1:4]
  mean_est2 <- object$Parameter_Estimates$Mean[5:8]

  # define piecewise function
  pw_eq <- function(x,est){
    return(est[1] + est[2]*x + est[3]*(max(0, x-est[4])))
  }

  mean_traj1 <- rep(0,100)
  mean_traj2 <- rep(0,100)
  for(i in 1:100){
    mean_traj1[i] <- pw_eq(xvec[i], mean_est1)
    mean_traj2[i] <- pw_eq(xvec[i], mean_est2)
  }

  plot(x[1, !is.na(y[1,])],
       y[1, !is.na(y[1,])],
       type = "l",
       col = "grey",
       ylim = c(min(y,na.rm=TRUE), max(y,na.rm=TRUE)),
       xlim= c(min(x,na.rm=TRUE), max(x,na.rm=TRUE)),
       xlab = xlab,
       ylab = ylab, ...)
  for(i in 2:n_subj){
    lines(x[i, !is.na(y[i,])],
          y[i, !is.na(y[i,])],
          type = "l",
          col = colors[1])
  }
  for(i in 1:n_subj){
    lines(x[i, !is.na(y2[i,])],
          y2[i, !is.na(y2[i,])],
          type = "l",
          col = colors[2])
  }
  points(xvec,
         mean_traj1,
         type='l',
         col=mean_colors[1],
         lwd=4)
  points(xvec,
         mean_traj2,
         type='l',
         col=mean_colors[2],
         lwd=4)
  legend(legend_pos, lty=1, col=colors[1:2], legend=c(y_var, y2_var))

}

#' Plot the results of a crossed random effects model (CREM)
#'
#' @description
#' Provides a fitted plot of a CREM model, as returned by `Bayes_CREM()`.
#'
#' @param x An object of class "CREM" (returned by `Bayes_CREM(...)`).
#' @param xlab X-axis label for the generated plot.
#' @param ylab Y-axis label for the generated plot.
#' @param colors Color for observed trajectories (optional). Default is "grey".
#' @param mean_colors Colors for the trajectory defined by the mean parameters for each outcome (optional). Default is "black".
#' @param legend_pos (optional) Option to change legend position (default = "topright").
#' @param ... (optional) Other parameters to pass to the `plot()` function.
#'
#' @returns No return value.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_pcrem)
#' # plot fitted results
#' plot(results_pcrem)
#'
#' @import graphics
#'
#' @export
plot.CREM <- function(x,
                      xlab='X', ylab='Y',
                      colors=NULL, mean_colors=NULL,
                      legend_pos="topright", ...){
  # Setup ----

  if(is.null(colors)) colors <- c('grey')
  if(is.null(mean_colors)) mean_colors = c('black')

  object <- x

  data <- object$Data
  id_var <- object$Call$ind_id_var
  time_var <- object$Call$time_var
  y_var <- object$Call$y_var

  ## outcome data - matrix form
  y <- reshape(data[,c(id_var, time_var, y_var)],
               idvar=id_var,
               timevar=time_var,
               direction='wide')
  y <- unname(as.matrix(y[,names(y)!=id_var]))

  ## time data - matrix form
  ## should be the same dimensions as y
  x <- matrix(data[,c(time_var)],
              byrow=TRUE,
              nrow=dim(y)[1],
              ncol=dim(y)[2])

  ## Define relevant variables
  n_subj <- nrow(y)
  n_time <- ncol(y)
  xvec <- seq(min(x), max(x), length.out=100)

  # Fitted Plot -----

  # determine form
  form <- object$Functional_Form

  # determine number of parameters
  if(form=="linear")                           n_param <- 2
  if(form=="quadratic" | form=="exponential")  n_param <- 3
  if(form=="piecewise")                        n_param <- 4

  # pull fixed effect estimates
  mean_est <- object$Parameter_Estimates$Mean[1:n_param]

  # define functional form equation

  fit_form_eq <- function(x,est){
    if(form=="linear")       return(est[1] + est[2]*x)
    if(form=="quadratic")    return(est[1] + est[2]*x + est[3]*(x^2))
    if(form=="exponential")  return(est[1] + est[2]*(1-exp(-est[3]*x)))
    if(form=="piecewise")    return(est[1] + est[2]*x + est[3]*(max(0, x-est[4])))
  }

  mean_traj <- rep(0,100)
  for(i in 1:100){
    mean_traj[i] <- fit_form_eq(xvec[i], mean_est)
  }

  plot(x[1, !is.na(y[1,])],
       y[1, !is.na(y[1,])],
       type = "l",
       col = colors,
       ylim = c(min(y,na.rm=TRUE), max(y,na.rm=TRUE)),
       xlim= c(min(x,na.rm=TRUE), max(x,na.rm=TRUE)),
       xlab = xlab,
       ylab = ylab, ...)
  for(i in 2:n_subj){
    lines(x[i, !is.na(y[i,])],
          y[i, !is.na(y[i,])],
          type = "l",
          col = colors)
  }
  points(xvec,
         mean_traj,
         type='l',
         col=mean_colors,
         lwd=4)

}

#' Summarize the results of a crossed random effects model (CREM)
#'
#' @description
#' Provides a summary of a CREM model, as returned by `Bayes_CREM()`.
#'
#' @param object An object of class "CREM" (returned by `Bayes_CREM(...)`).
#' @param ... Additional arguments.
#'
#' @returns Returns a list of key parameter estimates.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_pcrem)
#' # result summary
#' summary(results_pcrem)
#'
#' @export
summary.CREM <- function(object, ...){

  x <- object

  # determine form
  form <- x$Functional_Form

  # determine number of parameters
  if(form=="linear")                           n_param <- 2
  if(form=="quadratic" | form=="exponential")  n_param <- 3
  if(form=="piecewise")                        n_param <- 4

  # determine number of covariances
  n_covar <- (n_param*(n_param-1))/2

  # define parameters for labeling
  if(form=="linear")       param_names <- c("Intercept", "Slope")
  if(form=="linear")       names(param_names) <- c("0", "1")
  if(form=="quadratic")    param_names <- c("Intercept", "Linear Slope", "Quadratic Slope")
  if(form=="quadratic")    names(param_names) <- c("0", "1", "2")
  if(form=="exponential")  param_names <- c("Intercept", "Total Change", "Growth Rate")
  if(form=="exponential")  names(param_names) <- c("0", "1", "2")
  if(form=="piecewise")    param_names <- c("Intercept", "Slope", "Change in Slope", "Changepoint")
  if(form=="piecewise")    names(param_names) <- c("0", "1", "2", "g")

  # FIXED EFFECTS -----
  fix_eff_est <- x$Parameter_Estimates[1:n_param,]
  colnames(fix_eff_est) <- c("Estimate", "Lower CI", "Upper CI")
  rownames(fix_eff_est) <- paste0(param_names, " Mean")

  # RANDOM EFFECTS -----

  ## INDIVIDUALS -----
  ran_eff_b_var_mat <- x$Parameter_Estimates[(n_param+(1:n_param)),]
  colnames(ran_eff_b_var_mat) <- c("Estimate", "Lower CI", "Upper CI")
  rownames(ran_eff_b_var_mat) <- paste0(param_names, " Var")

  ran_eff_b_cov_mat <- x$Parameter_Estimates[(2*n_param+(1:n_covar)),]
  colnames(ran_eff_b_cov_mat) <- c("Estimate", "Lower CI", "Upper CI")
  # get rownames
  ss <- sub("cp", "g", sub("cov_b_", "", rownames(ran_eff_b_cov_mat))) ## replace cp with g to make following code simpler
  rownames(ran_eff_b_cov_mat) <- paste0("Cov(", param_names[substr(ss,1,1)], ", ", param_names[substr(ss,2,2)], ")")

  ## GROUPS -----
  ran_eff_g_var_mat <- x$Parameter_Estimates[(2*n_param+n_covar+(1:n_param)),]
  colnames(ran_eff_g_var_mat) <- c("Estimate", "Lower CI", "Upper CI")
  rownames(ran_eff_g_var_mat) <- paste0(param_names, " Var")

  ran_eff_g_cov_mat <- x$Parameter_Estimates[(3*n_param+n_covar+(1:n_covar)),]
  colnames(ran_eff_g_cov_mat) <- c("Estimate", "Lower CI", "Upper CI")
  # get rownames
  ss <- sub("cp", "g", sub("cov_g_", "", rownames(ran_eff_g_cov_mat))) ## replace cp with g to make following code simpler
  rownames(ran_eff_g_cov_mat) <- paste0("Cov(", param_names[substr(ss,1,1)], ", ", param_names[substr(ss,2,2)], ")")

  ## ERROR -----
  error_mat <- x$Parameter_Estimates[(3*n_param+2*n_covar+1),]
  colnames(error_mat) <- c("Estimate", "Lower CI", "Upper CI")
  rownames(error_mat) <- c("Error Var")

  ## return value
  out <- list("data" = x$Call$data,
              "y_var" = x$Call$y_var,
              "ind_id_var" = x$Call$ind_id_var,
              "cross_id_var" = x$Call$cross_id_var,
              "fix_eff_est" = fix_eff_est,
              "ran_eff_b_var_mat" = ran_eff_b_var_mat,
              "ran_eff_b_cov_mat" = ran_eff_b_cov_mat,
              "ran_eff_g_var_mat" = ran_eff_g_var_mat,
              "ran_eff_g_cov_mat" = ran_eff_g_cov_mat,
              "error_var" = error_mat,
              "msrf" = x$Convergence$multivariate_psrf,
              "mean_psrf" = x$Convergence$mean_psrf,
              "DIC" = x$Model_Fit$dic)
  class(out) <- c("summary.CREM", class(out))
  return(out)
}

#' @rdname summary.CREM
#' @param x An object of class "summary.CREM".
#' @export
print.summary.CREM <- function(x, ...){

  cat("Bayesian crossed random effects model\n")
  cat("Data:", x$data, "\n")
  cat("Outcome:", x$y_var, "\n")
  cat("Individuals:", x$ind_id_var, "\n")
  cat("Group:", x$cross_id_var, "\n")
  cat("\n")
  cat("Fixed Effect Parameters:\n")
  print(round(x$fix_eff_est,3))
  cat("\n")
  cat("Random Effect Parameters:\n")
  cat("Individual Random Effects Variance-Covariance:\n")
  print(round(rbind(x$ran_eff_b_var_mat, x$ran_eff_b_cov_mat),3), na.print="")
  cat("\n")
  cat("Group Random Effects Variance-Covariance:\n")
  print(round(rbind(x$ran_eff_g_var_mat, x$ran_eff_g_cov_mat),3), na.print="")
  cat("\n")
  cat("Error Variance:\n")
  print(round(x$error_var,3), na.print="")
  cat("\n")
  cat("Gelman's msrf:", round(x$msrf, 3), "\n")
  cat("Mean psrf:", round(x$mean_psrf, 3), "\n")
  cat("DIC:", round(x$DIC,3))
  cat("\n")

  invisible(x)
}

#' Extract fixed effects parameter estimates
#'
#' @description
#' Extracts the fixed effects parameter estimates from a fitted model of class "BPREM", "CREM", or "PREM".
#'
#' @param x An object of class "BPREM", "CREM", or "PREM".
#' @param ... Additional arguments.
#'
#' @returns Returns a vector or data frame of fixed effects parameter values.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_bprem)
#' # get fixed effects
#' getFixEf(results_bprem)
#'
#' @export
getFixEf <- function(x, ...) UseMethod(("getFixEf"))

#' @rdname getFixEf
#' @export
getFixEf.BPREM <- function(x, ...) {

  # determine number of parameters
  n_param <- 4
  # define parameters for labeling
  param_names <- c("Intercept", "Slope", "Change in Slope", "Changepoint")

  fix_eff_est <- matrix(x$Parameter_Estimates[1:(n_param*2),"Mean"], ncol=2)
  colnames(fix_eff_est) <- c(x$Call$y1_var, x$Call$y2_var)
  rownames(fix_eff_est) <- paste0(param_names)
  fix_eff_est <- as.data.frame(fix_eff_est)

  out <- list(fixef = fix_eff_est)
  class(out) <- c("getFixEf.BPREM", "getFixEf.BEND", class(out$fixef))
  return(out)
}

#' @rdname getFixEf
#' @export
getFixEf.CREM <- function(x, ...){

  # determine form
  form <- x$Functional_Form
  # determine number of parameters
  if(form=="linear")                           n_param <- 2
  if(form=="quadratic" | form=="exponential")  n_param <- 3
  if(form=="piecewise")                        n_param <- 4
  # define parameters for labeling
  if(form=="linear")       param_names <- c("Intercept", "Slope")
  if(form=="quadratic")    param_names <- c("Intercept", "Linear Slope", "Quadratic Slope")
  if(form=="exponential")  param_names <- c("Intercept", "Total Change", "Growth Rate")
  if(form=="piecewise")    param_names <- c("Intercept", "Slope", "Change in Slope", "Changepoint")

  fix_eff_est <- x$Parameter_Estimates[1:n_param,"Mean"]
  names(fix_eff_est) <- param_names

  out <- list(fixef = fix_eff_est)
  class(out) <- c("getFixEf.CREM", "getFixEf.BEND", class(out$fixef))
  return(out)
}

#' @rdname getFixEf
#' @export
getFixEf.PREM <- function(x, ...){

  # determine number of classes
  n_class <- length(unique(x$Class_Information$class_membership))

  # determine number of changepoints in each class (based on final model results)
  changepoints <- c()
  for(i in 1:n_class){
    class_num <- paste0("Class_", i)
    changepoints[i] <- which.max(x$Parameter_Estimates[[class_num]]$K_prob)-1
  }
  max_cp <- max(changepoints)

  # pull fixed effects for each class
  fixef_mat <- matrix(NA, nrow=2+2*max_cp, ncol=n_class)
  for(i in 1:n_class){
    class_num <- paste0("Class_", i)
    cp_num <- paste0("K_", changepoints)[i]
    # fixed effects
    fixef_mat[1,i] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[1]
    fixef_mat[2,i] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[2]
    # if there are changepoints (beyond linear)
    if(changepoints[i]>0){
      for(k in 1:changepoints[i]){
        # fixed effects
        fixef_mat[2*k+1,i] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_mean[k]
        fixef_mat[2*k+2,i] <- x$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[k+2]
      }
    }
  }
  colnames(fixef_mat) <- paste0("Class ", 1:n_class)

  my_rownames <- rep(NA, 2+2*max_cp)
  my_rownames[1] = "Intercept"
  my_rownames[2] = "Slope"
  for(k in 1:max_cp){
    my_rownames[2*k+1] = paste0("Changepoint ", k)
    my_rownames[2*k+2] = paste0("Change in Slope ", k)
  }
  rownames(fixef_mat) <- my_rownames

  out <- list(fixef = fixef_mat)
  class(out) <- c("getFixEf.PREM", "getFixEf.BEND", class(out$fixef))
  return(out)
}

#' @rdname getFixEf
#' @export
print.getFixEf.BEND <- function(x, ...){

  cat("Fixed Effects Parameters\n")
  print(x$fixef, na.print="")
  cat("\n")

  invisible(x)
}

#' Simulated bivariate piecewise random effects model data
#'
#' A simulated dataset for the bivariate piecewise random effects model (BPREM).
#'
#' @format ## `SimData_BPREM`
#' A data frame with 210 rows and 4 columns:
#' \describe{
#'   \item{id}{Individual ids}
#'   \item{time}{Time}
#'   \item{y1}{Outcome 1}
#'   \item{y2}{Outcome 2}
#' }
"SimData_BPREM"

#' Simulated piecewise crossed random effects model data
#'
#' A simulated dataset for the piecewise crossed random effects model (PCREM).
#'
#' @format ## `SimData_PCREM`
#' A data frame with 210 rows and 4 columns:
#' \describe{
#'   \item{id}{Individual ids}
#'   \item{teacherid}{Teacher ids}
#'   \item{time}{Time}
#'   \item{y}{Outcome}
#' }
"SimData_PCREM"

#' Simulated piecewise random effects model data
#'
#' A simulated dataset for the piecewise random effects model (PREM).
#'
#' @format ## `SimData_PREM`
#' A data frame with 540 rows and 6 columns:
#' \describe{
#'   \item{id}{Individual ids}
#'   \item{time}{Time}
#'   \item{y}{Outcome}
#'   \item{class_pred_1}{First Class-Predictive Covariate}
#'   \item{class_pred_2}{Second Class-Predictive Covariate}
#'   \item{outcome_pred_1}{First Outcome-Predictive Covariate}
#' }
"SimData_PREM"

#' Bivariate piecewise random effects model results
#'
#' Results from fitting a bivariate piecewise random effects model (BPREM) to `SimData_BPREM` (returned by `Bayes_BPREM(...)`).
#'
#' @format ## `results_bprem`
#' A list object with elements:
#' \describe{
#'   \item{Call}{Function arguments}
#'   \item{Sample_Size}{Data set sample size}
#'   \item{Data}{Data used for model fitting}
#'   \item{Convergence}{Convergence status}
#'   \item{Model_Fit}{Model fit information}
#'   \item{Fitted_Values}{Fitted outcome values}
#'   \item{Parameter_Estimates}{Parameter estimates}
#'   \item{Random_Coefficients}{Random coefficient estimates}
#'   \item{Run_Time}{Model run time}
#' }
"results_bprem"

#' Piecewise crossed random effects model results
#'
#' Results from fitting a piecewise crossed random effects model (PCREM) to `SimData_PCREM` (returned by `Bayes_CREM(...)`).
#'
#' @format ## `results_pcrem`
#' A list object with elements:
#' \describe{
#'   \item{Call}{Function arguments}
#'   \item{Sample_Size}{Data set sample size}
#'   \item{Data}{Data used for model fitting}
#'   \item{Convergence}{Convergence status}
#'   \item{Model_Fit}{Model fit information}
#'   \item{Fitted_Values}{Fitted outcome values}
#'   \item{Parameter_Estimates}{Parameter estimates}
#'   \item{Random_Coefficients}{Random coefficient estimates}
#'   \item{Run_Time}{Model run time}
#' }
"results_pcrem"

#' Piecewise random effects model results
#'
#' Results from fitting a piecewise random effects model (PREM) to `SimData_PREM` (returned by `Bayes_PREM(...)`).
#'
#' @format ## `results_prem`
#' A list object with elements:
#' \describe{
#'   \item{Call}{Function arguments}
#'   \item{Sample_Size}{Data set sample size}
#'   \item{Data}{Data used for model fitting}
#'   \item{Convergence}{Convergence status}
#'   \item{Model_Fit}{Model fit information}
#'   \item{Fitted_Values}{Fitted outcome values}
#'   \item{Parameter_Estimates}{Parameter estimates}
#'   \item{Random_Coefficients}{Random coefficient estimates}
#'   \item{Run_Time}{Model run time}
#' }
"results_prem"

#' Print the results of a bivariate piecewise random effects model (BPREM)
#'
#' @description
#' Provides a summary of a BPREM model, as returned by `Bayes_BPREM()`.
#'
#' @param x An object of class "BPREM" (returned by `Bayes_BPREM(...)`).
#' @param ... Additional arguments.
#'
#' @returns Returns a list of key parameter estimates.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_bprem)
#' # print results
#' print(results_bprem)
#'
#' @export
print.BPREM <- function(x, ...){

  # determine number of parameters
  n_param <- 4

  # determine number of covariances
  n_covar <- ((n_param*2)*((n_param*2)-1))/2

  # define parameters for labeling
  param_names <- c("Intercept", "Slope", "Change in Slope", "Changepoint")

  # FIXED EFFECTS -----
  fix_eff_est <- matrix(x$Parameter_Estimates[1:(n_param*2),"Mean"], ncol=2)
  colnames(fix_eff_est) <- c("Outcome 1", "Outcome 2")
  rownames(fix_eff_est) <- paste0(param_names, " Mean")

  # RANDOM EFFECTS -----

  ## COVARIANCE MATRIX -----
  ran_eff_cov_mat <- diag(n_param*2)
  diag(ran_eff_cov_mat) <- x$Parameter_Estimates[((n_param*2)+(1:(n_param*2))),"Mean"]

  ran_eff_cov_mat[upper.tri(ran_eff_cov_mat)] <- x$Parameter_Estimates[(2*(n_param*2)+(1:n_covar)),"Mean"] # covariances on upper triangle
  ran_eff_cov_mat[lower.tri(ran_eff_cov_mat)] <- t(ran_eff_cov_mat)[lower.tri(ran_eff_cov_mat)] # copy to lower triangle
  ran_eff_cov_mat[upper.tri(ran_eff_cov_mat)] <- NA

  colnames(ran_eff_cov_mat) <- rownames(ran_eff_cov_mat) <- paste0(rep(c("Outcome 1: ", "Outcome 2: "), e=4), rep(param_names,2))

  ## CORRELATION MATRIX -----
  ran_eff_corr_mat <- diag(n_param*2)

  ran_eff_corr_mat[upper.tri(ran_eff_corr_mat)] <- x$Parameter_Estimates[(2*(n_param*2)+n_covar+3+(1:n_covar)),"Mean"] # covariances on upper triangle
  ran_eff_corr_mat[lower.tri(ran_eff_corr_mat)] <- t(ran_eff_corr_mat)[lower.tri(ran_eff_corr_mat)] # copy to lower triangle
  ran_eff_corr_mat[upper.tri(ran_eff_corr_mat)] <- NA

  colnames(ran_eff_corr_mat) <- rownames(ran_eff_corr_mat) <- paste0(rep(c("Outcome 1: ", "Outcome 2: "), e=4), rep(param_names,2))

  # ERROR -----
  error_cov_mat <- diag(2)
  diag(error_cov_mat) <- x$Parameter_Estimates[(2*(n_param*2)+n_covar+(1:2)),"Mean"]

  error_cov_mat[upper.tri(error_cov_mat)] <- error_cov_mat[lower.tri(error_cov_mat)] <- x$Parameter_Estimates[(2*(n_param*2)+n_covar+(3)),"Mean"] # covariances on upper triangle
  error_cov_mat[upper.tri(error_cov_mat)] <- NA

  colnames(error_cov_mat) <- rownames(error_cov_mat) <- paste0(c("Outcome 1: ", "Outcome 2: "), "Error")

  ## return value
  out <- list("fix_eff_est" = fix_eff_est,
              "ran_eff_cov_mat" = ran_eff_cov_mat,
              "ran_eff_corr_mat" = ran_eff_corr_mat,
              "error_cov_mat" = error_cov_mat,
              "error_corr" = x$Parameter_Estimates[(2*(n_param*2)+2*n_covar+3+1),"Mean"],
              "msrf" = x$Convergence$multivariate_psrf,
              "mean_psrf" = x$Convergence$mean_psrf,
              "DIC" = x$Model_Fit$dic)

  cat("Bayesian bivariate piecewise random effects model\n")
  cat("Data:", x$Call$data, "\n")
  cat("Outcomes:", c(x$Call$y1_var, x$Call$y2_var), "\n")
  cat("\n")
  cat("Fixed Effect Parameters:\n")
  print(round(out$fix_eff_est,3))
  cat("\n")
  cat("Random Effect Parameters:\n")
  cat("Covariance Matrix:\n")
  print(round(out$ran_eff_cov_mat,3), na.print="")
  cat("\n")
  cat("Correlation Matrix:\n")
  print(round(out$ran_eff_corr_mat,3), na.print="")
  cat("\n")
  cat("Error:\n")
  cat("Covariance Matrix:\n")
  print(round(out$error_cov_mat,3), na.print="")
  cat("\n")
  cat("Error Corr:", round(out$error_corr,3), "\n")
  cat("Gelman's msrf:", round(out$msrf, 3), "\n")
  cat("Mean psrf:", round(out$mean_psrf, 3), "\n")
  cat("DIC:", round(out$DIC,3))
  cat("\n")

  invisible(x)
}

#' Plot the results of a piecewise random effects model (PREM)
#'
#' @description
#' Provides a fitted plot of a PREM model, as returned by `Bayes_PREM()`.
#'
#' @param x An object of class "PREM" (returned by `Bayes_PREM(...)`).
#' @param xlab X-axis label for the generated plot.
#' @param ylab Y-axis label for the generated plot.
#' @param colors Colors for each class (`PREMM` or `CI-PREMM`). By default, up to 5 colors are provided in the following order: "blue" (class 1), "red" (class 2), "green" (class 3), "gold" (class 4), "gray" (class 5).
#' @param mean_colors Colors for the trajectory defined by the mean parameters for each class (`PREMM` or `CI-PREMM`). By default, up to 5 colors are provided in the following order: "darkblue" (class 1), "darkred" (class 2), "darkgreen" (class 3), "gold4" (class 4), "darkgray" (class 5).
#' @param legend_pos (optional) Option to change legend position (default = "topright").
#' @param ... (optional) Other parameters to pass to the `plot()` function.
#'
#' @returns No return value.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_prem)
#' # plot fitted results
#' plot(results_prem)
#'
#' @import graphics
#'
#' @export
plot.PREM <- function(x,
                      xlab='X', ylab='Y',
                      colors=NULL, mean_colors=NULL,
                      legend_pos="topright", ...){
  # Setup ----

  object <- x

  data <- object$Data
  id_var <- object$Call$id_var
  time_var <- object$Call$time_var
  y_var <- object$Call$y_var

  if(is.null(colors)) colors <- c('blue','red','green','gold','gray')
  if(is.null(mean_colors)) mean_colors = c('darkblue','darkred','darkgreen','gold4','darkgray')

  ## outcome data - matrix form
  y <- reshape(data[,c(id_var, time_var, y_var)],
               idvar=id_var,
               timevar=time_var,
               direction='wide')
  y <- unname(as.matrix(y[,names(y)!=id_var]))

  ## time data - matrix form
  ## should be the same dimensions as y
  x <- matrix(data[,c(time_var)],
              byrow=TRUE,
              nrow=dim(y)[1],
              ncol=dim(y)[2])

  ## Define relevant variables
  n_subj <- nrow(y)
  n_time <- ncol(y)
  xvec <- seq(min(x), max(x), length.out=100)

  # Fitted Plot -----


  # determine number of classes
  n_class <- length(unique(object$Class_Information$class_membership))

  # determine number of changepoints in each class (based on final model results)
  changepoints <- c()
  class_data <- data.frame()
  for(i in 1:n_class){
    class_num <- paste0("Class_", i)
    changepoints[i] <- which.max(object$Parameter_Estimates[[class_num]]$K_prob)-1
  }
  max_cp <- max(changepoints)

  # determine who is in each class
  class_list <- list()
  for(i in 1:n_class){
    class_list[[i]] <- c(1:n_subj)[object$Class_Information$class_membership==i]
  }

  # class mean estimates
  class_means <- list()
  for(i in 1:n_class){
    class_num <- paste0("Class_", i)
    cp_num <- paste0("K_", changepoints)[i]
    k <- changepoints[i]

    if(k==0) I <- rep(0, max_cp)
    if(k>0) I <- c(rep(1,k), rep(0,max_cp-k+2))
    # i = k+1
    int <- object$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[1]
    slope1 <- object$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[2]

    slope_cp <- rep(0,max_cp)
    cp <- rep(0,max_cp)
    if(k>0){
      slope_cp[1:k] <- object$Parameter_Estimates[[class_num]]$K[[cp_num]]$beta_mean[3:(k+2)]
      cp[1:k] <- object$Parameter_Estimates[[class_num]]$K[[cp_num]]$cp_mean[1:k]
    }

    class_means[[i]] <- rep(0,100)
    for(j in 1:100){
      temp <- int + slope1*xvec[j]
      if(k>0){
        for(l in 1:k){
          temp <- temp + slope_cp[l]*(max(0, xvec[j]-cp[l]))}
      }
      class_means[[i]][j] <- temp
    }
  }

  plot(NULL, NULL,
       xlim = c(min(x), max(x)),
       ylim = c(min(y,na.rm=TRUE), max(y,na.rm=TRUE)),
       ylab = ylab,
       xlab = xlab, ...)
  for(i in 1:n_class){
    for(j in class_list[[i]]){
      points(x[j,!is.na(y[j,])],
             y[j,!is.na(y[j,])],
             type = "l",
             col=colors[i])
    }
    points(xvec,
           class_means[[i]],
           type='l',
           col=mean_colors[i],
           lwd=4)
  }
  legend(legend_pos, lty=1, col=colors[1:n_class], legend=paste0("Class ", 1:n_class))

}

#' Plot the results of a crossed random effects model (CREM)
#'
#' @description
#' Provides a fitted plot of a CREM model, as returned by `Bayes_CREM()`.
#'
#' @param x An object of class "CREM" (returned by `Bayes_CREM(...)`).
#' @param xlab X-axis label for the generated plot.
#' @param ylab Y-axis label for the generated plot.
#' @param colors Color for observed trajectories (optional). Default is "grey".
#' @param mean_colors Colors for the trajectory defined by the mean parameters for each outcome (optional). Default is "black".
#' @param legend_pos (optional) Option to change legend position (default = "topright").
#' @param ... (optional) Other parameters to pass to the `plot()` function.
#'
#' @returns No return value.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_pcrem)
#' # plot fitted results
#' plot(results_pcrem)
#'
#' @import graphics
#'
#' @export
plot.CREM <- function(x,
                      xlab='X', ylab='Y',
                      colors=NULL, mean_colors=NULL,
                      legend_pos="topright", ...){
  # Setup ----

  if(is.null(colors)) colors <- c('grey')
  if(is.null(mean_colors)) mean_colors = c('black')

  object <- x

  data <- object$Data
  id_var <- object$Call$ind_id_var
  time_var <- object$Call$time_var
  y_var <- object$Call$y_var

  ## outcome data - matrix form
  y <- reshape(data[,c(id_var, time_var, y_var)],
               idvar=id_var,
               timevar=time_var,
               direction='wide')
  y <- unname(as.matrix(y[,names(y)!=id_var]))

  ## time data - matrix form
  ## should be the same dimensions as y
  x <- matrix(data[,c(time_var)],
              byrow=TRUE,
              nrow=dim(y)[1],
              ncol=dim(y)[2])

  ## Define relevant variables
  n_subj <- nrow(y)
  n_time <- ncol(y)
  xvec <- seq(min(x), max(x), length.out=100)

  # Fitted Plot -----

  # determine form
  form <- object$Functional_Form

  # determine number of parameters
  if(form=="linear")                           n_param <- 2
  if(form=="quadratic" | form=="exponential")  n_param <- 3
  if(form=="piecewise")                        n_param <- 4

  # pull fixed effect estimates
  mean_est <- object$Parameter_Estimates$Mean[1:n_param]

  # define functional form equation

  fit_form_eq <- function(x,est){
    if(form=="linear")       return(est[1] + est[2]*x)
    if(form=="quadratic")    return(est[1] + est[2]*x + est[3]*(x^2))
    if(form=="exponential")  return(est[1] + est[2]*(1-exp(-est[3]*x)))
    if(form=="piecewise")    return(est[1] + est[2]*x + est[3]*(max(0, x-est[4])))
  }

  mean_traj <- rep(0,100)
  for(i in 1:100){
    mean_traj[i] <- fit_form_eq(xvec[i], mean_est)
  }

  plot(x[1, !is.na(y[1,])],
       y[1, !is.na(y[1,])],
       type = "l",
       col = colors,
       ylim = c(min(y,na.rm=TRUE), max(y,na.rm=TRUE)),
       xlim= c(min(x,na.rm=TRUE), max(x,na.rm=TRUE)),
       xlab = xlab,
       ylab = ylab, ...)
  for(i in 2:n_subj){
    lines(x[i, !is.na(y[i,])],
          y[i, !is.na(y[i,])],
          type = "l",
          col = colors)
  }
  points(xvec,
         mean_traj,
         type='l',
         col=mean_colors,
         lwd=4)

}

#' Extract changepoint probabilities
#'
#' @description
#' Extracts the K (number of changepoints) probabilities from a fitted model of class "PREM".
#'
#' @param x An object of class "PREM".
#' @param ... Additional arguments.
#'
#' @returns Returns a list of the posterior probabilities for each possible number of changepoints, for each class.
#'
#' @author Corissa T. Rohloff
#'
#' @examples
#' # load fitted model results
#' data(results_prem)
#' # get changepoint probabilities
#' getKProb(results_prem)
#'
#' @export
getKProb <- function(x, ...) UseMethod(("getKProb"))

#' @rdname getKProb
#' @export
getKProb.PREM <- function(x, ...){

  # determine number of classes
  n_class <- length(unique(x$Class_Information$class_membership))

  k_probs <- vector("list", n_class)
  for(i in 1:n_class){
    class_num <- paste0("Class_", i)
    k_probs[[i]] <- x$Parameter_Estimates[[class_num]]$K_prob
  }
  names(k_probs) <- paste0("Class ", 1:n_class)

  out <- k_probs
  class(out) <- c("getKProb.PREM", class(out))
  return(out)

}

#' @rdname getKProb
#' @export
print.getKProb.PREM <- function(x, ...){

  cat("K (Number of Changepoints) Probabilities\n\n")
  for(i in 1:length(x)){
    cat(paste0("Class ",i,"\n"))
    print(x[[i]], na.print="")
    cat("\n")
  }

  invisible(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getKProb.R
\name{getKProb}
\alias{getKProb}
\alias{getKProb.PREM}
\alias{print.getKProb.PREM}
\title{Extract changepoint probabilities}
\usage{
getKProb(x, ...)

\method{getKProb}{PREM}(x, ...)

\method{print}{getKProb.PREM}(x, ...)
}
\arguments{
\item{x}{An object of class "PREM".}

\item{...}{Additional arguments.}
}
\value{
Returns a list of the posterior probabilities for each possible number of changepoints, for each class.
}
\description{
Extracts the K (number of changepoints) probabilities from a fitted model of class "PREM".
}
\examples{
# load fitted model results
data(results_prem)
# get changepoint probabilities
getKProb(results_prem)

}
\author{
Corissa T. Rohloff
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getFixEf.R
\name{getFixEf}
\alias{getFixEf}
\alias{getFixEf.BPREM}
\alias{getFixEf.CREM}
\alias{getFixEf.PREM}
\alias{print.getFixEf.BEND}
\title{Extract fixed effects parameter estimates}
\usage{
getFixEf(x, ...)

\method{getFixEf}{BPREM}(x, ...)

\method{getFixEf}{CREM}(x, ...)

\method{getFixEf}{PREM}(x, ...)

\method{print}{getFixEf.BEND}(x, ...)
}
\arguments{
\item{x}{An object of class "BPREM", "CREM", or "PREM".}

\item{...}{Additional arguments.}
}
\value{
Returns a vector or data frame of fixed effects parameter values.
}
\description{
Extracts the fixed effects parameter estimates from a fitted model of class "BPREM", "CREM", or "PREM".
}
\examples{
# load fitted model results
data(results_bprem)
# get fixed effects
getFixEf(results_bprem)

}
\author{
Corissa T. Rohloff
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BEND.R, R/data.R
\docType{data}
\name{results_prem}
\alias{results_prem}
\title{Fitted results for a PREM}
\format{
A list (an object of class \code{PREM}) with fitted model results.

\subsection{\code{results_prem}}{

A list object with elements:
\describe{
\item{Call}{Function arguments}
\item{Sample_Size}{Data set sample size}
\item{Data}{Data used for model fitting}
\item{Convergence}{Convergence status}
\item{Model_Fit}{Model fit information}
\item{Fitted_Values}{Fitted outcome values}
\item{Parameter_Estimates}{Parameter estimates}
\item{Random_Coefficients}{Random coefficient estimates}
\item{Run_Time}{Model run time}
}
}
}
\usage{
data(results_prem)

results_prem
}
\description{
Fitted results for a piecewise random effects model (PREM) using \code{SimData_PREM}. Included to demonstrate the use of \code{plot_BEND()} and \code{summary.PREM()}.

Results from fitting a piecewise random effects model (PREM) to \code{SimData_PREM} (returned by \code{Bayes_PREM(...)}).
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_BPREM.R
\name{plot.BPREM}
\alias{plot.BPREM}
\title{Plot the results of a bivariate piecewise random effects model (BPREM)}
\usage{
\method{plot}{BPREM}(
  x,
  xlab = "X",
  ylab = "Y",
  colors = NULL,
  mean_colors = NULL,
  legend_pos = "topright",
  ...
)
}
\arguments{
\item{x}{An object of class "BPREM" (returned by \code{Bayes_BPREM(...)}).}

\item{xlab}{X-axis label for the generated plot.}

\item{ylab}{Y-axis label for the generated plot.}

\item{colors}{Colors for each class outcome. By default, up to 2 colors are provided in the following order: "blue" (outcome 1), "red" (outcome 2).}

\item{mean_colors}{Colors for the trajectory defined by the mean parameters for each outcome. By default, up to 2 colors are provided in the following order: "darkblue" (outcome 1), "darkred" (outcome 2).}

\item{legend_pos}{(optional) Option to change legend position (default = "topright").}

\item{...}{(optional) Other parameters to pass to the \code{plot()} function.}
}
\value{
No return value.
}
\description{
Provides a fitted plot of a BPREM model, as returned by \code{Bayes_BPREM()}.
}
\examples{
# load fitted model results
data(results_bprem)
# plot fitted results
plot(results_bprem)

}
\author{
Corissa T. Rohloff
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Bayes_PREM.R
\name{Bayes_PREM}
\alias{Bayes_PREM}
\title{Bayesian Piecewise Random Effects Model (PREM) + Extensions}
\usage{
Bayes_PREM(
  data,
  id_var,
  time_var,
  y_var,
  n_class = 1,
  max_cp = 2,
  class_predictive_vars = NULL,
  outcome_predictive_vars = NULL,
  scale_prior = "uniform",
  alpha = 1,
  cp_prior = "binomial",
  binom_prob = 0.5,
  iters_adapt = 1000,
  iters_burn_in = 20000,
  iters_sampling = 30000,
  thin = 15,
  save_full_chains = FALSE,
  save_conv_chains = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame in long format, where each row describes a measurement occasion for a given individual. It is assumed that each individual has the same number of assigned timepoints (a.k.a., rows). There can be missingness in the outcome (\code{y_var}), but there cannot be missingness in time (\code{time_var}).}

\item{id_var}{Name of column that contains ids for individuals with repeated measures in a longitudinal dataset.}

\item{time_var}{Name of column that contains the time variable. This column cannot contain any missing values.}

\item{y_var}{Name of column that contains the outcome variable. Missing values should be denoted by NA.}

\item{n_class}{Number of latent classes (default = 1). Note, CI-PREMM only allows for two classes.}

\item{max_cp}{Maximum number of changepoints in each latent class (default = 2).}

\item{class_predictive_vars}{Name(s) of column(s) that contain class-predictive covariates (time-invariant only). Give a vector of names if multiple covariates. Note, there cannot be any missingness in the covariates.}

\item{outcome_predictive_vars}{Name(s) of column(s) that contain outcome-predictive covariates (time-varying or -invariant). Give a vector of names if multiple covariates. Note, there cannot be any missingness in the covariates.}

\item{scale_prior}{Prior for the scale parameter for the hierarchical random effects. Options include: ‘uniform’ (scaled uniform prior; default) or ‘hc’ (scaled half-cauchy prior).}

\item{alpha}{Concentration parameter for Dirichlet prior for latent classes (default = 1). This can be a vector of values corresponding to the number of classes (specified by n_class). Note, this is not used for CI-PGMM.}

\item{cp_prior}{Prior for the number of changepoints in each class. Options include: 'binomial' (default) or 'uniform'.}

\item{binom_prob}{Probability for binomial prior, if specified (default = 0.5).}

\item{iters_adapt}{(optional) Number of iterations for adaptation of jags model (default = 1000).}

\item{iters_burn_in}{(optional) Number of iterations for burn-in (default = 20000).}

\item{iters_sampling}{(optional) Number of iterations for posterior sampling (default = 30000).}

\item{thin}{(optional) Thinning interval for posterior sampling (default = 15).}

\item{save_full_chains}{Logical indicating whether the MCMC chains from rjags should be saved (default = FALSE). Note, this should not be used regularly as it will result in an object with a large file size.}

\item{save_conv_chains}{Logical indicating whether the MCMC chains from rjags should be saved but only for the parameters monitored for convergence (default = FALSE). This would be useful for plotting traceplots for relevant model parameters to evaluate convergence behavior. Note, this should not be used regularly as it will result in an object with a large file size.}

\item{verbose}{Logical controlling whether progress messages/bars are generated (default = TRUE).}
}
\value{
A list (an object of class \code{PREM}) with elements:
\item{Call}{A list of all specified function arguments.}
\item{Sample Size}{Number of subjects (\code{n_subj}) and number of timepoints (\code{n_time}) in the dataset.}
\item{Data}{Dataset used for estimation.}
\item{Convergence}{Potential scale reduction factor (PSRF) for each parameter (\code{parameter_psrf}), Gelman multivariate scale reduction factor (\code{multivariate_psrf}), and mean PSRF (\code{mean_psrf}) to assess model convergence.}
\item{Model_Fit}{Deviance (\code{deviance}), effective number of parameters (\code{pD}), and Deviance information criterion (\code{dic}) to assess model fit.}
\item{Fitted_Values}{Data frame with the fitted values at each timepoint for each individual.}
\item{Parameter_Estimates}{Data frame with posterior mean and 95\% credible intervals for each model parameter.}
\item{Random_Coefficients}{List object with data frames providing random coefficients for each individual.}
\item{Run_Time}{Total run time for model fitting.}
\item{Full_MCMC_Chains}{If save_full_chains=TRUE, raw MCMC chains from rjags.}
\item{Convergence_MCMC_Chains}{If save_conv_chains=TRUE, raw MCMC chains from rjags but only for the parameters monitored for convergence.}
\code{Class_Information} contains a list with elements:
\item{class_membership}{Vector of length n with class membership assignments for each individual.}
\item{individ_class_probability}{nxC matrix with each individual’s probabilities of belonging to each class conditional on their class-predictive covariates (when applicable) and growth curve.}
\item{unconditional_class_probability}{This output will differ based on which model was fit. For a PREM or CI-PREM, this will equal 1 as there is only one class. For a PREMM or CI-PREMM with only outcome-predictive covariates, this will be a vector of length C denoting the population probability of belonging to each class. For a CI-PREMM with class-predictive covariates, this will be a vector of length n denoting the probability of each individual belonging to the non-reference class (Class 2) based on their class-predictive covariates only.}
}
\description{
Estimates a Bayesian piecewise random effects model (PREM), with some useful extensions. There are three model options included in this function:
\itemize{
\item \code{PREM} estimates a Bayesian piecewise random effects model with a latent number of changepoints (default). Allows the inclusion of outcome-predictive covariates (\code{CI-PREM}).
\item \code{PREMM} estimates a piecewise random effects mixture model for a given number of latent classes and a latent number of possible changepoints in each class.
\item \code{CI-PREMM} estimates a covariate influenced piecewise random effects mixture model for a given number of latent classes and a latent number of possible changepoints in each class. Allows the inclusion of outcome- and/or class-predictive covariates.
See Lock et al. (2018) and Lamm (2022) for more details.
}
}
\details{
For more information on the model equation and priors implemented in this function, see Lamm et al. (2022; CI-PREMM) and Lock et al. (2018; PREMM).
}
\examples{
\donttest{
# load simulated data
data(SimData_PREM)

# PREM ---------------------------------------------------------------------------------
# fit Bayes_PREM()
results_prem <- Bayes_PREM(data = SimData_PREM,
                           id_var = "id",
                           time_var = "time",
                           y_var = "y")
# result summary
summary(results_prem)
# plot fitted results
plot(results_prem)

# CI-PREM ---------------------------------------------------------------------------------
# fit Bayes_PREM()
results_ciprem <- Bayes_PREM(data = SimData_PREM,
                             id_var = "id",
                             time_var = "time",
                             y_var = "y",
                             outcome_predictive_vars = "outcome_pred_1")
# result summary
summary(results_ciprem)
# plot fitted results
plot(results_ciprem)

# PREMM ---------------------------------------------------------------------------------
# fit Bayes_PREM()
results_premm <- Bayes_PREM(data = SimData_PREM,
                            id_var = "id",
                            time_var = "time",
                            y_var = "y",
                            n_class = 2)
# result summary
summary(results_premm)
# plot fitted results
plot(results_premm)


# CI-PREMM ---------------------------------------------------------------------------------
# fit Bayes_PREM()
results_cipremm <- Bayes_PREM(data = SimData_PREM,
                              id_var = "id",
                              time_var = "time",
                              y_var = "y",
                              n_class = 2,
                              class_predictive_vars = c("class_pred_1", "class_pred_2"),
                              outcome_predictive_vars = "outcome_pred_1")
# result summary
summary(results_cipremm)
# plot fitted results
plot(results_cipremm)
}

}
\references{
Lamm, R. (2022). Incorporation of covariates in Bayesian piecewise growth mixture models. https://hdl.handle.net/11299/252533

Lock, E. F., Kohli, N., & Bose, M. (2018). Detecting multiple random changepoints in Bayesian piecewise growth mixture models. Psychometrika, 83(3), 733–750. https://doi.org/10.1007/s11336-017-9594-5
}
\author{
Corissa T. Rohloff, Rik Lamm, Eric F. Lock
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Bayes_CREM.R
\name{Bayes_CREM}
\alias{Bayes_CREM}
\title{Bayesian Crossed Random Effects Model (CREM)}
\usage{
Bayes_CREM(
  data,
  ind_id_var,
  cross_id_var,
  time_var,
  y_var,
  form = "linear",
  fixed_effects = NULL,
  iters_adapt = 5000,
  iters_burn_in = 50000,
  iters_sampling = 50000,
  thin = 15,
  save_full_chains = FALSE,
  save_conv_chains = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame in long format, where each row describes a measurement occasion for a given individual. It is assumed that each individual has the same number of assigned timepoints (a.k.a., rows). There can be missingness in the outcome (\code{y_var}), but there cannot be missingness in time (\code{time_var}).}

\item{ind_id_var}{Name of column that contains ids for individuals with repeated measures in a longitudinal dataset (e.g., students).}

\item{cross_id_var}{Name of column that contains ids for the crossed factor (e.g., teachers).}

\item{time_var}{Name of column that contains the time variable. This column cannot contain any missing values.}

\item{y_var}{Name of column that contains the outcome variable. Missing values should be denoted by NA.}

\item{form}{Name of the functional form. Options include: ‘linear’ (default), ‘quadratic’, ‘exponential’, ‘piecewise’.}

\item{fixed_effects}{(optional) Starting values for the fixed effects parameters.}

\item{iters_adapt}{(optional) Number of iterations for adaptation of jags model (default = 5000).}

\item{iters_burn_in}{(optional) Number of iterations for burn-in (default = 50000).}

\item{iters_sampling}{(optional) Number of iterations for posterior sampling (default = 50000).}

\item{thin}{(optional) Thinning interval for posterior sampling (default = 15).}

\item{save_full_chains}{Logical indicating whether the MCMC chains from rjags should be saved (default = FALSE). Note, this should not be used regularly as it will result in an object with a large file size.}

\item{save_conv_chains}{Logical indicating whether the MCMC chains from rjags should be saved but only for the parameters monitored for convergence (default = FALSE). This would be useful for plotting traceplots for relevant model parameters to evaluate convergence behavior. Note, this should not be used regularly as it will result in an object with a large file size.}

\item{verbose}{Logical controlling whether progress messages/bars are generated (default = TRUE).}
}
\value{
A list (an object of class \code{CREM}) with elements:
\item{Call}{A list of all specified function arguments.}
\item{Sample Size}{Number of subjects (\code{n_subj}), number of groups (\code{n_group}) and number of timepoints (\code{n_time}) in the dataset.}
\item{Data}{Dataset used for estimation.}
\item{Convergence}{Potential scale reduction factor (PSRF) for each parameter (\code{parameter_psrf}), Gelman multivariate scale reduction factor (\code{multivariate_psrf}), and mean PSRF (\code{mean_psrf}) to assess model convergence.}
\item{Model_Fit}{Deviance (\code{deviance}), effective number of parameters (\code{pD}), and Deviance information criterion (\code{dic}) to assess model fit.}
\item{Fitted_Values}{Data frame with the fitted values at each timepoint for each individual.}
\item{Functional_Form}{Functional form fitted.}
\item{Parameter_Estimates}{Data frame with posterior mean and 95\% credible intervals for each model parameter.}
\item{Random_Coefficients}{List object with data frames providing individual random effects, group random effects, and random coefficients for each individual and group.}
\item{Run_Time}{Total run time for model fitting.}
\item{Full_MCMC_Chains}{If save_full_chains=TRUE, raw MCMC chains from rjags.}
\item{Convergence_MCMC_Chains}{If save_conv_chains=TRUE, raw MCMC chains from rjags but only for the parameters monitored for convergence.}
}
\description{
Estimates a Bayesian crossed random effects models (CREM) for longitudinal data with dynamic group membership. Four different choices for functional forms are provided: linear, quadratic, exponential, and piecewise. See Rohloff et al. (2024) for more details.
}
\details{
For more information on the model equation and priors implemented in this function, see Rohloff et al. (2024).

Note, this function differs from the above reference by estimating the covariances between the random effects parameters. The variance-covariance matrices of the individual and group random effects have a scaled inverse-Wishart prior (see Peralta et al., 2022).
}
\examples{
\donttest{
# load simulated data
data(SimData_PCREM)
# fit Bayes_CREM()
results_pcrem <- Bayes_CREM(data = SimData_PCREM,
                            ind_id_var = "id",
                            cross_id_var = "teacherid",
                            time_var = "time",
                            y_var = "y",
                            form="piecewise")
# result summary
summary(results_pcrem)
# plot fitted results
plot(results_pcrem)
}

}
\references{
Peralta, Y., Kohli, N., Lock, E. F., & Davison, M. L. (2022). Bayesian modeling of associations in bivariate piecewise linear mixed-effects models. Psychological Methods, 27(1), 44–64. https://doi.org/10.1037/met0000358

Rohloff, C. T., Kohli, N., & Lock, E. F. (2024). Identifiability and estimability of Bayesian linear and nonlinear crossed random effects models. British Journal of Mathematical and Statistical Psychology. https://doi.org/10.1111/bmsp.12334
}
\author{
Corissa T. Rohloff
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getFitted.R
\name{getFitted}
\alias{getFitted}
\alias{getFitted.BEND}
\alias{print.getFitted}
\title{Extract fitted values}
\usage{
getFitted(x, ...)

\method{getFitted}{BEND}(x, ...)

\method{print}{getFitted}(x, ...)
}
\arguments{
\item{x}{An object of class "BPREM", "CREM", or "PREM".}

\item{...}{Additional arguments.}
}
\value{
Returns a data frame of the fitted values for each individual and timepoint.
}
\description{
Extracts the individual fitted values from a fitted model of class "BPREM", "CREM", or "PREM".
}
\examples{
# load fitted model results
data(results_prem)
# get fitted values
getFitted(results_prem)

}
\author{
Corissa T. Rohloff
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{summary.CREM}
\alias{summary.CREM}
\alias{print.summary.CREM}
\title{Summarize the results of a crossed random effects model (CREM)}
\usage{
\method{summary}{CREM}(object, ...)

\method{print}{summary.CREM}(x, ...)
}
\arguments{
\item{object}{An object of class "CREM" (returned by \code{Bayes_CREM(...)}).}

\item{...}{Additional arguments.}

\item{x}{An object of class "summary.CREM".}
}
\value{
Returns a list of key parameter estimates.
}
\description{
Provides a summary of a CREM model, as returned by \code{Bayes_CREM()}.
}
\examples{
# load fitted model results
data(results_pcrem)
# result summary
summary(results_pcrem)

}
\author{
Corissa T. Rohloff
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Bayes_BPREM.R
\name{Bayes_BPREM}
\alias{Bayes_BPREM}
\title{Bayesian Bivariate Piecewise Random Effects Model (BPREM)}
\usage{
Bayes_BPREM(
  data,
  id_var,
  time_var,
  y1_var,
  y2_var,
  iters_adapt = 5000,
  iters_burn_in = 1e+05,
  iters_sampling = 50000,
  thin = 15,
  save_full_chains = FALSE,
  save_conv_chains = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame in long format, where each row describes a measurement occasion for a given individual. It is assumed that each individual has the same number of assigned timepoints (a.k.a., rows). There can be missingness in the outcomes (\code{y1_var} and \code{ y2_var}), but there cannot be missingness in time (\code{time_var}).}

\item{id_var}{Name of column that contains ids for individuals with repeated measures in a longitudinal dataset.}

\item{time_var}{Name of column that contains the time variable. This column cannot contain any missing values.}

\item{y1_var}{Name of column that contains the first outcome variable. Missing values should be denoted by NA.}

\item{y2_var}{Name of column that contains the second outcome variable. Missing values should be denoted by NA.}

\item{iters_adapt}{(optional) Number of iterations for adaptation of jags model (default = 5000).}

\item{iters_burn_in}{(optional) Number of iterations for burn-in (default = 100000).}

\item{iters_sampling}{(optional) Number of iterations for posterior sampling (default = 50000).}

\item{thin}{(optional) Thinning interval for posterior sampling (default = 15).}

\item{save_full_chains}{Logical indicating whether the MCMC chains from rjags should be saved (default = FALSE). Note, this should not be used regularly as it will result in an object with a large file size.}

\item{save_conv_chains}{Logical indicating whether the MCMC chains from rjags should be saved but only for the parameters monitored for convergence (default = FALSE). This would be useful for plotting traceplots for relevant model parameters to evaluate convergence behavior. Note, this should not be used regularly as it will result in an object with a large file size.}

\item{verbose}{Logical controlling whether progress messages/bars are generated (default = TRUE).}
}
\value{
A list (an object of class \code{BPREM}) with elements:
\item{Call}{A list of all specified function arguments.}
\item{Sample Size}{Number of subjects (\code{n_subj}) and number of timepoints (\code{n_time}) in the dataset.}
\item{Data}{Dataset used for estimation.}
\item{Convergence}{Potential scale reduction factor (PSRF) for each parameter (\code{parameter_psrf}), Gelman multivariate scale reduction factor (\code{multivariate_psrf}), and mean PSRF (\code{mean_psrf}) to assess model convergence.}
\item{Model_Fit}{Deviance (\code{deviance}), effective number of parameters (\code{pD}), and Deviance information criterion (\code{dic}) to assess model fit.}
\item{Fitted_Values}{Data frame with the fitted values for y1 and y2 at each timepoint for each individual.}
\item{Parameter_Estimates}{Data frame with posterior mean and 95\% credible intervals for each model parameter.}
\item{Random_Coefficients}{List object with data frames providing random effects and random coefficients for each individual.}
\item{Run_Time}{Total run time for model fitting.}
\item{Full_MCMC_Chains}{If save_full_chains=TRUE, raw MCMC chains from rjags.}
\item{Convergence_MCMC_Chains}{If save_conv_chains=TRUE, raw MCMC chains from rjags but only for the parameters monitored for convergence.}
}
\description{
Estimates a Bayesian bivariate piecewise random effects models (BPREM) for longitudinal data with two interrelated outcomes. See Peralta et al. (2022) for more details.
}
\details{
For more information on the model equation and priors implemented in this function, see Peralta et al. (2022).
}
\examples{
\donttest{
# load simulated data
data(SimData_BPREM)
# fit Bayes_BPREM()
results_bprem <- Bayes_BPREM(data = SimData_BPREM,
                             id_var = "id",
                             time_var = "time",
                             y1_var = "y1",
                             y2_var = "y2")
# result summary
summary(results_bprem)
# plot fitted results
plot(results_bprem)
}

}
\references{
Peralta, Y., Kohli, N., Lock, E. F., & Davison, M. L. (2022). Bayesian modeling of associations in bivariate piecewise linear mixed-effects models. Psychological Methods, 27(1), 44–64. https://doi.org/10.1037/met0000358
}
\author{
Corissa T. Rohloff, Yadira Peralta
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot.CREM}
\alias{plot.CREM}
\title{Plot the results of a crossed random effects model (CREM)}
\usage{
\method{plot}{CREM}(
  x,
  xlab = "X",
  ylab = "Y",
  colors = NULL,
  mean_colors = NULL,
  legend_pos = "topright",
  ...
)
}
\arguments{
\item{x}{An object of class "CREM" (returned by \code{Bayes_CREM(...)}).}

\item{xlab}{X-axis label for the generated plot.}

\item{ylab}{Y-axis label for the generated plot.}

\item{colors}{Color for observed trajectories (optional). Default is "grey".}

\item{mean_colors}{Colors for the trajectory defined by the mean parameters for each outcome (optional). Default is "black".}

\item{legend_pos}{(optional) Option to change legend position (default = "topright").}

\item{...}{(optional) Other parameters to pass to the \code{plot()} function.}
}
\value{
No return value.
}
\description{
Provides a fitted plot of a CREM model, as returned by \code{Bayes_CREM()}.
}
\examples{
# load fitted model results
data(results_pcrem)
# plot fitted results
plot(results_pcrem)

}
\author{
Corissa T. Rohloff
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BEND.R, R/data.R
\docType{data}
\name{SimData_PREM}
\alias{SimData_PREM}
\title{Simulated data for a PREM + Extensions}
\format{
A data frame with 540 rows and 6 variables.

\subsection{\code{SimData_PREM}}{

A data frame with 540 rows and 6 columns:
\describe{
\item{id}{Individual ids}
\item{time}{Time}
\item{y}{Outcome}
\item{class_pred_1}{First Class-Predictive Covariate}
\item{class_pred_2}{Second Class-Predictive Covariate}
\item{outcome_pred_1}{First Outcome-Predictive Covariate}
}
}
}
\usage{
data(SimData_PREM)

SimData_PREM
}
\description{
Simulated data for a piecewise random effects model (PREM) and useful extensions (CI-PREM, PREMM, CI-PREMM) with 18 timepoints collected on 30 individuals.

A simulated dataset for the piecewise random effects model (PREM).
}
\details{
\itemize{
\item \code{id} ID for each individual.
\item \code{time} Timepoints for each individual.
\item \code{y} Outcome.
\item \code{class_pred_1} First class predictive covariate (time-invariant).
\item \code{class_pred_2} Second class predictive covariate (time-invariant).
\item \code{outcome_pred_1} Outcome predictive covariate (time-varying).
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getClassProb.R
\name{getClassProb}
\alias{getClassProb}
\alias{getClassProb.PREM}
\alias{print.getClassProb.PREM}
\title{Extract class probabilities}
\usage{
getClassProb(x, ...)

\method{getClassProb}{PREM}(x, ...)

\method{print}{getClassProb.PREM}(x, ...)
}
\arguments{
\item{x}{An object of class "PREM".}

\item{...}{Additional arguments.}
}
\value{
Returns a list of the posterior class probabilities for each individual.
}
\description{
Extracts the posterior class probabilities for each individual from a fitted model of class "PREM".
}
\examples{
# load fitted model results
data(results_prem)
# get class probabilities
getClassProb(results_prem)

}
\author{
Corissa T. Rohloff
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print_CREM.R
\name{print.CREM}
\alias{print.CREM}
\title{Print the results of a crossed random effects model (CREM)}
\usage{
\method{print}{CREM}(x, ...)
}
\arguments{
\item{x}{An object of class "CREM" (returned by \code{Bayes_CREM(...)}).}

\item{...}{Additional arguments.}
}
\value{
Returns a list of key parameter estimates.
}
\description{
Provides a summary of a CREM model, as returned by \code{Bayes_CREM()}.
}
\examples{
# load fitted model results
data(results_pcrem)
# print results
print(results_pcrem)

}
\author{
Corissa T. Rohloff
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BEND.R, R/data.R
\docType{data}
\name{SimData_BPREM}
\alias{SimData_BPREM}
\title{Simulated data for a BPREM}
\format{
A data frame with 210 rows and 4 variables.

\subsection{\code{SimData_BPREM}}{

A data frame with 210 rows and 4 columns:
\describe{
\item{id}{Individual ids}
\item{time}{Time}
\item{y1}{Outcome 1}
\item{y2}{Outcome 2}
}
}
}
\usage{
data(SimData_BPREM)

SimData_BPREM
}
\description{
Simulated data for a bivariate piecewise random effects model (BPREM) with 7 timepoints collected on 30 individuals.

A simulated dataset for the bivariate piecewise random effects model (BPREM).
}
\details{
\itemize{
\item \code{id} ID for each individual.
\item \code{time} Timepoints for each individual.
\item \code{y1} Outcome 1.
\item \code{y2} Outcome 2.
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getModelFit.R
\name{getModelFit}
\alias{getModelFit}
\alias{getModelFit.BEND}
\alias{print.getModelFit.BEND}
\title{Extract model fit}
\usage{
getModelFit(x, ...)

\method{getModelFit}{BEND}(x, ...)

\method{print}{getModelFit.BEND}(x, ...)
}
\arguments{
\item{x}{An object of class "BPREM", "CREM", or "PREM".}

\item{...}{Additional arguments.}
}
\value{
Returns a vector of the model fit information (deviance, pD, DIC).
}
\description{
Extracts the model fit information from a fitted model of class "BPREM", "CREM", or "PREM".
}
\examples{
# load fitted model results
data(results_prem)
# get model fit
getModelFit(results_prem)

}
\author{
Corissa T. Rohloff
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_BEND.R
\name{plot_BEND}
\alias{plot_BEND}
\title{Plot a BEND Model (PREM, CREM, BPREM)}
\usage{
plot_BEND(
  data,
  id_var,
  time_var,
  y_var,
  y2_var = NULL,
  results = NULL,
  xlab = "X",
  ylab = "Y",
  colors = NULL,
  mean_colors = NULL,
  legend_pos = "topright",
  ...
)
}
\arguments{
\item{data}{Data frame in long format, where each row describes a measurement occasion for a given individual. It is assumed that each individual has the same number of assigned timepoints (a.k.a., rows).}

\item{id_var}{Name of column that contains ids for individuals with repeated measures in a longitudinal dataset.}

\item{time_var}{Name of column that contains the time variable.}

\item{y_var}{Name of column that contains the outcome variable.}

\item{y2_var}{(for \code{BPREM} only) Name of column that contains the second outcome variable.}

\item{results}{The output of \code{BEND} model to the data. If results=NULL, only a spaghetti plot of the data will be generated.}

\item{xlab}{X-axis label for the generated plot.}

\item{ylab}{Y-axis label for the generated plot.}

\item{colors}{Colors for each class (\code{PREMM} or \code{CI-PREMM}) or outcome (\code{BPREM}). By default, up to 5 colors are provided in the following order: “blue” (class 1 and outcome 1), “red” (class 2 and outcome 2), “green” (class 3), “gold” (class 4), “gray” (class 5).}

\item{mean_colors}{Colors for the trajectory defined by the mean parameters for each class (\code{PREMM} or \code{CI-PREMM}) or outcome (\code{BPREM}). By default, up to 5 colors are provided in the following order: “darkblue” (class 1 and outcome 1), “darkred” (class 2 and outcome 2), “darkgreen” (class 3), “gold4” (class 4), “darkgray” (class 5).}

\item{legend_pos}{(optional) Option to change legend position (default = "topright").}

\item{...}{(optional) Other parameters to pass to the \code{plot()} function.}
}
\value{
No return value.
}
\description{
Generates a "spaghetti plot" of observed longitudinal trajectories for each individual. If the results from a \code{BEND} function are supplied, the trajectory defined by the mean parameters is shown in bold. If fitting a mixture (\code{PREMM} or \code{CI-PREMM}) or bivariate model (\code{BPREM}), the mean trajectories for classes or outcomes will be distinguished by color.
}
\examples{
# load simulated data
data(SimData_PREM)
# plot observed data
plot_BEND(data = SimData_PREM,
          id_var = "id",
          time_var = "time",
          y_var = "y")
# load fitted model results
data(results_prem)
# plot fitted results
plot_BEND(data = SimData_PREM,
          id_var = "id",
          time_var = "time",
          y_var = "y",
          results = results_prem)

}
\author{
Corissa T. Rohloff
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getRanEf.R
\name{getRanEf}
\alias{getRanEf}
\alias{getRanEf.BPREM}
\alias{getRanEf.CREM}
\alias{print.getRanEf.BPREM}
\alias{print.getRanEf.CREM}
\title{Extract random effects}
\usage{
getRanEf(x, ...)

\method{getRanEf}{BPREM}(x, ...)

\method{getRanEf}{CREM}(x, ...)

\method{print}{getRanEf.BPREM}(x, ...)

\method{print}{getRanEf.CREM}(x, ...)
}
\arguments{
\item{x}{An object of class "BPREM" or "CREM".}

\item{...}{Additional arguments.}
}
\value{
Returns a list of the random effects for each individual/group.
}
\description{
Extracts the random effects from a fitted model of class "BPREM" or "CREM".
}
\examples{
# load fitted model results
data(results_bprem)
# get random effects
getRanEf(results_bprem)

}
\author{
Corissa T. Rohloff
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BEND.R, R/data.R
\docType{data}
\name{SimData_PCREM}
\alias{SimData_PCREM}
\title{Simulated data for a PCREM}
\format{
A data frame with 210 rows and 4 variables.

\subsection{\code{SimData_PCREM}}{

A data frame with 210 rows and 4 columns:
\describe{
\item{id}{Individual ids}
\item{teacherid}{Teacher ids}
\item{time}{Time}
\item{y}{Outcome}
}
}
}
\usage{
data(SimData_PCREM)

SimData_PCREM
}
\description{
Simulated data for a piecewise crossed random effects model (PCREM) with 7 timepoints collected on 30 individuals.

A simulated dataset for the piecewise crossed random effects model (PCREM).
}
\details{
\itemize{
\item \code{id} ID for each individual.
\item \code{teacherid} ID for each teacher.
\item \code{time} Timepoints for each individual.
\item \code{y} Outcome.
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getCoef.R
\name{getCoef}
\alias{getCoef}
\alias{getCoef.BEND}
\alias{print.getCoef}
\title{Extract random coefficients}
\usage{
getCoef(x, ...)

\method{getCoef}{BEND}(x, ...)

\method{print}{getCoef}(x, ...)
}
\arguments{
\item{x}{An object of class "BPREM", "CREM", or "PREM".}

\item{...}{Additional arguments.}
}
\value{
Returns a data frame of the random coefficients for each individual/group.
}
\description{
Extracts the random coefficients from a fitted model of class "BPREM", "CREM", or "PREM".
}
\examples{
# load fitted model results
data(results_prem)
# get random coefficients
getCoef(results_prem)

}
\author{
Corissa T. Rohloff
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/BEND.R, R/data.R
\docType{data}
\name{results_pcrem}
\alias{results_pcrem}
\title{Fitted results for a PCREM}
\format{
A list (an object of class \code{CREM}) with fitted model results.

\subsection{\code{results_pcrem}}{

A list object with elements:
\describe{
\item{Call}{Function arguments}
\item{Sample_Size}{Data set sample size}
\item{Data}{Data used for model fitting}
\item{Convergence}{Convergence status}
\item{Model_Fit}{Model fit information}
\item{Fitted_Values}{Fitted outcome values}
\item{Parameter_Estimates}{Parameter estimates}
\item{Random_Coefficients}{Random coefficient estimates}
\item{Run_Time}{Model run time}
}
}
}
\usage{
data(results_pcrem)

results_pcrem
}
\description{
Simulated data for a piecewise crossed random effects model (PCREM) using \code{SimData_CREM}. Included to demonstrate the use of \code{summary.CREM()}.

Results from fitting a piecewise crossed random effects model (PCREM) to \code{SimData_PCREM} (returned by \code{Bayes_CREM(...)}).
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getVarCov.R
\name{getVarCov}
\alias{getVarCov}
\alias{getVarCov.BPREM}
\alias{getVarCov.CREM}
\alias{getVarCov.PREM}
\alias{print.getVarCov.BPREM}
\alias{print.getVarCov.CREM}
\alias{print.getVarCov.PREM}
\title{Extract random effects variance-covariance matrix}
\usage{
getVarCov(x, ...)

\method{getVarCov}{BPREM}(x, ...)

\method{getVarCov}{CREM}(x, ...)

\method{getVarCov}{PREM}(x, ...)

\method{print}{getVarCov.BPREM}(x, ...)

\method{print}{getVarCov.CREM}(x, ...)

\method{print}{getVarCov.PREM}(x, ...)
}
\arguments{
\item{x}{An object of class "BPREM", "CREM", or "PREM".}

\item{...}{Additional arguments.}
}
\value{
Returns a list of the random effects variance-covariance matrices.
}
\description{
Extracts the random effects variance-covariance matrix from a fitted model of class "BPREM", "CREM", or "PREM".
}
\examples{
# load fitted model results
data(results_prem)
# get random effects variance-covariance matrices
getVarCov(results_prem)

}
\author{
Corissa T. Rohloff
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.PREM}
\alias{print.PREM}
\title{Print the results of a piecewise random effects model (PREM)}
\usage{
\method{print}{PREM}(x, ...)
}
\arguments{
\item{x}{An object of class "PREM" (returned by \code{Bayes_PREM(...)}).}

\item{...}{Additional arguments.}
}
\value{
Returns a list of key parameter estimates.
}
\description{
Provides a summary of a PREM model, as returned by \code{Bayes_PREM()}.
}
\examples{
# load fitted model results
data(results_prem)
# print results
print(results_prem)

}
\author{
Corissa T. Rohloff
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary.R
\name{summary.PREM}
\alias{summary.PREM}
\alias{print.summary.PREM}
\title{Summarize the results of a piecewise random effects model (PREM)}
\usage{
\method{summary}{PREM}(object, ...)

\method{print}{summary.PREM}(x, ...)
}
\arguments{
\item{object}{An object of class "PREM" (returned by \code{Bayes_PREM(...)}).}

\item{...}{Additional arguments.}

\item{x}{An object of class "summary.PREM".}
}
\value{
Returns a list of key parameter estimates.
}
\description{
Provides a summary of a PREM model, as returned by \code{Bayes_PREM()}.
}
\examples{
# load fitted model results
data(results_prem)
# result summary
summary(results_prem)

}
\author{
Corissa T. Rohloff
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot.PREM}
\alias{plot.PREM}
\title{Plot the results of a piecewise random effects model (PREM)}
\usage{
\method{plot}{PREM}(
  x,
  xlab = "X",
  ylab = "Y",
  colors = NULL,
  mean_colors = NULL,
  legend_pos = "topright",
  ...
)
}
\arguments{
\item{x}{An object of class "PREM" (returned by \code{Bayes_PREM(...)}).}

\item{xlab}{X-axis label for the generated plot.}

\item{ylab}{Y-axis label for the generated plot.}

\item{colors}{Colors for each class (\code{PREMM} or \code{CI-PREMM}). By default, up to 5 colors are provided in the following order: "blue" (class 1), "red" (class 2), "green" (class 3), "gold" (class 4), "gray" (class 5).}

\item{mean_colors}{Colors for the trajectory defined by the mean parameters for each class (\code{PREMM} or \code{CI-PREMM}). By default, up to 5 colors are provided in the following order: "darkblue" (class 1), "darkred" (class 2), "darkgreen" (class 3), "gold4" (class 4), "darkgray" (class 5).}

\item{legend_pos}{(optional) Option to change legend position (default = "topright").}

\item{...}{(optional) Other parameters to pass to the \code{plot()} function.}
}
\value{
No return value.
}
\description{
Provides a fitted plot of a PREM model, as returned by \code{Bayes_PREM()}.
}
\examples{
# load fitted model results
data(results_prem)
# plot fitted results
plot(results_prem)

}
\author{
Corissa T. Rohloff
}
